//****************************************************************
//
//    Class dense
//
//    Description:
//      Dense is a class for implementing dense matrices.
//
//      If you discover any bugs, it would be very helpful if
//      you sent me e-mail.  Thank you.
//
//      Author:                Peter Braaten
//      e-mail:                braa0007@mermaid.micro.umn.edu
//      Date last modified:    July 18, 1995
//
//****************************************************************

/*==========================================================================
				PRIVATE :
				--------

    int nrow;		// row dimension of the matrix
    int ncol;		// col dimension of the matrix
    double *matrix;	// double array which contains the values of the matrix

				FUNCTIONS :
				---------
constructors
destructor
cnrms : 	        computes the norms of the columns
coscal :		scales the columns of of a dense matrix by their norms
getbwd :	        gives the bandwith of the upper and the lower part 
			of the mat
getbwdu :	        gives the bandwith of the upper part of the matrix
getbwdl :	        gives the bandwith of the lower part of the matrix
getdiag : 		extracts a specified diagonal from a matrix.

get_pointer : 		returns a pointer to the matrix
get_nrow : 		returns the row dimension of a matrix
get_ncol : 		returns the col dimension of a matrix
get_nnz :		returns the number of non zero element of a matrix
getu :			extracts the upper triangular part of a matrix
getl :			extracts the lower triangular part of a matrix
get_element(i,j) : 	returns the element of the matrix with the coordinates
			(i,j)
initialize(x) : 	sets all the elements of the matrix to the value x
make_identity : 	makes the matrix into an identity matrix
rnrms : 	        computes the norms of the rows
rmdiag : 		removes a specified diagonal from a matrix.
roscal :		scales the rows of a dense matrix by their norms
set_element(i,j,v) : 	sets the element (i,j) to the value v
submat :		extracts a submatrix from a dense one.


operators: 		=, <<= , !, +, -, *, &&, ||, <, <=, >, >=, ==, !=

filter(mat,f) : 	Returns a new matrix where each element (i,j) is
			calculated by applying the function to the i,j element
			of mat.
transp : 		calculates the transpose of a matrix
emult : 		multiplies 2 matrices element by element
ediv : 			divides 2 matrices element by element
dot : 			if it is possible, calculates the dot product
output : 		outputs the content of a matrix
==========================================================================*/

#include "spmath.h"

//#define debugging


//******************************************************************
//       Member Constructors
//******************************************************************

dense::dense () {
  // Create an empty dense matrix.

  matrix = new double[1];
  nrow = 0;
  ncol = 0;
  
} // end constructor dense

dense::dense (int i, int j) {
  // Create the zero matrix with the dimensions i,j.
 
  matrix = new double[i * j];
  nrow = i;
  ncol = j;

  initialize_zero();

} // end constructor dense  

dense::dense (const dense & y) {
  // Copy constructor
 
  int i,j;       // Simple loop indecis

  nrow = y.get_nrow();
  ncol = y.get_ncol();
  matrix = new double[nrow*ncol];

  for (i = 1; i <= nrow; i += 1) 
    for (j = 1; j <= ncol; j += 1)
      matrix[(i-1)*ncol + (j-1)] = y.get_element(i,j);

} // end constructor dense  

dense::dense (int i, int j, double x) {
  // Create a matrix with the dimensions i,j and all of the
  // elements initialized to x.
 
  matrix = new double[i * j];
  nrow = i;
  ncol = j;

  initialize(x);

} // end constructor dense  

dense::dense (sparse & mat) {
  // Create a dense matrix from a sparse matrix.
  int i,j,k;

  nrow = mat.get_nrow();
  ncol = mat.get_ncol();
  matrix = new double[nrow*ncol];
  
  initialize_zero ();

  int *mja = mat.get_ja();
  int *mia = mat.get_ia();
  double *ma = mat.get_a();
  int upper = mia[0] -1;

  for (i = 1; i <= nrow; i++) {
    upper = mia[i-1+1] -1;
    for (k = mia[i-1]; k <= upper; k++) {
      j = mja[k-1];
      matrix[(i-1)*ncol + (j-1)] = ma[k-1];
    }
  }

} // end constructor dense




//******************************************************************
//       Member Destructor
//******************************************************************


void dense::free () {
  // Free up memory dynamically allocated.

  if(matrix) { delete [] matrix; matrix = NULL; }

  nrow = 0;
  ncol = 0;
} // end destructor dense

dense::~dense() {
  //  if(matrix) { delete [] matrix; matrix = NULL; }
}

//******************************************************************
//       Member Selectors
//******************************************************************

/*=========================================================================
#			GET_POINTER :
#			-------------
# Entries : none
# Returns : the pointer to the matrix
#
=========================================================================*/

double* dense::get_pointer () {
  // Return the pointer to the matrix.


  return matrix;
} // end get_pointer

/*=========================================================================
#			GET_NROW :
#			---------
# Entries : none
# Returns : the row dimension of the matrix
#
=========================================================================*/

int dense::get_nrow () const {
  // Return the row dimension of the matrix.

  return nrow;
} // end get_nrow
 
/*=========================================================================
#			GET_NCOL :
#			---------
# Entries : none
# Returns : the col dimension the matrix
#
=========================================================================*/

int dense::get_ncol () const {
  // Return the col dimension of the matrix.

  return ncol;
} // end get_ncol

/*=========================================================================
#			GET_NNZ :
#			---------
# Entries : none
# Returns : the number of non zero elements of the matrix
#
=========================================================================*/

int dense::get_nnz () const {
  // Return the number of non zero elements.

  int i,j;
  int nnz = 0;

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++)
      if (matrix[(i-1)*ncol + (j-1)] != 0)
        nnz += 1;

  return nnz;
} // end get_ncol

/*============================================================================
#				GET_ELEMENT		
#				-----------    	
# in entry  : 	i -> row of the element
#		j -> column of the element
#
# in return : The value of the element whose coordinates are (i,j)
#
# details   :
#
#   Since the matrix is represented as a one dimensional
#   array, the index is calculated by (i-1)(ncol) + (j-1)
#   Consider the following matrix.
#  
#                 1    2    3    4
#               --------------------
#           1 |   0    1    2    3
#             |  
#           2 |   4    5    6    7
#             |
#           3 |   8    9    10   11
#  
#   The index of the element 2,3 is 6.  The formula 
#   (i-1)(ncol) + (j-1) yields  (2-1)(4) + (3-1) = 6
#   as expected.
#
#
============================================================================*/


double dense::get_element (int i, int j) const {
  // Return the element with the  coordinates i,j.
  
  #ifdef debugging
    assert (i <= nrow  &&  i > 0  && j <= ncol  && j > 0);
  #endif

  return matrix[(i-1)*ncol + (j-1)];

} // end get_element


int dense::size() {
  return (sizeof(double) * (nrow*ncol) +
	  sizeof(*this));
}


/*============================================================================
#				GET_ELEMENT_REF		
#				---------------   	
# in entry  : 	i -> row of the element
#		j -> column of the element
#
# in return : The value of the element whose coordinates are (i,j)
#
# details   :
#
#   Since the matrix is represented as a one dimensional
#   array, the index is calculated by (i-1)(ncol) + (j-1)
#   Consider the following matrix.
#  
#                 1    2    3    4
#               --------------------
#           1 |   0    1    2    3
#             |  
#           2 |   4    5    6    7
#             |
#           3 |   8    9    10   11
#  
#   The index of the element 2,3 is 6.  The formula 
#   (i-1)(ncol) + (j-1) yields  (2-1)(4) + (3-1) = 6
#   as expected.
#
#
============================================================================*/




double& dense::get_element_ref (int i, int j) {
  // Return the element at the position i,j by  reference.
  // See documentation for get_element

  #ifdef debugging
    assert (i <= nrow  &&  i > 0  && j <= ncol  && j > 0);
  #endif

  return matrix[(i-1)*ncol + (j-1)];
} // end get_element_ref 
  

//******************************************************************
//       Member Functions
//******************************************************************

/*============================================================================
#				SET_ELEMENT		
#				-----------    	
# in entry  : 	i     -> row of the element
#		j     -> column of the element
#		value -> value you want to give to the element
#
# in return :	nothing
#
# details   :	see documentation for get_element
#
#
============================================================================*/

void dense::set_element (int i, int j, double value) {
  // Set the element i,j to value.

  #ifdef debugging
    assert (i <= nrow  &&  i > 0  && j <= ncol  && j > 0);
  #endif

  // See documentation for get_element.
  matrix[(i-1)*ncol + (j-1)] = value;
} // end set_element

/*============================================================================
#			INITIALIZE_ZERO		
#			----------------    	
# in entry  : nothing
#
# in return : nothing
#
# details   : sets all the elements of the matrix to the value 0
#
============================================================================*/

void dense::initialize_zero () {
  // Initialize the matrix to the zero matrix.
  int i, j;        // Simple loop counters

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++)
      matrix[(i-1)*ncol + (j-1)] = 0;
} // end initialize_zero      

/*============================================================================
#				INITIALIZE		
#				----------    	
# in entry  : x -> value you want to give to all the elements 
#
# in return : nothing
#
# details   : sets all the elements to the value x
#
============================================================================*/

void dense::initialize (double x) {
  // Initialize the matrix so that all of the elements are x.
  int i, j;        // Simple loop counters

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++)
      matrix[(i-1)*ncol + (j-1)] = x;
} // end initialize_zero      

/*============================================================================
#			MAKE_IDENTITY		
#			-------------    	
# in entry  : nothing 
#
# in return : nothing
#
# details   : makes the matrix into an identity matrix
#
============================================================================*/

void dense::make_identity () {
  // Make the matrix into an identity matrix.

  int i;             // Simple loop index.

  #ifdef debugging
    assert (nrow == ncol);
  #endif

  // Make everything zeroes
  initialize_zero();

  // Make the main diagonal ones.
  for (i = 1; i <= nrow; i++) 
    matrix[(i-1)*ncol + (i-1)] = 1;

} // end make_identity
  
//******************************************************************
//       Member Operators
//******************************************************************

/*============================================================================
#				OPERATOR =		
#				----------    	
# in entry  : y -> a dense matrix 
#
# in return : nothing
#
# details   : copies the value of y into the dense matrix.
#		- free the memory from the old matrix 
#		- create a new matrix with y's dimensions
#		- copy all the elements of y in the matrix (this[i,j] = y[i,j])
#
============================================================================*/



void dense::operator = (const dense & y) {
  // Copy the value of y into this dense matrix.

  int i,j;          // Simple loop counter.

  // Free the memory (if any) from the old matrix.
  delete [] matrix;

  nrow = y.get_nrow();
  ncol = y.get_ncol();
  matrix = new double[nrow * ncol];

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++)
      matrix[(i-1)*ncol + (j-1)] = y.get_element(i,j);

} // end operator =

/*============================================================================
#				OPERATOR <<=		
#				------------    	
# in entry  : y -> a dense matrix 
#
# in return : nothing
#
# details   : copies the value of the pointer to y into the pointer to the 
# 	      matrix
#
============================================================================*/

void dense::operator <<= (dense & y) {
  // Copy the value of y into this dense matrix.  The buffer
  // for the matrix is not copied. 

  // Free the memory (if any) from the old matrix.
  delete [] matrix;

  nrow = y.get_nrow();
  ncol = y.get_ncol();
  matrix = y.get_pointer();

} // end operator <<=


/*============================================================================
#				SUBMAT		
#				------    	
# in entry  : 	n -> the number of rows of the matrix
#	      	i1, i2, j1, j2 -> the indecis of the first and the last 
#		element to extract 
#
# in return : 	nothing
#
# details   : 	It just extracts the elements of the matrix between (i1, j1)
#		and (i2, j2) and puts the result in the same matrix!
#
============================================================================*/

void dense::submat(int n, int i1, int i2, int j1, int j2)
{


  // Store the values of the matrix
	dense input(*this);

  
  //Allocate enough space for the results
	free();
	nrow = i2 - i1 + 1;
	ncol = j2 - j1 + 1;
	matrix = new double[nrow*ncol];
	
  // Set the values of the submatrix:
	int i, j;
	for (i=1;i<=nrow;i++)
	{
		for (j=1;j<=ncol;j++)
		{
		  matrix[(i-1)*ncol +j-1] = input.get_element(i+i1-1,j+j1-1);
		}
	}

  // Free the memory  
	input.free();
}

/*============================================================================
#				GETDIAG		
#				------    	
# in entry  : 	ioff -> int (number of the diagonal) 
#
# in return : 	nothing
#
# details   : 	It extracts a diagonal from a dense matrix
#		- it stores the matrix (input)
#		- it frees the memory of the old matrix
#		- it allocates enough memory to store the diagonal in the
#		  old matrix (as a dense matrix)
#		- It calculates the diagonal (two different possibilities 
#		  according the sign of ioff)
#		- It frees the memory of input
#
============================================================================*/

void dense::getdiag(int ioff)

{

  // Store the matrix

	dense *input=new dense(*this);
	int n=nrow - abs(ioff);

  // Free the memory of the matrix

	free();

  // Allocate enough space for the diag
	matrix= new double [n];
	nrow=1;
	ncol=n;

  // Calculate the diagonal

int i;
if (ioff>=0)
	{for (i=0; i<n; i++)
		{ set_element(1,i+1,input->get_element(i + 1, i + ioff + 1)); } }

else 	{for (i=0; i<n; i++)
		{ set_element(1,i+1,input->get_element(i-ioff+1,i+1)); } }

  input->free();

}

/*============================================================================
#				RMDIAG		
#				------    	
# in entry  : 	ioff -> int (number of the diagonal) 
#
# in return : 	nothing
#
# details   : 	It removes the diagonal from a dense matrix
#		- n is the number of elements of this diagonal
#		- There are 2 differents possibilities according to 
#		  the sign of ioff
#
============================================================================*/

void dense::rmdiag(int ioff)

{
int n=nrow - abs(ioff);
int i;
if (ioff>=0)
	{
	for (i=0; i<n; i++)
		{ set_element(i + 1, i + ioff + 1, 0); }
	}
else {
	for (i=0; i<n; i++)
		{ set_element(i-ioff+1,i+1, 0); }
	}
}

/*============================================================================
#				GETL/GETU		
#				---------    	
# in entry  : 	nothing 
#
# in return : 	nothing
#
# details   : 	It extracts either the lower or the upper triangular part 
#		of the dense matrix, which means that it sets to zero all
#		the other elements.
#
============================================================================*/

void dense::getu()
{
int i,j;
for (i=2;i<=nrow;i++)
	{
	for (j=1;j<=i-1;j++)
		{set_element(i,j,0); }
	}
}

void dense::getl()
{
int i,j;
for (i=1;i<nrow;i++)
	{
	for (j=i+1;j<=nrow;j++)
		{set_element(i,j,0); }
	}
}

/*============================================================================
#				  GETBWD		
#				  ------    	
# in entry  : 	nothing 
#
# in return : 	int * or int
#
# details   : 	It returns the bandwith of the upper and of the lower part 
#		of the matrix
#		- getbwdu returns the bandwith of the upper part
#		- getbwdl returns the bandwith of the lower part
#		- getbwd returns an array containing 2 elements: the first
#		  one is the bandwith of the lower part and the 2nd is the
#		  bandwith of the upper part
#
============================================================================*/

int * dense::getbwd()
	{
	int *res = new int[2];
	res[0] = getbwdl();
	res[1] = getbwdu();
	return res;
	
	}
int dense::getbwdu()
	{
	int max=(int)(nrow/2);
	int i,j;
	int res=0;
	int startpt, endpt;
	for (i=1;i<=max;i++)
		{
		startpt=0;endpt=0;
		for (j=1;j<=ncol;j++)
			{
			if (get_element(i,j)!=0) 
			   	{
			    	if (startpt == 0) {startpt =j ; endpt = j;}
				else endpt = j; 
				}
			}
		if (endpt-startpt>res) res=endpt-startpt;
		}
	return (res);
	}

int dense::getbwdl()
	
	{
	int min=nrow - (int)(nrow/2) + 1;
	int i,j;
	int res=0;
	int startpt, endpt;
	for (i=min;i<=nrow;i++)
		{
		startpt=0;endpt=0;
		for (j=1;j<=ncol;j++)
			{
			if (get_element(i,j)!=0) 
			   	{
			    	if (startpt == 0) {startpt =j ; endpt = j;}
				else endpt = j; 
				}
			}
		if (endpt-startpt>res) res=endpt-startpt;
		}
	return (res);
	}
	
	
/*============================================================================
#				CNRMS :		
#				------    	
# in entry  :   nrm -> norm indicator 
#
# in return :	a double array (norms of each columns)
#
# details   :	It computes the norms of the columns of a matrix
#		- If nrm = 0, it returns the max of each column
#		- If nrm = 1, it returns the sum of the elements of each column
#		- If nrm = 2, it returns the sqrt of the sum of the sqr of 
#		  the elements of each column
#		- If nrm = 3, it returns the sum of the sqr of the elements
#		  of each column
#
============================================================================*/

double * dense::cnrms(int nrm)
{

double * result=new double[ncol];
int i,j;

if (nrm == 0)
	{
	double max;
	
	for (j=1;j<=ncol;j++)
		{
		max = get_element(1,j);
		for (i=2;i<=nrow;i++)
			{ if (get_element(i,j) > max) max=get_element(i,j);} 
		result[j-1] = max;
		}

	return result;

	}

if (nrm == 1)
	{
	double sum;
	for (j=1;j<=ncol;j++)
		{
		sum = get_element(1,j);
		for (i=2;i<=nrow;i++)
			{ sum = sum + get_element(i,j);} 
		result[j-1] = sum;		
		} 
	return result;
	
	}

if (nrm == 2)

{
	double sum2;
	for (j=1;j<=ncol;j++)
		{
		sum2 = pow(get_element(1,j),2) ;
		for (i=2;i<=nrow;i++)
			{ sum2 = sum2 + pow(get_element(i,j),2);} 
		result[j-1] = pow(sum2,0.5);		
		} 
	return result;
	
	}


else {
	double sum2;
	for (j=1;j<=ncol;j++)
		{
		sum2 = pow(get_element(1,j),2) ;
		for (i=2;i<=nrow;i++)
			{ sum2 = sum2 + pow(get_element(i,j),2);} 
		result[j-1] = sum2;		
		} 
	return result;
	
	}

}

/*============================================================================
#				RNRMS :		
#				------    	
# in entry  :   nrm -> norm indicator 
#
# in return :	a double array (norms of each rows)
#
# details   :	It computes the norms of the rows of a matrix
#		- If nrm = 0, it returns the max of each row
#		- If nrm = 1, it returns the sum of the elements of each row
#		- If nrm = 2, it returns the sqrt of the sum of the sqr of 
#		  the elements of each row
#		- If nrm = 3, it returns the sum of the sqr of the elements
#		  of each row
#
============================================================================*/

double * dense::rnrms(int nrm)
{

double * result=new double[nrow];
int i,j;

if (nrm == 0)
	{
	double max;
	
	for (i=1;i<=nrow;i++)
		{
		max = get_element(i,1);
		for (j=2;j<=ncol;j++)
			{ if (get_element(i,j) > max) max=get_element(i,j);} 
		result[i-1] = max;
		}

	return result;

	}

if (nrm == 1)
	{
	double sum;
	for (i=1;i<=nrow;i++)
		{
		sum = get_element(i,1);
		for (j=2;j<=ncol;j++)
			{ sum = sum + get_element(i,j);} 
		result[i-1] = sum;		
		} 
	return result;
	
	}

if (nrm == 2)

{
	double sum2;
	for (i=1;i<=nrow;i++)
		{
		sum2 = pow(get_element(i,1),2) ;
		for (j=2;j<=ncol;j++)
			{ sum2 = sum2 + pow(get_element(i,j),2);} 
		result[i-1] = pow(sum2,0.5);		
		} 
	return result;
	
	}


else {
	double sum2;
	for (i=1;i<=nrow;i++)
		{
		sum2 = pow(get_element(i,1),2) ;
		for (j=2;j<=ncol;j++)
			{ sum2 = sum2 + pow(get_element(i,j),2);} 
		result[i-1] = sum2;		
		} 
	return result;
	
	}

}

/*============================================================================
#			     COSCAL/ROSCAL :		
#			     -------------    	
# in entry  :   nrm -> norm indicator 
#
# in return :	nothing
#
# details   :	It scales either the rows or the columns of a dense matrix
#		by their norms
#
============================================================================*/

void dense::roscal(int nrm)

{
	double *norm = new double[nrow];
	norm = rnrms(nrm);

	int i,j;
	for (i=1;i<=nrow;i++)
		{
		for (j=1;j<=ncol;j++)
			{
			set_element(i,j,get_element(i,j)/(norm[i-1]));	
			}
		}
	delete [] norm;

}

void dense::coscal(int nrm)

{
	double *norm = new double[ncol];
	norm = cnrms(nrm);

	int i,j;
	for (i=1;i<=nrow;i++)
		{
		for (j=1;j<=ncol;j++)
			{
			set_element(i,j,get_element(i,j)/(norm[j-1]));	
			}
		}
	delete [] norm;

}

/*============================================================================
#			       RANDOM :		
#			       -------    	
# in entry  :   max -> all the elements will be in [0,max] 
#
# in return :	nothing
#
# details   :	It generates a random dense matrix. All its elements are 
#		random double between 0 and max. 
#
============================================================================*/

void dense::rando(int max)
{
int i;
int j;
for (i=0;i<nrow;i++)
	{
	for (j=0;j<ncol;j++) 
		{ set_element(i+1,j+1,rnd0nb(max,0.0)); }
	}
}

//******************************************************************
//      Operators
//******************************************************************

/*============================================================================
#				OPERATOR !		
#				----------    	
# in entry  : x -> a dense matrix 
#
# in return : a dense matrix whose elements are either 0 or 1.
#
# details   : calculates the logical not of a matrix element by element:
#		the matrix returned by this function is calculated as follows:
#		result[i,j]=!(x[i,j])	
#
============================================================================*/

dense operator ! (const dense &x) {
  // Calculate the logical not of a matrix element by element.

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);     // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++)
      if (x.get_element (i,j) == 0)
        z.set_element (i,j,1);
      else 
	z.set_element (i,j,0);

  return z;

} // end operator !

/*============================================================================
#				OPERATOR +		
#				----------    	
# in entry  : y -> a dense matrix
#	      x -> a dense matrix 
#
# in return : a dense matrix : result
#
# details   : calculates result = x + y
#	      	- it checks if the two entries have the same dimension.
#		- it creates a new matrix (result) with the same dimensions
#		- it adds the matrices element by element, which means that
#		  result[i,j] = x[i,j] + y[i,j]
#
============================================================================*/

dense operator + (const dense &x, const dense &y) {
  // Calculate z = x + y and return the result.

  #ifdef debugging
    assert (x.get_nrow() == y.get_nrow());
    assert (x.get_ncol() == y.get_ncol());
  #endif

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);     // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++)
      z.set_element(i,j, x.get_element(i,j) + 
	                 y.get_element(i,j));

  return z;
} // end operator +

/*============================================================================
#				OPERATOR +		
#				----------    	
# in entry  : y -> a dense matrix
#	      x -> a double 
#
# in return : a dense matrix : result
#
# details   : calculates result = x + y
#		- it creates a new matrix (result) with the same dimensions as y
#		- it adds x to each element of the matrix y, which means that
#		  result[i,j] = x + y[i,j]
#
============================================================================*/

dense operator + (const double &x, const dense &y) {
  // Calculate z = x + y and return the result.

  int nrow = y.get_nrow();    // Row dimension of the matrix.
  int ncol = y.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);     // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++)
      z.set_element(i,j, x + y.get_element(i,j));

  return z;
} // end operator +

/*============================================================================
#				OPERATOR +		
#				----------    	
# in entry  : x -> a dense matrix
#	      y -> a double 
#
# in return : a dense matrix : result
#
# details   : calculates result = x + y
#		- it creates a new matrix (result) with the same dimensions as y
#		- it adds y to each element of the matrix x, which means that
#		  result[i,j] = y + x[i,j]
#
============================================================================*/

dense operator + (const dense &x, const double &y) {
  // Calculate z = x + y and return the result.

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);     // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++)
      z.set_element(i,j, x.get_element(i,j) + y);

  return z;
} // end operator +


/*============================================================================
#				OPERATOR -		
#				----------    	
# in entry  : y -> a dense matrix
#	      x -> a dense matrix 
#
# in return : a dense matrix : result
#
# details   : calculates result = x - y
#	      	- it checks if the two entries have the same dimension.
#		- it creates a new matrix (result) with the same dimensions
#		- it adds the matrices element by element, which means that
#		  result[i,j] = x[i,j] - y[i,j]
#
============================================================================*/

dense operator - (const dense & x, const dense &y) {
  // Calculate z = x - y and return the result.

  #ifdef debugging
    assert (x.get_nrow() == y.get_nrow());
    assert (x.get_ncol() == y.get_ncol());
  #endif

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow,  ncol);    // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++)
      z.set_element(i,j, x.get_element(i,j) - 
			 y.get_element(i,j));

  return z;
} // end operator -

/*============================================================================
#				OPERATOR -		
#				----------    	
# in entry  : x -> a dense matrix
#	      y -> a double 
#
# in return : a dense matrix : result
#
# details   : calculates result = x - y
#		- it creates a new matrix (result) with the same dimensions as x
#		- it substracts y to each element of the matrix y, which means
#		  that result[i,j] = x[i,j] - y
#
============================================================================*/

dense operator - (const dense & x, const double &y) {
  // Calculate z = x - y and return the result.

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow,  ncol);    // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++)
      z.set_element(i,j, x.get_element(i,j) - y);

  return z;
} // end operator -

/*============================================================================
#				OPERATOR +		
#				----------    	
# in entry  : y -> a dense matrix
#	      x -> a double 
#
# in return : a dense matrix : result
#
# details   : calculates result = x - y
#		- it creates a new matrix (result) with the same dimensions as y
#		- it substracts each element of the matrix y to x, which means
#		  that
#		  result[i,j] = x - y[i,j]
#
============================================================================*/

dense operator - (const double & x, const dense &y) {
  // Calculate z = x - y and return the result.

  int nrow = y.get_nrow();    // Row dimension of the matrix.
  int ncol = y.get_ncol();    // Column dimension of the matrix.
  dense z(nrow,  ncol);    // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++)
      z.set_element(i,j, x - y.get_element(i,j));

  return z;
} // end operator -

/*============================================================================
#				OPERATOR +		
#				----------    	
# in entry  : x -> a dense matrix 
#
# in return : a dense matrix : result
#
# details   : calculates result = -x
#		- it creates a new matrix (result) with the same dimensions as x
#		- result[i,j] = x + y[i,j]
#
============================================================================*/

dense operator - (const dense & x) {
  // Calculate z = -x and return the result.

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow,  ncol);    // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++)
      z.set_element(i, j, -x.get_element(i,j));

  return z;
} // end operator -

/*============================================================================
#				OPERATOR *		
#				----------    	
# in entry  : y -> a dense matrix
#	      x -> a dense matrix
#
# in return : a dense matrix : result
#
# details   : calculates result = x * y
#		- it checks if the two entries have the right dimensions
#		- it creates a new matrix (result) with the right dimension
#		- it calculates x*y
#
============================================================================*/

dense operator * (const dense & x, const dense &y) {
  // Calculate z = x * y and return the result. An 
  // element of z is calculated with the following.
  //         x.ncol
  // z(i,j) =   E   x(i,k) * y(k,j)
  //           k=1


  #ifdef debugging
    assert (x.get_ncol() == y.get_nrow());
  #endif

  int xnrow = x.get_nrow();
  int xncol = x.get_ncol();
  // int ynrow = y.get_nrow();
  int yncol = y.get_ncol();

  dense z (xnrow, yncol);   // Matrix to hold the result.
  int i,j,k;                                 // Simple loop counters.
  double sum;                                     // Tempory summing variable

  for (i = 1; i <= xnrow; i++) {
    for (j = 1; j <= yncol; j++) {
      // Calculate the value of z(i,j) int sum.
      sum = 0;
      for (k = 1; k <= xncol; k++) 
        sum = sum + x.get_element(i,k) * y.get_element(k,j);
      z.set_element(i,j,sum);
    } // end for
  } // end for

  return z;
} // end operator *

/*============================================================================
#				OPERATOR *		
#				----------    	
# in entry  : a -> a double
#	      x -> a dense matrix
#
# in return : a dense matrix : result
#
# details   : calculates result = a * y
#		- it creates a new matrix (result) with the dimensions of y
#		- it multiplies each element of y by a, which means that
#		  result[i,j] = a* y[i,j]
#
============================================================================*/

dense operator * (const double & a, const dense & x) {
  // Calculate y = ax and return the result.


  int xnrow = x.get_nrow();
  int xncol = x.get_ncol();
  dense y (xnrow, xncol);  // Array to hold the  result in.
  int i,j;                                  // Simple  loop counter.

  
  for (i = 1; i <= xnrow; i++) 
    for (j = 1; j <= xncol; j++) 
      y.set_element (i,j, a * x.get_element(i,j));

  return y;
} // end operator *

/*============================================================================
#				OPERATOR *		
#				----------    	
# in entry  : a -> a double
#	      x -> a dense matrix
#
# in return : a dense matrix : result
#
# details   : calculates result = a * y
#		- it creates a new matrix (result) with the dimensions of y
#		- it multiplies each element of y by a, which means that
#		  result[i,j] = a* y[i,j]
#
============================================================================*/

dense operator * (const dense & x, const double & a) {
  // Calculate y = ax and return the result.

  return a * x;
} // end operator *

/*============================================================================
#				OPERATOR /		
#				----------    	
# in entry  : a -> a double
#	      x -> a dense matrix
#
# in return : a dense matrix : result
#
# details   : calculates result = y/a
#		- it calls (1/a) * x
#
============================================================================*/

dense operator / (const dense & x, const double & a) {
  // Calculate y = x/a and return the result.

  return (1/a) * x;
} // end operator *

/*============================================================================
#				OPERATOR &&		
#				-----------    	
# in entry  : y -> a dense matrix
#	      x -> a double
#
# in return : a dense matrix : result
#
# details   : calculates result = x && y
#		- it creates a new matrix (result) with the dimensions of y
#		- it calculates the logical and of each element with x
#		  result[i,j] = x && y[i,j]
#
============================================================================*/

dense operator && (const double &x, const dense &y) {
  // Calculate x && y.

  int nrow = y.get_nrow();    // Row dimension of the matrix.
  int ncol = y.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
      if (x == 0 || y.get_element(i,j) == 0)
        z.set_element(i,j, 0);
      else
	z.set_element(i,j, 1);
    } // end for

  return z;
  
} // end operator &&

/*============================================================================
#				OPERATOR &&		
#				-----------    	
# in entry  : x -> a dense matrix
#	      y -> a double
#
# in return : a dense matrix : result
#
# details   : calculates result = x && y
#		- it creates a new matrix (result) with the dimensions of x
#		- it calculates the logical and of each element with y
#		  result[i,j] = y && x[i,j]
#
============================================================================*/

dense operator && (const dense &x, const double &y) {
  // Calculate x && y.

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
      if (x.get_element(i,j) == 0 || y == 0)
        z.set_element(i,j, 0);
      else
	z.set_element(i,j, 1);
    } // end for

  return z;
  
} // end operator &&

/*============================================================================
#				OPERATOR &&		
#				-----------    	
# in entry  : y -> a dense matrix
#	      x -> a dense matrix
#
# in return : a dense matrix : result
#
# details   : calculates result = x && y
#		- it checks if the 2 entries have the same dimensions
#		- it creates a new matrix (result) with the dimensions of y
#		- result[i,j] = x[i,j] && y[i,j]
#
============================================================================*/

dense operator && (const dense &x, const dense &y) {
  // Calculate x && y.

  #ifdef debugging
    assert (x.get_nrow() == y.get_nrow());
    assert (x.get_ncol() == y.get_ncol());
  #endif

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
      if (x.get_element(i,j) == 0 || y.get_element(i,j) == 0)
        z.set_element(i,j, 0);
      else
	z.set_element(i,j, 1);
    } // end for

  return z;
  
} // end operator &&

/*============================================================================
#				OPERATOR ||		
#				-----------    	
# in entry  : y -> a dense matrix
#	      x -> a double
#
# in return : a dense matrix : result
#
# details   : calculates result = x || y
#		- it creates a new matrix (result) with the dimensions of y
#		- it calculates the logical OR of each element with x
#		  result[i,j] = x || y[i,j]
#
============================================================================*/

dense operator || (const double &x, const dense &y) {
  // Calculate x || y.

  int nrow = y.get_nrow();    // Row dimension of the matrix.
  int ncol = y.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
      if (x != 0 || y.get_element(i,j) != 0)
        z.set_element(i,j,1);
      else
	z.set_element(i,j,0);
    } // end for

  return z;
  
} // end operator ||

/*============================================================================
#				OPERATOR ||		
#				-----------    	
# in entry  : x -> a dense matrix
#	      y -> a double
#
# in return : a dense matrix : result
#
# details   : calculates result = x || y
#		- it creates a new matrix (result) with the dimensions of x
#		- it calculates the logical OR of each element with y
#		  result[i,j] = y || x[i,j]
#
============================================================================*/

dense operator || (const dense &x, const double &y) {
  // Calculate x || y.

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
      if (x.get_element(i,j) != 0 || y != 0)
        z.set_element(i,j, 1);
      else
	z.set_element(i,j, 0);
    } // end for

  return z;
  
} // end operator ||

/*============================================================================
#				OPERATOR ||		
#				-----------    	
# in entry  : y -> a dense matrix
#	      x -> a dense matrix
#
# in return : a dense matrix : result
#
# details   : calculates result = x || y
#		- it checks if the 2 entries have the same dimensions
#		- it creates a new matrix (result) with the dimensions of y
#		- result[i,j] = x[i,j] || y[i,j]
#
============================================================================*/

dense operator || (const dense &x, const dense &y) {
  // Calculate x || y.

  #ifdef debugging
    assert (x.get_nrow() == y.get_nrow());
    assert (x.get_ncol() == y.get_ncol());
  #endif

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
      if (x.get_element(i,j) != 0 || y.get_element(i,j) != 0)
        z.set_element(i,j, 1);
      else
	z.set_element(i,j, 0);
    } // end for

  return z;
  
} // end operator ||

/*============================================================================
#				OPERATOR <		
#				-----------    	
# in entry  : y -> a dense matrix
#	      x -> a double
#
# in return : a dense matrix : result
#
# details   : calculates result = x < y
#		- it creates a new matrix (result) with the dimensions of y
#		- result[i,j] = x < y[i,j]
#
============================================================================*/

dense operator < (const double &x, const dense &y) {
  // Calculate x < y.

  int nrow = y.get_nrow();    // Row dimension of the matrix.
  int ncol = y.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
        z.set_element(i,j, x < y.get_element(i,j));
    } // end for

  return z;
  
} // end operator <

/*============================================================================
#				OPERATOR <		
#				-----------    	
# in entry  : x -> a dense matrix
#	      y -> a double
#
# in return : a dense matrix : result
#
# details   : calculates result = x < y
#		- it creates a new matrix (result) with the dimensions of y
#		- result[i,j] = x[i,j] < y
#
============================================================================*/

dense operator < (const dense &x, const double &y) {
  // Calculate x < y.

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
        z.set_element(i,j, x.get_element(i,j) < y);
    } // end for

  return z;
  
} // end operator <

/*============================================================================
#				OPERATOR <		
#				-----------    	
# in entry  : x -> a dense matrix
#	      y -> a dense matrix
#
# in return : a dense matrix : result
#
# details   : calculates result = x < y
#		- it checks if the 2 entries have the same dimensions
#		- it creates a new matrix (result) with those dimensions
#		- result[i,j] = x[i,j] < y[i,j]
#
============================================================================*/

dense operator < (const dense &x, const dense &y) {
  // Calculate x < y.

  #ifdef debugging
    assert (x.get_nrow() == y.get_nrow());
    assert (x.get_ncol() == y.get_ncol());
  #endif

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
	z.set_element(i,j, x.get_element(i,j) < y.get_element(i,j));
    } // end for

  return z;
  
} // end operator <

/*============================================================================
#				OPERATOR <=		
#				-----------    	
# in entry  : y -> a dense matrix
#	      x -> a double
#
# in return : a dense matrix : result
#
# details   : calculates result = x <= y
#		- it creates a new matrix (result) with the dimensions of y
#		- result[i,j] = x <= y[i,j]
#
============================================================================*/

dense operator <= (const double &x, const dense &y) {
  // Calculate x <= y.

  int nrow = y.get_nrow();    // Row dimension of the matrix.
  int ncol = y.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
        z.set_element(i,j, x <= y.get_element(i,j));
    } // end for

  return z;
  
} // end operator <=

/*============================================================================
#				OPERATOR <=		
#				-----------    	
# in entry  : x -> a dense matrix
#	      y -> a double
#
# in return : a dense matrix : result
#
# details   : calculates result = x <= y
#		- it creates a new matrix (result) with the dimensions of y
#		- result[i,j] = x[i,j] <= y
#
============================================================================*/

dense operator <= (const dense &x, const double &y) {
  // Calculate x <= y.

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
        z.set_element(i,j, x.get_element(i,j) <= y);
    } // end for

  return z;
  
} // end operator <=


/*============================================================================
#				OPERATOR <=		
#				-----------    	
# in entry  : x -> a dense matrix
#	      y -> a dense matrix
#
# in return : a dense matrix : result
#
# details   : calculates result = x <= y
#		- it checks if the 2 entries have the same dimensions
#		- it creates a new matrix (result) with those dimensions
#		- result[i,j] = x[i,j] <= y[i,j]
#
============================================================================*/

dense operator <= (const dense &x, const dense &y) {
  // Calculate x <= y.

  #ifdef debugging
    assert (x.get_nrow() == y.get_nrow());
    assert (x.get_ncol() == y.get_ncol());
  #endif

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
	z.set_element(i,j, x.get_element(i,j) <= y.get_element(i,j));
    } // end for

  return z;
  
} // end operator <=

/*============================================================================
#				OPERATOR ==		
#				-----------    	
# in entry  : y -> a dense matrix
#	      x -> a double
#
# in return : a dense matrix : result
#
# details   : calculates result = x == y
#		- it creates a new matrix (result) with the dimensions of y
#		- result[i,j] = x == y[i,j]
#
============================================================================*/

dense operator == (const double &x, const dense &y) {
  // Calculate x == y.

  int nrow = y.get_nrow();    // Row dimension of the matrix.
  int ncol = y.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
        z.set_element(i,j, x == y.get_element(i,j));
    } // end for

  return z;
  
} // end operator ==

/*============================================================================
#				OPERATOR ==		
#				-----------    	
# in entry  : x -> a dense matrix
#	      y -> a double
#
# in return : a dense matrix : result
#
# details   : calculates result = y == x
#		- it creates a new matrix (result) with the dimensions of y
#		- result[i,j] = x[i,j] == y
#
============================================================================*/

dense operator == (const dense &x, const double &y) {
  // Calculate x == y.

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
        z.set_element(i,j, x.get_element(i,j) == y);
    } // end for

  return z;
  
} // end operator ==


/*============================================================================
#				OPERATOR ==		
#				-----------    	
# in entry  : x -> a dense matrix
#	      y -> a dense matrix
#
# in return : a dense matrix : result
#
# details   : calculates result = x == y
#		- it checks if the 2 entries have the same dimensions
#		- it creates a new matrix (result) with those dimensions
#		- result[i,j] = x[i,j] == y[i,j]
#
============================================================================*/

dense operator == (const dense &x, const dense &y) {
  // Calculate x == y.

  #ifdef debugging
    assert (x.get_nrow() == y.get_nrow());
    assert (x.get_ncol() == y.get_ncol());
  #endif

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
	z.set_element(i,j, x.get_element(i,j) == y.get_element(i,j));
    } // end for

  return z;
  
} // end operator ==

/*============================================================================
#				OPERATOR !=		
#				-----------    	
# in entry  : y -> a dense matrix
#	      x -> a double
#
# in return : a dense matrix : result
#
# details   : calculates result = x != y
#		- it creates a new matrix (result) with the dimensions of y
#		- result[i,j] = x != y[i,j]
#
============================================================================*/

dense operator != (const double &x, const dense &y) {
  // Calculate x != y.

  int nrow = y.get_nrow();    // Row dimension of the matrix.
  int ncol = y.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
        z.set_element(i,j, x != y.get_element(i,j));
    } // end for

  return z;
  
} // end operator !=

/*============================================================================
#				OPERATOR!=		
#				-----------    	
# in entry  : x -> a dense matrix
#	      y -> a double
#
# in return : a dense matrix : result
#
# details   : calculates result = y != x
#		- it creates a new matrix (result) with the dimensions of y
#		- result[i,j] = x[i,j] != y
#
============================================================================*/

dense operator != (const dense &x, const double &y) {
  // Calculate x != y.

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
        z.set_element(i,j, x.get_element(i,j) != y);
    } // end for

  return z;
  
} // end operator !=


/*============================================================================
#				OPERATOR !=		
#				-----------    	
# in entry  : x -> a dense matrix
#	      y -> a dense matrix
#
# in return : a dense matrix : result
#
# details   : calculates result = x != y
#		- it checks if the 2 entries have the same dimensions
#		- it creates a new matrix (result) with those dimensions
#		- result[i,j] = x[i,j] != y[i,j]
#
============================================================================*/

dense operator != (const dense &x, const dense &y) {
  // Calculate x != y.

  #ifdef debugging
    assert (x.get_nrow() == y.get_nrow());
    assert (x.get_ncol() == y.get_ncol());
  #endif

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
	z.set_element(i,j, x.get_element(i,j) != y.get_element(i,j));
    } // end for

  return z;
  
} // end operator !=

/*============================================================================
#				OPERATOR >=		
#				-----------    	
# in entry  : y -> a dense matrix
#	      x -> a double
#
# in return : a dense matrix : result
#
# details   : calculates result = x >= y
#		- it creates a new matrix (result) with the dimensions of y
#		- result[i,j] = x >= y[i,j]
#
============================================================================*/


dense operator >= (const double &x, const dense &y) {
  // Calculate x >= y.

  int nrow = y.get_nrow();    // Row dimension of the matrix.
  int ncol = y.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
        z.set_element(i,j, x >= y.get_element(i,j));
    } // end for

  return z;
  
} // end operator >=

/*============================================================================
#				OPERATOR >=		
#				-----------    	
# in entry  : x -> a dense matrix
#	      y -> a double
#
# in return : a dense matrix : result
#
# details   : calculates result = x >= y
#		- it creates a new matrix (result) with the dimensions of y
#		- result[i,j] = x[i,j] >= y
#
============================================================================*/

dense operator >= (const dense &x, const double &y) {
  // Calculate x >= y.

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
        z.set_element(i,j, x.get_element(i,j) >= y);
    } // end for

  return z;
  
} // end operator >=


/*============================================================================
#				OPERATOR >=		
#				-----------    	
# in entry  : x -> a dense matrix
#	      y -> a dense matrix
#
# in return : a dense matrix : result
#
# details   : calculates result = x >= y
#		- it checks if the 2 entries have the same dimensions
#		- it creates a new matrix (result) with those dimensions
#		- result[i,j] = x[i,j] >= y[i,j]
#
============================================================================*/

dense operator >= (const dense &x, const dense &y) {
  // Calculate x >= y.

  #ifdef debugging
    assert (x.get_nrow() == y.get_nrow());
    assert (x.get_ncol() == y.get_ncol());
  #endif

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
	z.set_element(i,j, x.get_element(i,j) >= y.get_element(i,j));
    } // end for

  return z;
  
} // end operator >=

/*============================================================================
#				OPERATOR >		
#				-----------    	
# in entry  : y -> a dense matrix
#	      x -> a double
#
# in return : a dense matrix : result
#
# details   : calculates result = x > y
#		- it creates a new matrix (result) with the dimensions of y
#		- result[i,j] = x > y[i,j]
#
============================================================================*/

dense operator > (const double &x, const dense &y) {
  // Calculate x > y.

  int nrow = y.get_nrow();    // Row dimension of the matrix.
  int ncol = y.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
        z.set_element(i,j, x > y.get_element(i,j));
    } // end for

  return z;
  
} // end operator >

/*============================================================================
#				OPERATOR >		
#				-----------    	
# in entry  : x -> a dense matrix
#	      y -> a double
#
# in return : a dense matrix : result
#
# details   : calculates result = x > y
#		- it creates a new matrix (result) with the dimensions of y
#		- result[i,j] = x[i,j] > y
#
============================================================================*/

dense operator > (const dense &x, const double &y) {
  // Calculate x < y.

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
        z.set_element(i,j, x.get_element(i,j) > y);
    } // end for

  return z;
  
} // end operator >


/*============================================================================
#				OPERATOR >		
#				-----------    	
# in entry  : x -> a dense matrix
#	      y -> a dense matrix
#
# in return : a dense matrix : result
#
# details   : calculates result = x > y
#		- it checks if the 2 entries have the same dimensions
#		- it creates a new matrix (result) with those dimensions
#		- result[i,j] = x[i,j] > y[i,j]
#
============================================================================*/

dense operator > (const dense &x, const dense &y) {
  // Calculate x > y.

  #ifdef debugging
    assert (x.get_nrow() == y.get_nrow());
    assert (x.get_ncol() == y.get_ncol());
  #endif

  int nrow = x.get_nrow();    // Row dimension of the matrix.
  int ncol = x.get_ncol();    // Column dimension of the matrix.
  dense z(nrow, ncol);        // Matrix to hold the result.
  int i, j;                   // Simple loop counters.

  for (i = 1; i <= nrow; i++)
    for (j = 1; j <= ncol; j++) {
	z.set_element(i,j, x.get_element(i,j) > y.get_element(i,j));
    } // end for

  return z;
  
} // end operator >


//******************************************************************
//       Functions
//******************************************************************


/*============================================================================
#				FILTER		
#				------    	
# in entry  : mat -> a dense matrix
#	      funct -> a function whose argument is a double and which also
#	      	       returns a double
#
# in return : a dense matrix : result
#
# details   : It returns a matrix where element (i,j) is calculating by 
#	      applying the function funct to the (i,j) element of mat
#		- it creates a matrix with the same dimensions as mat
#		- result[i,j] = funct (mat[i,j])
#
============================================================================*/

dense filter (const dense &mat, ffunc funct) {
  // Return a new matrix where each element (i,j) is  calculated
  // by applying the function funct to the i,j element of mat.

  int i,j;          // Simple loop index.
  int nrow = mat.get_nrow();
  int ncol = mat.get_ncol();
  dense result (nrow, ncol);

  for (i = 1; i <= nrow; i++) 
    for (j = 1; j <= ncol; j++)
      result.set_element(i,j, funct(mat.get_element(i,j)));

  return result;

} // end filter

/*============================================================================
#				TRANSP		
#				------    	
# in entry  : mat -> a dense matrix
#	    
# in return : a dense matrix : result
#
# details   : It calculates the transpose matrix of the matrix mat.
#		- it creates a matrix with the same dimension as mat : result
#		- result[i,j] = mat[j,i]
#
============================================================================*/

dense transp (const dense & mat) {
  // Calculate the transpose of a matrix, and return the result.

  int i,j;          // Simple loop index.
  int xnrow = mat.get_nrow();
  int xncol = mat.get_ncol();
  dense result (xncol, xnrow);

  for (i = 1; i <= xnrow; i++) 
    for (j = 1; j <= xncol; j++)
      result.get_element_ref(j,i) = mat.get_element (i,j);

  return result; 
} // end transp

/*============================================================================
#				EMULT		
#				------    	
# in entry  : left -> a dense matrix
#	      right -> a dense matrix
#
# in return : a dense matrix : result
#
# details   : It multiplies the 2 entries element by element
#		- it checks if the 2 entries have the same dimensions
#		- it creates a matrix with the same dimensions as mat
#		- result[i,j] = left[i,j]*right[i,j]
#
============================================================================*/

dense emult (const dense &left, const dense & right) {
  // Multiply two matrices element by element.

  #ifdef debugging
    assert (left.get_ncol() == right.get_ncol());
    assert (left.get_nrow() == right.get_nrow());
  #endif
  
  int i,j;
  int nrow = left.get_nrow();
  int ncol = left.get_ncol();
  dense result (nrow, ncol);

  for (i = 1; i <= nrow; i++) 
    for (j = 1; j <= ncol; j++)
      result.get_element_ref(i,j) = left.get_element (i,j) * 
				    right.get_element (i,j);
  
  return result;

} // end emult


/*============================================================================
#				EDIV		
#				-----    	
# in entry  : left -> a dense matrix
#	      right -> a dense matrix
#
# in return : a dense matrix : result
#
# details   : It divides the 2 entries element by element
#		- it checks if the 2 entries have the same dimensions
#		- it creates a matrix with the same dimensions as mat
#		- result[i,j] = left[i,j]/right[i,j]
#
============================================================================*/

dense ediv (const dense &left, const dense & right) {
  // Divides two matrices element by element.

  #ifdef debugging
    assert (left.get_ncol() == right.get_ncol());
    assert (left.get_nrow() == right.get_nrow());
  #endif
  
  int i,j;
  int nrow = left.get_nrow();
  int ncol = left.get_ncol();
  dense result (nrow, ncol);

  for (i = 1; i <= nrow; i++) 
    for (j = 1; j <= ncol; j++)
      result.get_element_ref(i,j) = left.get_element (i,j) / 
				    right.get_element (i,j);
  
  return result;

} // end ediv


/*============================================================================
#				DOT		
#				---    	
# in entry  : left -> a dense matrix
#	      right -> a dense matrix
#
# in return : a double : result
#
# details   : It calculates the dot product of the 2 dense matrices
#		- it checks if the 2 entries are vectors and have the same size
#		- result = sum (left[i,1]*right[i,1])
#
============================================================================*/

double dot (const dense & left, const dense & right) {
  // If both of the matrices are vectors (ncol == 1), and they have
  // the same size, calculate the dot product and return  it.

  #ifdef debugging
    assert (left.get_ncol() == 1);
    assert (right.get_ncol() == 1);
    assert (left.get_nrow() == right.get_nrow());
  #endif

  int i;             //   Simple loop index.
  double result = 0;      //   The result of the dot product.
  int lnrow = left.get_nrow();

  for (i = 1; i <= lnrow; i++) 
    result = result + left.get_element(i,1) * right.get_element(i,1);

  return result;

} // end dot


/*============================================================================
#				OUTPUT	
#				------    	
# in entry  : mat -> a dense matrix
#
# in return : nothing
#
# details   : It outputs all the elements of the matrix mat
#
============================================================================*/

void output (dense & mat) {
  // Print out the contents of the matrix mat.

  int i, j;                   // Simple loop counters
  int nrow = mat.get_nrow();  // The row dimension of the matrix. 
  int ncol = mat.get_ncol();  // The col dimension of the matrix. 
  

  for (i = 1; i <= nrow; i++) {
    for (j = 1; j <= ncol; j++) {
      double el = mat.get_element(i, j);
      sp_out(el);
      sp_out("\t");
    } // end for
    sp_out("\n");
  } // end for

} // end output

int absol (int x) {
	
	if (x>=0) return x;
	else return -x; 

}

    
/*======================= PERM ============================================*/

dense cperm (dense &x, int *perm) {
  // Calculate the column permutation of a dense matrix
  // and return the result.

  dense y;
  y = x;
  int i,j;
  for (j=1;j<=x.get_ncol();j++)
	{
	for (i=1;i<=x.get_nrow();i++)
		{
		if ((perm[j-1]>0) && (perm[j-1]<=x.get_ncol()))
		y.set_element(i, perm[j-1], x.get_element(i, j));
		

		}
	}

  

  return y;

} // end cperm

dense rperm (dense &x, int *perm) {
  // Calculate the row permutation of a dense matrix
  // and return the result.

  dense y;
  y = x;
  int i,j;
for (i=1;i<=x.get_nrow();i++)  
	{
	for (j=1;j<=x.get_ncol();j++)

		{
		if ((perm[i-1]>0) && (perm[i-1]<=x.get_nrow()))
		y.set_element(perm[i-1], j, x.get_element(i,j));
		

		}
	}

  

  return y;

} // end cperm

dense dperm (dense &x, int *perm, int *qperm) {

  dense d = rperm(x, perm);
return ( cperm (d, qperm ) );

}

