// ****************************************************************
//
//    Builtin Functions
//
//    Description:
//	Contains all of the builtin functions.
//
//      If you discover any bugs, it would be very helpful if
//      you sent me e-mail.  Thank you.
//
//      Author:                Peter Braaten
//      e-mail:                braa0007@mermaid.micro.umn.edu
//      Date last modified:    July 18, 1995
//
// ****************************************************************

/*===========================================================================
				FUNCTIONS :

				----------
csort :				sorts the elements in increasing order 
				of columns
coscal : 			scales the columns of a matrix by their norms
drawmat : 			Shows the non-zero pattern of a sparse matrix
				and allows you to zoom.
eye(arg) : 			returns an identity matrix
function(arg,f) : 		applies the function f to arg
getdia : 			extracts a specified diagonal from a matrix.
gen57pt :			generates a 5-point and 7-point matrix
getu :				extracts the upper triangular part of a matrix
getl :				extracts the lower triangular part of a matrix
getbwd :	   		gives the bandwith of the upper and the lower
				part of the mat
getbwdu :	   		gives the bandwith of the upper part of matrix
getbwdl :	   		gives the bandwith of the lower part of matrix
cnrms : 			computes the norms of the columns
get_element(name,args) : 	gets the element of the matrix name whose
				coordinates are stored in args
indset0 :			greedy algorithm for independent set ordering
indset1 :			idem  but using minimal degree traversal
filter	:			filters the elements from a matrix according to 
				their magnitude
help(topic) : 			provides a documentation on the topic
multic,colors,colors2 : 	greedy algorithm for multicoloring (the only 
				difference between these functions is what they 
				return)
prtmt(mat,file) : 		prints the matrix mat int he Boeing-Harwell file
readmt(file) : 			reads the matrix stored in the Boeing-Harwell
				file
rmdia : 			removes a specified diagonal from a matrix.
roscal : 			scales the rows of a matrix by their norms
rperm(mat,perm) : 		permutes mat with perm
set_element(name,args,elt) : 	sets the element of the matrix name whose
				coordinates are stored in args to the value 
solve_system :			Solve a linear system Ax=y where A is sparse.
                                (now with 3-arg version)
submat :			extracts a submatrix from a matrix
vec : 				creates a vector 
zeroes(arg) : 			returns a matrix composed with zeroes elts
===========================================================================*/

#include "spmath.h"

/*============================================================================
#				FUNCTION :		
#				----------    	
# in entry  :   - a value : arg (the value you will apply the function to)
#		- a function whose argument is a double and which returns also
#		  a double
#
# in return :	a value : result 
#
# details   :	There are five possibilities :
#		 - arg=NUM      -> result = funct (arg)
#		 - arg=DENSE    -> it calls the function filter of the class 
#		   		   dense.c and so returns a dense matrix where
#		   		   result[i,j] = funct ( arg[i,j] )
#		 - arg = SPARSE -> if funct(0) = 0, then the result will be a
# 				   sparse matrix, so it calls the function
#				   filter of the class sparse.c 
#				   else it calls the function filter of dense.c
#		 - else 	-> it returns an error
#
============================================================================*/

value * function (value *arg, ffunc func, int &err) {
  // Apply the function func to arg and return the result.

  if (arg->type == NUM) 
    return new value (func(arg->num));

  if (arg->type == DENSE) 
    return new value (new dense (filter (*arg->dmat,func)));

  if (arg->type == SPARSE) 
    if (func(0) == 0)
      return new value (new sparse (filter (*arg->smat,func)));
    else
      return new value (new dense (filter (dense(*arg->smat),func)));


  err = 1;
  return new value (0);

} // end function

/*============================================================================
#				RPERM :		
#				-------    	
# in entry  :   a value : mat
#		a value : perm
#		a value : result
#
# in return :	a value : result
#
# details   :	- it checks if mat is a Sparse Matrix
#		- it checks if perm is a vector (verifies that perm is a dense
#		  matrix and that at least one of its dimension is equal to 1 )
#		- it distinguishes 2 similar cases : ncol=1 or nrow=1
#		  if NCOL=1
#		    + it checks if the row dimension of mat is the same as the
#		      length of the permutation vector.	
#		    + it stored the values of perm in a vector p
#		  if NROW=1
#		    + it checks if the row dimension of mat is the same as the
#		      length of the permutation vector.	
#		    + it stored the values of perm in a vector p	
#		- it calls the function rperm(mat,p) of sparse.c
#		- it returns the sparse matrix	
#		  
============================================================================*/

value * rperm (value *mat, value *perm, int &err) {
  // Permute mat with perm


  if (!(  (mat->type == SPARSE) || (mat->type == DENSE))) {
    err = 1;
    yyerror ("Error, argument 1 of rperm must be a matrix.");
    return mat;
  } // end if

  if (perm->type != DENSE) {
    err = 1;
    yyerror ("Error, argument 2 of rperm must be a vector.");
    return mat;
  } // end if
      
  int prow = perm->dmat->get_nrow();
  int pcol = perm->dmat->get_ncol();

  if (prow != 1 && pcol != 1) {
    err = 1;
    yyerror ("Error, argument 2 of rperm must be a vector.");
    return mat;
  } // end if

  int *p = NULL; // NULL to avoid warning messages

  if (pcol == 1) {
    if (prow != mat->smat->get_nrow()) {
      err = 1;
      yyerror ("Error, in function rperm, row dimension of matrix must be the same as the length of the permutation vector.");
      return mat;
    } // end if

    p = new int [prow];
    int i;

    for (i = 0; i <= prow -1; i += 1) 
      p[i] = int (perm->dmat->get_element(i+1,1));
  } // end if

  if (prow == 1) {
    if (pcol != mat->smat->get_nrow()) {
      err = 1;
      yyerror ("Error, in function rperm, row dimension of matrix must be the same as the length of the permutation vector.");
      return mat;
    } // end if

    p = new int [pcol];
    int i;

    for (i = 0; i <= pcol -1; i += 1) 
      p[i] = int (perm->dmat->get_element(1,i+1));
  } // end if

if (mat->type == SPARSE)

  return new value (new sparse (rperm (*mat->smat, p)));

else 

  return new value (new dense (rperm (*mat->dmat, p)));

} // end rperm

value * cperm (value *mat, value *perm, int &err) {
  // Permute mat with perm


   if (!(  (mat->type == SPARSE) || (mat->type == DENSE))) {
    err = 1;
    yyerror ("Error, argument 1 of cperm must be a matrix.");
    return mat;
  } // end if

  if (perm->type != DENSE) {
    err = 1;
    yyerror ("Error, argument 2 of cperm must be a vector.");
    return mat;
  } // end if
      
  int prow = perm->dmat->get_nrow();
  int pcol = perm->dmat->get_ncol();

  if (prow != 1 && pcol != 1) {
    err = 1;
    yyerror ("Error, argument 2 of cperm must be a vector.");
    return mat;
  } // end if

  int *p = NULL;

  if (pcol == 1) {
    if (prow != mat->smat->get_ncol()) {
      err = 1;
      yyerror ("Error, in function cperm, col dimension of matrix must be the same as the length of the permutation vector.");
      return mat;
    } // end if

    p = new int [prow];
    int i;

    for (i = 0; i <= prow -1; i += 1) 
      p[i] = int (perm->dmat->get_element(i+1,1));
  } // end if

  if (prow == 1) {
    if (pcol != mat->smat->get_ncol()) {
      err = 1;
      yyerror ("Error, in function cperm, col dimension of matrix must be the same as the length of the permutation vector.");
      return mat;
    } // end if

    p = new int [pcol];
    int i;

    for (i = 0; i <= pcol -1; i += 1) 
      p[i] = int (perm->dmat->get_element(1,i+1));
  } // end if

if (mat->type == SPARSE)
  return new value (new sparse (cperm (*mat->smat, p)));

else 

  return new value (new dense (cperm (*mat->dmat, p)));

} // end rperm

value * dperm (value *mat, value *perm, value *qperm ,int &err) {
  // Permute mat with perm


   if (!(  (mat->type == SPARSE) || (mat->type == DENSE))) {
    err = 1;
    yyerror ("Error, argument 1 of dperm must be a matrix.");
    return mat;
  } // end if

/*------------------- treat perm*/
  if (perm->type != DENSE) {
    err = 1;
    yyerror ("Error, argument 2 of rperm must be a vector.");
    return mat;
  } // end if
      
  int prow = perm->dmat->get_nrow();
  int pcol = perm->dmat->get_ncol();

  if (prow != 1 && pcol != 1) {
    err = 1;
    yyerror ("Error, argument 2 of rperm must be a vector.");
    return mat;
  } // end if

  int *p = NULL;

  if (pcol == 1) {
    if (prow != mat->smat->get_nrow()) {
      err = 1;
      yyerror ("Error, in function rperm, row dimension of matrix must be the same as the length of the permutation vector.");
      return mat;
    } // end if

    p = new int [prow];
    int i;

    for (i = 0; i <= prow -1; i += 1) 
      p[i] = int (perm->dmat->get_element(i+1,1));
  } // end if

  if (prow == 1) {
    if (pcol != mat->smat->get_nrow()) {
      err = 1;
      yyerror ("Error, in function dperm, row dimension of matrix must be the same as the length of the permutation vector.");
      return mat;
    } // end if

    p = new int [pcol];
    int i;

    for (i = 0; i <= pcol -1; i += 1) 
      p[i] = int (perm->dmat->get_element(1,i+1));
  } // end if

/*------------------- treat qperm*/
  if (qperm->type != DENSE) {
    err = 1;
    yyerror ("Error, argument 3 of dperm must be a vector.");
    return mat;
  } // end if
      
  int qrow = qperm->dmat->get_nrow();
  int qcol = qperm->dmat->get_ncol();

  if (qrow != 1 && qcol != 1) {
    err = 1;
    yyerror ("Error, argument 3 of dperm must be a vector.");
    return mat;
  } // end if

  int *q = NULL;

  if (qcol == 1) {
    if (qrow != mat->smat->get_ncol()) {
      err = 1;
      yyerror ("Error, in function dperm, col dimension of matrix must be the same as the length of the permutation vector.");
      return mat;
    } // end if

    q = new int [qrow];
    int i;

    for (i = 0; i <= qrow -1; i += 1) 
      q[i] = int (qperm->dmat->get_element(i+1,1));
  } // end if

  if (qrow == 1) {
    if (qcol != mat->smat->get_ncol()) {
      err = 1;
      yyerror ("Error, in function dperm, col dimension of matrix must be the same as the length of the permutation vector.");
      return mat;
    } // end if

    q = new int [qcol];
    int i;

    for (i = 0; i <= qcol -1; i += 1) 
      q[i] = int (qperm->dmat->get_element(1,i+1));
  } // end if

if (mat->type == SPARSE)
  return new value (new sparse (dperm (*mat->smat, p, q, 3)));

else 
  return new value (new dense (dperm (*mat->dmat, p,q)));

} // end rperm
  
/*============================================================================
#				EYE :
#				-----		
#				   	
# in entry  :   a value    : arg (the dimension of the future matrix)
#		an integer : an error
#
# in return :	a value    : an identity matrix whose size is arg 
#
# details   :	- if arg=NUM then it will be the dimension of the matrix
#		- if arg=DENSE, the dimension of the returned matrix will 
#		  be the row dimension of arg
#		- if arg=SPARSE, idem
#		- it calls the function init_identity of sparse.c and returns
#		  its result
#
============================================================================*/

value * eye (value *arg, int &err) {
  // Return an identity matrix.  
  // If arg is a number the size will be that number.
  // If arg is a matrix, the size will be the row dimension
  // of that matrix.

  sparse * sp = new sparse;

  if (arg->type == NUM) {
    sp->init_identity(int(arg->num));
    return new value (sp);
  } // end if

  if (arg->type == DENSE) {
    sp->init_identity(int(arg->dmat->get_nrow()));
    return new value (sp);
  } // end if

  if (arg->type == SPARSE) {
    sp->init_identity(int(arg->smat->get_nrow()));
    return new value (sp);
  } // end if

  yyerror ("Error, invalid type for function eye.");
  return new value();

} // end eye

/*============================================================================
#				ZEROES :
#				-------		
#				   	
# in entry  :   a value    : arg (the dimension of the future matrix)
#		an integer : an error
#
# in return :	a value    : a matrix composed of zeroes 
#
# details   :	- if arg=NUM then the returned matrix will be square and its
#		  dimension will be arg
#		- if arg=DENSE, the dimensions of the returned matrix will 
#		  be the same as arg
#		- if arg=SPARSE, idem
#		- it calls the function init_zero of sparse.c and returns
#		  its result
#
============================================================================*/

value * zeroes (value *arg, int &err) {
  // Return a dense matrix composed of zeroes.  
  // If arg is a number the size will be that number.
  // If arg is a matrix, the size will be the same size as
  // the matrix

  dense *d;

  if (arg->type == NUM) {
    d = new dense(int(arg->num), int(arg->num), 0.0);
    return new value (d);
  } // end if

  if (arg->type == DENSE) {
    d = new dense(arg->dmat->get_nrow(), arg->dmat->get_ncol(), 0.0);
    return new value (d);
  } // end if

  if (arg->type == SPARSE) {
    d = new dense(arg->smat->get_nrow(), arg->smat->get_ncol(), 0.0);
    return new value (d);
  } // end if

  yyerror ("Error, invalid paramater type for function zeroes.");
  return new value();

} // end zeroes


value * spzeroes (value *arg, int &err) {
  // Return a matrix composed of zeroes.  
  // If arg is a number the size will be that number.
  // If arg is a matrix, the size will be the same size as
  // the matrix

  sparse * sp = new sparse;

  if (arg->type == NUM) {
    sp->init_zero(int(arg->num), int(arg->num));
    return new value (sp);
  } // end if

  if (arg->type == DENSE) {
    sp->init_zero(arg->dmat->get_nrow(), arg->dmat->get_ncol());
    return new value (sp);
  } // end if

  if (arg->type == SPARSE) {
    sp->init_zero(arg->smat->get_nrow(), arg->smat->get_ncol());
    return new value (sp);
  } // end if

  yyerror ("Error, invalid paramater type for function zeroes.");
  return new value();

} // end zeroes

/*============================================================================
#				ZEROES :
#				-------		
#				   	
# in entry  :   2 values    : arg1 and arg2 (the dimensions of the future
#			      matrix)
#		an integer : an error
#
# in return :	a value    : a matrix composed of zeroes 
#
# details   :	- if arg1=NUM then the row dimension of the returned matrix 
#		  will be arg1
#		- if arg1=DENSE or SPARSE, the row dimension of the returned
#		  matrix will be the same as arg
#		- if arg2=NUM then the col dimension of the returned matrix 
#		  will be arg1
#		- if arg2=DENSE or SPARSE, the col dimension of the returned
#		  matrix will be the same as arg
#		- it calls the function init_zero of sparse.c and returns
#		  its result
#
============================================================================*/

value * zeroes (value *arg1, value *arg2, int &err) {
  // Return a matrix composed of zeroes.  
  // Arg1 is the row dimension, and arg2 is the column dimension
  // If arg1 is a matrix, use its row dimension
  // If arg2 is a matrix, use its col dimension

  dense *d;
  int nrow = 0, ncol = 0;

  if (arg1->type == NUM) {
    nrow = int(arg1->num);
  } // end if

  if (arg2->type == NUM) {
    ncol = int(arg2->num);
  } // end if

  if (arg1->type == DENSE) {
    nrow = arg1->dmat->get_nrow();
  } // end if

  if (arg2->type == DENSE) {
    ncol = arg1->dmat->get_ncol();
  } // end if

  if (arg1->type == SPARSE) {
    nrow = arg1->smat->get_nrow();
  } // end if

  if (arg2->type == SPARSE) {
    ncol = arg1->smat->get_ncol();
  } // end if

  if (nrow > 0 && ncol > 0) {
    d = new dense(nrow, ncol, 0.0);
    return new value (d);
  } // end if

  yyerror ("Error, invalid type or dimension for function zeroes.");
  return new value();

} // end zeroes

value * spzeroes (value *arg1, value *arg2, int &err) {
  // Return a matrix composed of zeroes.  
  // Arg1 is the row dimension, and arg2 is the column dimension
  // If arg1 is a matrix, use its row dimension
  // If arg2 is a matrix, use its col dimension

  sparse * sp = new sparse;
  int nrow = 0, ncol = 0;

  if (arg1->type == NUM) {
    nrow = int(arg1->num);
  } // end if

  if (arg2->type == NUM) {
    ncol = int(arg2->num);
  } // end if

  if (arg1->type == DENSE) {
    nrow = arg1->dmat->get_nrow();
  } // end if

  if (arg2->type == DENSE) {
    ncol = arg1->dmat->get_ncol();
  } // end if

  if (arg1->type == SPARSE) {
    nrow = arg1->smat->get_nrow();
  } // end if

  if (arg2->type == SPARSE) {
    ncol = arg1->smat->get_ncol();
  } // end if


  if (nrow > 0 && ncol > 0) {
    sp->init_zero(nrow, ncol);
    return new value (sp);
  } // end if

  yyerror ("Error, invalid type or dimension for function zeros.");
  return new value();

} // end zeroes


value * ones (value *arg, int &err) {
  // Return a dense matrix composed of ones.  
  // If arg is a number the size will be that number.
  // If arg is a matrix, the size will be the same size as
  // the matrix

  dense *d;

  if (arg->type == NUM) {
    d = new dense(int(arg->num), int(arg->num), 1);
    return new value (d);
  } // end if

  if (arg->type == DENSE) {
    d = new dense(arg->dmat->get_nrow(), arg->dmat->get_ncol(), 1);
    return new value (d);
  } // end if

  if (arg->type == SPARSE) {
    d = new dense(arg->smat->get_nrow(), arg->smat->get_ncol(), 1);
    return new value (d);
  } // end if

  yyerror ("Error, invalid paramater type for function ones.");
  return new value();
}

value * ones (value *arg1, value *arg2, int &err) {
  // Return a matrix composed of ones.  
  // Arg1 is the row dimension, and arg2 is the column dimension
  // If arg1 is a matrix, use its row dimension
  // If arg2 is a matrix, use its col dimension

  dense *d;
  int nrow = 0, ncol = 0;

  if (arg1->type == NUM) {
    nrow = int(arg1->num);
  } // end if

  if (arg2->type == NUM) {
    ncol = int(arg2->num);
  } // end if

  if (arg1->type == DENSE) {
    nrow = arg1->dmat->get_nrow();
  } // end if

  if (arg2->type == DENSE) {
    ncol = arg1->dmat->get_ncol();
  } // end if

  if (arg1->type == SPARSE) {
    nrow = arg1->smat->get_nrow();
  } // end if

  if (arg2->type == SPARSE) {
    ncol = arg1->smat->get_ncol();
  } // end if

  if (nrow > 0 && ncol > 0) {
    d = new dense(nrow, ncol, 1);
    return new value (d);
  } // end if

  yyerror ("Error, invalid type or dimension for function ones.");
  return new value();
}

/*============================================================================
#				SET_ELEMENT		
#				-----------    	
# in entry  :   a pointer to a char : name (name of the element you want to set)
#		a list              : args (the coordinates)
#		a value 	    : elem (the value to give to the element)
#		an integer	    : err  (an error) 
#
# in return :	a value		    : 0 if there is an error, otherwise elem
#
# details   :   It sets the element (whose coordinates are stored in args) of
#		the matrix (whose name is name) to the value elem
#	
#		- it checks if the type for the element is a number
#		- checks to make sure that the matrix name exists (looks if 
#		  symtab contains the name)
#		- it stores the value of the matrix in the variable mat
#		- if mat is a DENSE MATRIX
#		    + checks if there is at least one argument in args
#		    + checks if the args are numbers
#		    + stores the row coordinates in i (top value of the arg) 
#		      and remove it from the list arg
#		    + if there were only 1 argument, checks if mat is really a 
# 		      vector
#		    + checks if i doesn't exceed the matrix dimensions
#		    + stores the col coordinates in j (top value of arg) and
#		      removes it from the list arg	
#		    + checks if there is no more arguments in the list arg
#		    + checks if j doesn't exceed the matrix dimensions
#		    + calls the function set_element of dense.c and returns elem
#		- if mat is a SPARSE MATRIX
#		    + idem but calls the function of sparse.c
#
============================================================================*/

value * set_element (char *name, list *args, value *elem, int &err) {
  // Return the element of a matrix.  
  int i,j;

  // Check to make sure that the type for the elem is a number.
  if (elem->type != NUM) {
    args->remove_all();
    yyerror ("Error, matrix elements must be numbers.");
    err = 1;
    return new value ((0));
  } // end if

  // Check to make  sure that the matrix exists
  if (!syms->contains(name)) {
    args->remove_all();
    yyerror ("Error, undefined matrix");
    err = 1;
    return new value ((0));
  }
  value mat (syms->lookup(name).get_value());

  if (syms->lookup(name).get_value().type == DENSE) {
   
    // There must be at least one argument.
    if (args->empty()) { 
      yyerror ("Error, invalid number of indecis for matrix");
      err = 1;
      return new value ((0));
    } // end if

    // Indecis must be numbers
    if (args->top().get_value().type != NUM) {
      args->remove_all();
      yyerror ("Error, invalid type for matrix index.");
      err = 1;
      return new value ((0));
    } // end if
    
    // The row index
    i = int(args->top().get_value().num);
    args->remove_top();

    // If there is only one index, then check to see if the matrix
    // is really a vector.
    if (args->empty()) { 
      if (mat.dmat->get_nrow() == 1) {
        if (i < 1 || i > mat.dmat->get_ncol()) {
          yyerror ("Error, index exceeds matrix dimensions.");
	  err = 1;
          return new value ((0));
	} // end if
        mat.dmat->set_element (1,i,elem->num);
        return new value ((mat.dmat->get_element(1,i)));
      } // end if
      if (mat.dmat->get_ncol() == 1) {
        if (i < 1 || i > mat.dmat->get_nrow()) {
          yyerror ("Error, index exceeds matrix dimensions.");
	  err = 1;
          return new value ((0));
	} // end if
        mat.dmat->set_element (i,1,elem->num);
        return new value ((mat.dmat->get_element(i,1)));
      } // end if
      yyerror ("Error, invalid number of indecis for matrix");
      err = 1;
      return new value ((0));
    } // end if

    // Indecis must be numbers.
    if (args->top().get_value().type != NUM) {
      args->remove_all();
      yyerror ("Error, invalid type for matrix index.");
      err = 1;
      return new value ((0));
    } // end if
    
    // The collumn index.
    j = int(args->top().get_value().num);
    args->remove_top();

    // Maximum of two indecies are allowed.
    if (!args->empty()) { 
      yyerror ("Error, invalid number of indecis for matrix");
      err = 1;
      return new value ((0));
    } // end if

    // Check the dimensions.
    if (i < 1 || i > mat.dmat->get_nrow() || j < 1 || j > mat.dmat->get_ncol()) {
      yyerror ("Error, index exceeds matrix dimensions.");
      err = 1;
      return new value ((0));
    } // end if

    // Return the matrix element
    mat.dmat->set_element (i,j,elem->num);
    return new value ((mat.dmat->get_element(i,j)));
  } // end if

  if (syms->lookup(name).get_value().type == SPARSE) {
   
    // There must be at least one argument.
    if (args->empty()) { 
      yyerror ("Error, invalid number of indecis for matrix");
      err = 1;
      return new value ((0));
    } // end if

    // Indecis must be numbers
    if (args->top().get_value().type != NUM) {
      args->remove_all();
      yyerror ("Error, invalid type for matrix index.");
      err = 1;
      return new value ((0));
    } // end if
    
    // The row index
    i = int(args->top().get_value().num);
    args->remove_top();

    // If there is only one index, then check to see if the matrix
    // is really a vector.
    if (args->empty()) { 
      if (mat.smat->get_nrow() == 1) {
        if (i < 1 || i > mat.smat->get_ncol()) {
          yyerror ("Error, index exceeds matrix dimensions.");
	  err = 1;
          return new value ((0));
	} // end if
        mat.smat->set_element (1,i,elem->num);
        return new value ((mat.smat->get_element(1,i)));
      } // end if
      if (mat.smat->get_ncol() == 1) {
        if (i < 1 || i > mat.smat->get_nrow()) {
          yyerror ("Error, index exceeds matrix dimensions.");
	  err = 1;
          return new value ((0));
	} // end if
        mat.smat->set_element (i,1,elem->num);
        return new value ((mat.smat->get_element(i,1)));
      } // end if
      yyerror ("Error, invalid number of indecis for matrix");
      err = 1;
      return new value ((0));
    } // end if

    // Indecis must be numbers.
    if (args->top().get_value().type != NUM) {
      args->remove_all();
      yyerror ("Error, invalid type for matrix index.");
      err = 1;
      return new value ((0));
    } // end if
    
    // The collumn index.
    j = int(args->top().get_value().num);
    args->remove_top();

    // Maximum of two indecies are allowed.
    if (!args->empty()) { 
      yyerror ("Error, invalid number of indecis for matrix");
      err = 1;
      return new value ((0));
    } // end if

    // Check the dimensions.
    if (i < 1 || i > mat.smat->get_nrow() || j < 1 || j > mat.smat->get_ncol()) {
      yyerror ("Error, index exceeds matrix dimensions.");
      err = 1;
      return new value ((0));
    } // end if

    // Return the matrix element
    mat.smat->set_element (i,j,elem->num);
    return new value ((mat.smat->get_element(i,j)));
  } // end if

  // Function should never get this far.
  assert (0);

  return new value((0));
} // end set_element

value * get_range(char * name, value * begin, value * stride, value * end,
		  int &err) {

  if(begin->type  != NUM ||
     stride->type != NUM || 
     end->type    != NUM) {

    yyerror("Error, range arguments must be numbers");
    err = 1;
    return new value();
  }

  int b = (int)begin->get_num();
  int e = (int)end->get_num();
  int s = (int)stride->get_num();

  // Check to make  sure that the matrix exists
  if (!syms->contains(name)) {
    yyerror ("Error, undefined symbol");
    err = 1;
    return new value();
  }
  value v= (syms->lookup(name).get_value());

  if(b < 1 || e < 1 || s < 1) {
    yyerror("Error, range arguments must be positive and non-zero");
    err = 1;
    return new value();
  }

  if(v.type == FUNCT) {

    yyerror("Error, cannot find range of function");
    err = 1; return new value();
  }
  if(v.type == NUM) {

    if(b > 1 || e > 1) {
      yyerror("Error, arguments may not exceed symbol\'s boundaries");
      err = 1; return new value();
    }
    return new value(v);
  }
  if(v.type == DENSE) {

    int ct;
    int len;

    //    cout << "Hello" << endl;

    if(e < b) {
      return new value();
    }

    int nrow = v.dmat->get_nrow();
    int ncol = v.dmat->get_ncol();

    if(e > (nrow * ncol)) {
      yyerror("Error, arguments may not exceed symbol\'s boundaries");
      err = 1; return new value();
    }

    len = (int)ceil((e-b+1)/float(s));
    //    cout << "len is " << len;

    // Make a vertical vector len elements tall
    value * ret = new value(new dense(len, 1, 0));

    // Fill it with the appropriate entries from the dense matrix
    for(ct = b; ct <= e; ct += s) {
      int col = ((ct-1) % ncol)+1;
      int row = ((ct-1) / ncol)+1;

      ret->dmat->set_element((int)ceil((ct-b+1)/float(s)), 
			     1, v.dmat->get_element(row, col));
    }

    return(ret);
  }
  if(v.type == SPARSE) {

    int ct;
    int len;

    //    cout << "Hello" << endl;

    if(e < b) {
      return new value();
    }

    int nrow = v.smat->get_nrow();
    int ncol = v.smat->get_ncol();

    if(e > (nrow * ncol)) {
      yyerror("Error, arguments may not exceed symbol\'s boundaries");
      err = 1; return new value();
    }

    len = (int)ceil((e-b+1)/float(s));
    //    cout << "len is " << len;

    // Make a vertical vector len elements tall
    value * ret = new value(new dense(len, 1, 0));

    // Fill it with the appropriate entries from the dense matrix
    for(ct = b; ct <= e; ct += s) {
      int col = ((ct-1) % ncol)+1;
      int row = ((ct-1) / ncol)+1;

      ret->dmat->set_element((int)ceil((ct-b+1)/float(s)), 
			     1, v.smat->get_element(row, col));
    }

    return(ret);
  }
  if(v.type == STRING) {

    int ct;
    int len;

    if(e < b) {
      return new value();
    }

    char *str = v.get_str();
    int slen = strlen(str);

    if(e > slen) {
      yyerror("Error, arguments may not exceed symbol\'s boundaries");
      err = 1; return new value();
    }

    // Adjust for string indeces
    b--; e--;

    len = (int)ceil((e-b+1)/float(s));
    //    cout << "len is " << len;

    // Make a new string of length len
    char *out = new char[len+1];
    out[len] = '\0';

    int ind = 0;
    // Fill it with the appropriate entries from the source string
    for(ct = b; ct <= e; ct += s, ind++) {
      out[ind] = str[ct];
    }

    return(new value(out));
  }

  assert(0);
  return new value();
}

/*============================================================================
#				GET_ELEMENT		
#				-----------    	
# in entry  : 	- a pointer to a char : name (name of the matrix)
#		- a list	      : args (the coordinates of the element)
#		- an integer	      : err
#
# in return :	the value of the element of the matrix name whose coordinates 
#		are stored in args
#
# details   :	- checks to make sure that the matrix name exists (looks if 
#		  symtab contains the name)
#		- it stores the value of the matrix in the variable mat
#		- if mat is a DENSE MATRIX
#		    + checks if there is at least one argument in args
#		    + checks if the args are numbers
#		    + stores the row coordinates in i (top value of the arg) 
#		      and remove it from the list arg
#		    + if there were only 1 argument, checks if mat is really a 
# 		      vector
#		    + checks if i doesn't exceed the matrix dimensions
#		    + stores the col coordinates in j (top value of arg) and
#		      removes it from the list arg	
#		    + checks if there is no more arguments in the list arg
#		    + checks if j doesn't exceed the matrix dimensions
#		    + calls the function get_element of dense.c and returns 
#		      the result	
#		- if mat is a SPARSE MATRIX
#		    + idem but calls the function of sparse.c
#
============================================================================*/

value * get_element (char *name, list *args, int &err) {
  // Return the element of a matrix.  
  int i,j;
  // Check to make  sure that the matrix exists
  if (!syms->contains(name)) {
    args->remove_all();
    yyerror ("Error, undefined matrix");
    err = 1;
    return new value ((0));
  }
  value mat (syms->lookup(name).get_value());
  dense ivec, jvec;

  if (syms->lookup(name).get_value().type == DENSE) {
   
    // There must be at least one argument.
    if (args->empty()) { 
      yyerror ("Error, invalid number of indecis for matrix");
      err = 1;
      return new value ((0));
    } // end if

    // if the index is a number, convert it into a vector
    if (args->top().get_value().type == NUM) {
      ivec = dense (1,1,args->top().get_value().num);
      args->remove_top();
    } // end if
    else   
      if (args->top().get_value().type == DENSE) {
        ivec = *args->top().get_value().dmat;
        args->remove_top();
        if (ivec.get_nrow() != 1 && ivec.get_ncol() != 1) {
          yyerror ("Error, indecis must be either numbers, or vectors");
          args->remove_all();
          err = 1;
          return new value ((0));
        }
      } // end if
      else {
          yyerror ("Error, indecis must be either numbers, or vectors");
          args->remove_all();
          err = 1;
          return new value ((0));
      } // end else

    // If there is only one index, then check to see if the matrix
    // is really a vector.
    if (args->empty()) { 
      if (mat.dmat->get_nrow() == 1) {
        jvec = ivec;
	ivec = dense (1,1,1);
      } // end if
      else
        if (mat.dmat->get_ncol() == 1) 
          jvec = dense (1,1,1);
	else {
          yyerror ("Error, invalid number of indecis for matrix");
          err = 1;
          return new value ((0));
        } // end else
    } // end if
    else {
      // if the index is a number, convert it into a vector
      if (args->top().get_value().type == NUM) {
        jvec = dense (1,1,args->top().get_value().num);
        args->remove_top();
      } // end if
      else   
        if (args->top().get_value().type == DENSE) {
          jvec = *args->top().get_value().dmat;
          args->remove_top();
          if (jvec.get_nrow() != 1 && jvec.get_ncol() != 1) {
            yyerror ("Error, indecis must be either numbers, or vectors");
            args->remove_all();
            err = 1;
            return new value ((0));
          }
        } // end if
        else {
          yyerror ("Error, indecis must be either numbers, or vectors");
          args->remove_all();
          err = 1;
          return new value ((0));
        } // end else
    } // end else

    // Maximum of two indecies are allowed.
    if (!args->empty()) { 
      yyerror ("Error, invalid number of indecis for matrix");
      err = 1;
      return new value ((0));
    } // end if

    // Make sure both vectors are row vectors
    if (ivec.get_nrow() != 1)
      ivec = transp (ivec);
    if (jvec.get_nrow() != 1)
      jvec = transp (jvec);

    int nrow = ivec.get_ncol();
    int ncol = jvec.get_ncol();
    // If both nrow, and ncol are 1, then return a number instead of a matrix
    // Check the boundary conditions too though.
    if (nrow == 1 &&  ncol == 1) {
      i = int (ivec.get_element (1,1));
      j = int (jvec.get_element (1,1));
      if (i < 1 || i > mat.dmat->get_nrow() || j < 1 || j > mat.dmat->get_ncol()) {
        yyerror ("Error, index exceeds matrix dimensions.");
        err = 1;
        return new value ((0));
      } // end if
      return new value (mat.dmat->get_element(i,j));
    } // end if

    int ii, jj;
    dense result (nrow, ncol);
    for (i = 1; i <= nrow; i += 1) {
      for (j = 1; j <= ncol; j += 1) {
        ii = int (ivec.get_element (1,i));
        jj = int (jvec.get_element (1,j));
        if (i < 1 || i > mat.dmat->get_nrow() || j < 1 || j > mat.dmat->get_ncol()) {
          yyerror ("Error, index exceeds matrix dimensions.");
          err = 1;
          return new value ((0));
        } // end if
        result.set_element(i,j,mat.dmat->get_element(ii,jj));
      } // end for
    } // end for
    return new value (new dense (result));
  } // end if


  // Repeat the same thing for sparse matrix.  
  if (syms->lookup(name).get_value().type == SPARSE) {
   
    // Must contain at leasd one index.
    if (args->empty()) { 
      yyerror ("Error, invalid number of indecis for matrix");
      err = 1;
      return new value ((0));
    } // end if

    // Indecis must be numbers.
    if (args->top().get_value().type != NUM) {
      args->remove_all();
      yyerror ("Error, invalid type for matrix index.");
      err = 1;
      return new value ((0));
    } // end if
    
    // Get the row index
    i = int(args->top().get_value().num);
    args->remove_top();

    // If there is only one index, possibly the matrix is a vector.
    if (args->empty()) { 
      if (mat.smat->get_nrow() == 1) {
        if (i < 1 || i > mat.smat->get_ncol()) {
          yyerror ("Error, index exceeds matrix dimensions.");
	  err = 1;
          return new value ((0));
	} // end if
        return new value ((mat.smat->get_element(1,i)));
      } // end if
      if (mat.smat->get_ncol() == 1) {
        if (i < 1 || i > mat.smat->get_nrow()) {
          yyerror ("Error, index exceeds matrix dimensions.");
	  err = 1;
          return new value ((0));
	} // end if
        return new value ((mat.smat->get_element(i,1)));
      } // end if
      yyerror ("Error, invalid number of indecis for matrix");
      err = 1;
      return new value ((0));
    } // end if

    // Indecies must be numbers
    if (args->top().get_value().type != NUM) {
      args->remove_all();
      yyerror ("Error, invalid type for matrix index.");
      err = 1;
      return new value ((0));
    } // end if
    
    // Get the collumn index.
    j = int(args->top().get_value().num);
    args->remove_top();

    // Maximum of two indexis allowed.
    if (!args->empty()) { 
      yyerror ("Error, invalid number of indecis for matrix");
      err = 1;
      return new value ((0));
    } // end if

    // Check the boundaries for the matrix
    if (i < 1 || i > mat.smat->get_nrow() || j < 1 || j > mat.smat->get_ncol()) {
      yyerror ("Error, index exceeds matrix dimensions.");
      err = 1;
      return new value ((0));
    } // end if

    // return the matrix element
    return new value ((mat.smat->get_element(i,j)));
  } // end if

  // This function should never get this far,
  assert (0);

  return new value((0));
} // end get_element

/*==========================================================================
				HELP
==========================================================================*/

value * help (char *topic) {

  struct stat statbuff;  // buffer to hold the status of the file
  char temp[128];
   
  int i = 0;

  // Determine the number  of preceding spaces
  while (topic[i] == ' ' || topic[i] == '\t')
    i += 1;
  string top(topic+i);

  strcpy (temp, spmathpath);
  strcat (temp, MANPAGE_DIR_NAME);

  if ((top == *(new string("is_dense")))  || 
      (top == *(new string("is_sparse"))) || 
      (top == *(new string("is_vector"))) || 
      (top == *(new string("is_string"))) || 
      (top == *(new string("is_number")))) {

    strcat (temp, "/is_**.manpage");
    spit_file(temp);

    return new value ();
  }
 
  if ((top == *(new string("indset0")))  || 
      (top == *(new string("indset1")))  || 
      (top == *(new string("colors")))   || 
      (top == *(new string("multic")))   || 
      (top == *(new string("colorset"))) || 
      (top == *(new string("colors2")))) {
    
    strcat (temp, "/set_order.manpage");
    spit_file(temp);

    return new value ();
  }

  if ((top == *(new string("nrow"))) || 
      (top == *(new string("ncol"))) || 
      (top == *(new string("getdia"))) || 
      (top == *(new string("rmdia"))) ||
      (top == *(new string("nnz"))) || 
      (top == *(new string("size")))) {

    strcat (temp, "/dim.manpage");
    spit_file(temp);

    return new value ();
  }

  if ((top == *(new string("extract"))) || 
      (top == *(new string("extract_nrow"))) || 
      (top == *(new string("extract_ncol")))) {

    strcat (temp, "/submat.manpage");
    spit_file(temp);

    return new value ();
  }

  if ((top == *(new string("getbwdl"))) || 
      (top == *(new string("getbwdu")))) {

    strcat (temp, "/getbwd.manpage");
    spit_file(temp);

    return new value ();
  }

  if ((top == *(new string("irand")))) {

    strcat (temp, "/rand.manpage");
    spit_file(temp);

    return new value ();
  }

  if ((top == *(new string("roscal"))) || 
      (top == *(new string("coscal"))) || 
      (top == *(new string("cnrms")))) {

    strcat (temp, "/normscal.manpage");
    spit_file(temp);

    return new value ();
  }

  if ((top == *(new string("getu"))) || 
      (top == *(new string("getl"))) ) {

    strcat (temp, "/triang_mat.manpage");
    spit_file(temp);

    return new value ();
  }

  if ((top == *(new string("dense")))  || 
      (top == *(new string("sparse"))) || 
      (top == *(new string("rperm")))  || 
      (top == *(new string("transp"))) || 
      (top == *(new string("cperm")))  || 
      (top == *(new string("dperm"))))  {

    strcat (temp, "/manip_matrix.manpage");
    spit_file(temp);

    return new value ();
  }


  if ((top == *(new string("hbin"))) || 
      (top == *(new string("hbout")))) {

    strcat (temp, "/inout.manpage");
    spit_file(temp);

    return new value ();
  }

  if ((top == *(new string("solve"))) ||
      (top == *(new string("solve_system")))) {

    strcat (temp, "/solve.manpage");
    spit_file(temp);

    return new value ();    
  }

  if ((top == *(new string("eye"))) || 
      (top == *(new string("zeroes"))) ||
      (top == *(new string("zeros"))) ||
      (top == *(new string("spzeroes"))) ||
      (top == *(new string("spzeros"))) ||
      (top == *(new string("ones")))) {

    strcat (temp, "/generate.manpage");
    spit_file(temp);

    return new value ();
  }

  strcat (temp, "/");
  strcat (temp, topic + i);
  strcat (temp, ".manpage");

  if (stat (temp, &statbuff) != 0) {

    strcpy (temp, spmathpath);
    strcat (temp, MANPAGE_DIR_NAME);
    strcat (temp, "/help");
    strcat (temp, ".manpage");

    if (stat (temp, &statbuff) != 0) {
      strcpy (temp, "Sorry, no online help for ");
      strcat (temp, topic + i);

      yyerror(temp);
    }
    else {
      spit_file(temp);
    }

    return new value ();
  } // end if

  strcpy (temp, spmathpath);
  strcat (temp, MANPAGE_DIR_NAME);
  strcat (temp, "/");
  strcat (temp, topic + i);
  strcat (temp, ".manpage");
  spit_file(temp);

  return new value ();

} // end help

value * spit_file (char *filename) {

  char buf[1025];

#ifndef NO_TK_COMPILE
  if(no_tk) {
#endif
    char * temp = new char [5 + strlen(filename)];
    sprintf(temp, "more %s", filename);

    system(temp);
#ifndef NO_TK_COMPILE
  }
  else {
    int lines = 0;
    
    ifstream tfile(filename);
    if(!tfile.is_open()) {
      sp_error("sorry, could not open file ");
      sp_error(filename);
      
      return new value();
    }
    
    while(tfile.getline(buf, 1023)) {
      lines++;
      sp_out(buf);
      sp_out("\n");
    }
    
    tfile.close();
    
    if(lines > 30) {
      
      // This allows the next string of output to be
      // entered into the tk text box without it scrolling
      // all the way to the bottom of the output.  This
      // is annoying...
      if(!no_tk) { TCL("turn_autoscroll_off"); }
      
      // This pages the textbox down one screen
      if(!no_tk) { TCL("page_down"); }

    }
  }
#endif

  return new value ();
}

/*============================================================================
#				READMT:		
#				--------    	
# in entry  :   a value    : file (name of the file where is stored the matrix)
#		an integer : err
#
# in return :	a value (the sparse matrix)
#
# details   :	It reads a sparse matrix from a Harwell-Boeing file
#		- checks to make sure that file is a string
#		- checks to make sure that the file exists
#		- calls the function readmt of sparse.c and return its result
#
============================================================================*/

value * readmt (value *file, int &err) {

  struct stat statbuff;  // buffer to hold the status of the file

  if (file->type != STRING) {
    yyerror ("Error, 1st argument of readmt must be a string.");
    err = 1;
    return file;
  } // end if

  //  cout << "String is " << file->str << endl;

  //  if (stat (file->str, &statbuff) != 0) {
  //    yyerror ("Error, file does not exist.");
  //    err = 1;
  //    return file;
  //  } // end if

  //  cout << "In readmt" << endl;

  // sparse *sptemp = new sparse();
  // sptemp->readmt(file->str);
  // return new value (sptemp);

  sparse *m = NULL;

  if(!load_matrix_from_file(file->str, m)) {
    yyerror ("Error, file does not contain a valid Harwell-Boeing or\n"
	     "Matrix Market matrix.");
    err = 1;
    return file;
  }

  return new value(m);
} // end readmt

/*============================================================================
#				SUBMAT :		
#				------    	
# in entry  :   i1 -> value (row of the first element to extract)
#		i2 -> value (row of the last element to extract)
#		j1 -> value (column of the first element to extract)
#		j2 -> value (column of the last element to extract)	
#		err -> integer
#
# in return :	a value (the submatrix)
#
# details   :	It extracts a submatrix from a matrix
#		- it checks to make sure that the first 4 elements are numbers
#		- it checks to make sure that the 5th element is a matrix
#		- it checks to make sure that i2>=i1 and that j2>=j1
#		- it checks to make sure that the indecis don't exceed
#		  the dimensions of the input matrix.
#		- according to the type of the matrix, it calls the function
#		  submat of sparse.C or of dense.C.
#
============================================================================*/

value * submat (value *i1, value *i2, value *j1, value *j2, 
                value *mat, int &err ) 
{
  
  if (i1->type != NUM) {
    yyerror ("Error, 1st argument of submat must be an integer .");
    err = 1;
    return i1;
  } // end if

  if (i2->type != NUM) {
    yyerror ("Error, 2nd argument of submat must be an integer .");
    err = 1;
    return i2;
  } // end if

  if (j1->type != NUM) {
    yyerror ("Error, 3st argument of submat must be an integer .");
    err = 1;
    return j1;
  } // end if

  if (j2->type != NUM) {
    yyerror ("Error, 4st argument of submat must be an integer .");
    err = 1;
    return j2;
  } // end if

  if (!( (mat->type == SPARSE) || (mat->type == DENSE))) {
    yyerror ("Error, 5st argument of submat must be a matrix .");
    err = 1;
    return mat;
  } // end if

  if ((i2->num < i1->num) || (j2->num < j1->num)) {
    yyerror ("Error, the conditions aren't respected: i2>=i1 and j2>=j1 ");
    err = 1;
    return new value(0);
  } // end if

  if (mat->type == SPARSE)
    {

      if ( (i1->num<=0) || (i2->num<=0) || (j1->num<=0) || (j2->num<=0) || (i1->num>(mat->smat)->get_nrow()) || (i2->num>(mat->smat)->get_nrow()) || (j1->num>(mat->smat)->get_ncol()) || (j2->num>(mat->smat)->get_ncol()) ) {
	yyerror ("Error, at least one of the indecis exceeds the matrix dimensions ");
	err = 1;
	return new value(0);
      } // end if


      sparse *sptemp = new sparse(*(mat->smat)); 
      sptemp->submat(sptemp->get_nrow(), (int)i1->num, (int)i2->num, (int)j1->num, (int)j2->num);
      return new value (sptemp);
    } //end if

  else {
    if ( (i1->num<=0) || (i2->num<=0) || 
	 (j1->num<=0) || (j2->num<=0) || 
	 (i1->num>(mat->dmat)->get_nrow()) || 
	 (i2->num>(mat->dmat)->get_nrow()) || 
	 (j1->num>(mat->dmat)->get_ncol()) || 
	 (j2->num>(mat->dmat)->get_ncol()) ) {

      yyerror ("Error, at least one of the indecis exceeds the matrix "
	       "dimensions ");
      err = 1;
      return new value(0);
    }

    dense *sptemp = new dense(*(mat->dmat)); 
    sptemp->submat(sptemp->get_nrow(), 
		   (int)i1->num, (int)i2->num, (int)j1->num, (int)j2->num);

    return new value (sptemp);
  }
  
} // end submat


/*============================================================================
#				FILTER :		
#				------    	
# in entry  :   n 	-> value (row dimension of the matrix mat)
#		job 	-> value (job indicator)
#		drptol 	-> value (drop tolerance)
#		mat 	-> value (the matrix) 
#
# in return :	a value (the resulting matrix)
#
# details   :	This module removes any elements whose absolute value is small
#		from an input matrix A and puts the resulting matrix in B.  
#		The input parameter job selects a definition of small.
#		- It checks if n, job and drptol are numbers
#		- It checks to amke sure that job really defines a job!
#		- It checks to make sure that mat is a sparse matrix
#		- It calls the function filter from sparse.C
#
============================================================================*/


value * filter (value *n, value *job, value *drptol, value *mat, int &err)

{
  if (n->type != NUM) {
    yyerror ("Error, 1st argument of filter must be a number.");
    err = 1;
    return n;
  } // end if

  if (job->type != NUM) {
    yyerror ("Error, 2nd argument of filter must be a number.");
    err = 1;
    return job;
  } // end if

  if (!( (job->num == 1) || (job->num == 2) || (job->num == 3) )) {
    yyerror ("Error, 2nd argument must be equal to 1, 2 or 3    	");
    err = 1;
    return job;
  } // end if

  if (drptol->type != NUM) {
    yyerror ("Error, 3rd argument of filter must be a number.");
    err = 1;
    return drptol;
  } // end if

  if (mat->type != SPARSE) {
    yyerror ("Error, 4st argument of filter must be a sparse matrix.");
    err = 1;
    return mat;
  } // end if

  sparse *sptemp = new sparse(*(mat->smat));
  sptemp->filter((int)n->num, (int)job->num, drptol->num);
  return new value(sptemp);
}

/*============================================================================
#				CSORT :		
#				------    	
# in entry  :   values -> job indicator
#		mat    -> the matrix
#		err 
#
# in return :	a value (the resulting matrix)
#
# details   :	sorts the elements in increasing order of columns
#		- Checks to make sure that values is a number
#		- Checks to make sure that mat is a sparse matrix
#		- Calls thew function csort from sparse.C
#
============================================================================*/

value * csort (value *values, value *mat, int &err)

{
  if (values->type != NUM) {
    yyerror ("Error, 1st argument of filter must be a number.");
    err = 1;
    return values;
  } // end if

  if (mat->type != SPARSE) {
    yyerror ("Error, 2nd argument of filter must be a sparse matrix.");
    err = 1;
    return mat;
  } // end if

  sparse *sptemp = new sparse(*(mat->smat));
  sptemp->csort((int)values->num);
  return new value(sptemp);

}

/*============================================================================
#				RMDIA		
#				-----    	
# in entry  : 	ioff -> int (number of the diagonal) 
#		mat  -> the matrix
#		err
#
# in return : 	the new matrix
#
# details   : 	It removes a diagonal from a dense matrix
#		- It checks to make sure that the arguments have the right type
#		- It checks to make sure that the number of the diagonal 
#		  (ioff) doesn't exceed the dimension of the matrix
#		- According to the type of mat, it calls rmdiag from dense.C 
#		  or rmdia from sparse.C
#
============================================================================*/

value * rmdia (value *ioff, value *mat, int &err)

{
  if (ioff->type != NUM) {
    yyerror ("Error, 1st argument of filter must be a number.");
    err = 1;
    return ioff;
  } // end if

  if (!( (mat->type == SPARSE) || (mat->type == DENSE)  )) {
    yyerror ("Error, 2nd argument of filter must be a matrix.");
    err = 1;
    return mat;
  } // end if

  if (mat->type == SPARSE) 
    {

      if ( abs((int)ioff->num)>= (mat->smat)->get_nrow()){
	yyerror ("Error, the number of the diagonal exceeds the matrix dimension");
	err = 1;
	return ioff;
      } // end if

      sparse *sptemp = new sparse(*(mat->smat));
      sptemp->rmdia((int)ioff->num);
      return new value(sptemp);
    }
  else 
    {

      if ( abs((int)ioff->num)>= (mat->dmat)->get_nrow()){
	yyerror ("Error, the number of the diagonal exceeds the matrix dimension");
	err = 1;
	return ioff;

  } // end if	
	dense *dtemp = new dense (*(mat->dmat));
	dtemp->rmdiag((int)ioff->num);
	return new value(dtemp);
	}

}

/*============================================================================
#				GETDIA		
#				------    	
# in entry  : 	ioff -> int (number of the diagonal) 
#		mat  -> the matrix
#		err
#
# in return : 	the diagonal (dense matrix)
#
# details   : 	It extracts a diagonal from a dense matrix
#		- It checks to make sure that the arguments have the right type
#		- It checks to make sure that the number of the diagonal 
#		  (ioff) doesn't exceed the dimension of the matrix
#		- If mat is SPARSE
#		    + it calls getdia from sparse.C ans stores the result
#		      which is a double array (the elements of the diag)
#		    + it creates a dense matrix from this array and 
#		      returns the result
#		- If mat is DENSE
#		    + it calls getdiag from dense.C and stores the result
#		      which is already a dense matrix
#		    + it returns the result
#
============================================================================*/

value * getdia (value *ioff, value *mat, int &err)

{
if (ioff->type != NUM) {
    yyerror ("Error, 1st argument of filter must be a number.");
    err = 1;
    return ioff;
  } // end if

if (!(  (mat->type == SPARSE) || (mat->type == DENSE) )) {
    yyerror ("Error, 2nd argument of filter must be a matrix.");
    err = 1;
    return mat;
  } // end if

if (mat->type == SPARSE)
	{
if ( abs((int)ioff->num)>= (mat->smat)->get_nrow()){
    yyerror ("Error, the number of the diagonal exceeds the matrix dimension");
    err = 1;
    return ioff;
  } // end if

	double * temp = new double[(mat->smat)->get_nrow()];
	sparse *sptemp = new sparse(*(mat->smat));
	temp = sptemp->getdia ((int)ioff->num);

	int nro= (mat->smat)->get_nrow() - abs ((int)ioff->num) ;
	dense *res= new dense(1 , nro);
	int i;
	for (i=1;i<=nro; i++)
		{ res->set_element(1, i, temp[i-1]); }

	delete [] temp;
	sptemp->free();

	return new value(res);
	}

else 
	{
if ( abs((int)ioff->num)>= (mat->dmat)->get_nrow()){
    yyerror ("Error, the number of the diagonal exceeds the matrix dimension");
    err = 1;
    return ioff;}

	dense *dtemp = new dense(*(mat->dmat));
	dtemp->getdiag ((int)ioff->num);

	return new value(dtemp);
	}
}

/*============================================================================
#				GETL/GETU		
#				---------    	
# in entry  : 	mat -> the matrix
#		err -> an error message 
#
# in return : 	a value (a dense or a sparse matrix)
#
# details   : 	It extracts either the lower or the upper triangular part 
#		of the dense matrix.
#		- It checks to make sure that mat is a square matrix
#		- According to the type of the matrix, it calls getu/getl
#		  from sparse.C or from dense.C	
#
============================================================================*/

value * getu(value *mat, int &err)
{
 if (!( (mat->type == SPARSE) || (mat->type == DENSE))) {
    yyerror ("Error, 1st argument of getu must be a matrix.");
    err = 1;
    return mat;
  } // end if

if (mat->type == SPARSE)
	{
 	if ((mat->smat)->get_nrow() != (mat->smat)->get_ncol()) {
    	yyerror ("Error, the matrix must be square.");
    	err = 1;
    	return mat;
  	} // end if

	sparse *sptemp = new sparse(*(mat->smat));
	sptemp->getu();
	return new value(sptemp);
	}
else {
	if ((mat->dmat)->get_nrow() != (mat->dmat)->get_ncol()) {
    	yyerror ("Error, the matrix must be square.");
    	err = 1;
    	return mat;
  	} // end if

	dense *dtemp = new dense(*(mat->dmat));
	dtemp->getu();
	return new value(dtemp);

	}

}

value * getl(value *mat, int &err)
{
  if (!( (mat->type == SPARSE) || (mat->type == DENSE))) {
    yyerror ("Error, 1st argument of getl must be a matrix.");
    err = 1;
    return mat;
  } // end if

if (mat->type == SPARSE)
	{
 	if ((mat->smat)->get_nrow() != (mat->smat)->get_ncol()) {
    	yyerror ("Error, the matrix must be square.");
    	err = 1;
    	return mat;
  	} // end if

	sparse *sptemp = new sparse(*(mat->smat));
	sptemp->getl();
	return new value(sptemp);
	}
else {
	if ((mat->dmat)->get_nrow() != (mat->dmat)->get_ncol()) {
    	yyerror ("Error, the matrix must be square.");
    	err = 1;
    	return mat;
  	} // end if

	dense *dtemp = new dense(*(mat->dmat));
	dtemp->getl();
	return new value(dtemp);

	}

}

/*============================================================================
#				GETBWD*		
#				-------    	
# in entry  : 	mat -> the matrix
#		err -> an error message 
#
# in return : 	a value (an integer or an array of integer)
#
# details   : 	It calculates the bandwith of the upper and the lower part
#		of the matrix
#		- It checks to make sure that mat is a matrix
#		- getbwdu returns the bandwith of the upper part
#		- getbwdl returns the bandwith of the lower part
#		- getbwd returns an array containing 2 elements: the first
#		  one is the bandwith of the lower part and the 2nd is the
#		  bandwith of the upper part 
#
============================================================================*/

value * getbwd(value *mat, int &err)
{
if (!(  (mat->type == SPARSE) || (mat->type == DENSE) )) {
    	yyerror ("Error, the argument of getbwdu must be a matrix.");
    	err = 1;
    	return mat;
  	} // end if

	int *res=new int[2];

if (mat->type == SPARSE)
	{
	res = (mat->smat)->getbwd();
	dense *dtemp = new dense(1,2);
	dtemp->set_element(1,1,res[0]);
	dtemp->set_element(1,2,res[1]);
	delete [] res;
	return new value(dtemp);}

else 
	{res = (mat->dmat)->getbwd();

	dense *dtemp = new dense(1,2);
	dtemp->set_element(1,1,res[0]);
	dtemp->set_element(1,2,res[1]);
	delete [] res;
	return new value(dtemp);}
	
}

value * getbwdu(value *mat, int &err)
{
  if (!(  (mat->type == SPARSE) || (mat->type == DENSE) )) {
    yyerror ("Error, the argument of getbwdu must be a matrix.");
    err = 1;
    return mat;
  } // end if
  if (mat->type == SPARSE)
    return new value((mat->smat)->getbwdu());
  else 
    return new value((mat->dmat)->getbwdu());
}

value * getbwdl(value *mat, int &err)
{
  if (!(  (mat->type == SPARSE) || (mat->type == DENSE) )) {
    yyerror ("Error, the argument of getbwdl must be a matrix.");
    err = 1;
    return mat;
  } // end if
  if (mat->type == SPARSE)
    return new value((mat->smat)->getbwdl());
  else 
    return new value((mat->dmat)->getbwdl());
}

value * blkfnd(value *mat, int &err)
{
  if (mat->type != SPARSE) {
    yyerror ("Error, the argument of blkfnd must be a sparse matrix.");
    err = 1;
    return mat;
  } // end if
  return new value((mat->smat)->blkfnd());
}

/*============================================================================
#				CNRMS :		
#				------    	
# in entry  :   nrm -> norm indicator
#		mat    -> the matrix
#		err 
#
# in return :	a value (an array)
#
# details   :	It computes the norms of the columns of a matrix
#		- It checks to make sure that the matrix is sparse or dense and #		  that the norm equals to 0, 1, 2 or 3.
#		- It creates the dense matrix which will be the result
#		- It calls cnrms from sparse.C
#		- It returns the result
#
============================================================================*/

value * cnrms(value *mat, value *nrm, int &err)
{
  if (!(  (mat->type == SPARSE) || (mat->type == DENSE)  )) {
    yyerror ("Error, the 1st argument of cnrms must be a matrix.");
    err = 1;
    return mat;
  } // end if

  if (nrm->type != NUM) {
    yyerror ("Error, the 2nd argument of cnrms must be an integer.");
    err = 1;
    return nrm;
  } // end if

  if (!( (nrm->num == 0) || (nrm->num ==1)  || (nrm->num == 2) || (nrm->num ==3)))
    {
      yyerror ("Error, the 2nd argument of cnrms must be an integer between 0 and 3.");
      err = 1;
      return nrm;
    } // end if
		
  if (mat->type == SPARSE)
    {
      int nco = (mat->smat)->get_ncol();
      double * temp = new double[nco];
      sparse *sptemp = new sparse(*(mat->smat));
      dense *res = new dense(1, nco);
	
      temp = sptemp->cnrms((int)nrm->num);
	
      int i;
      for (i=0;i<nco;i++)
	{ res->set_element(1, i+1, temp[i]); }

      delete [] temp;
      sptemp->free();
	
      return new value(res);
    }

  else {
    int nco = (mat->dmat)->get_ncol();
    double * temp = new double[nco];
    dense *dtemp = new dense(*(mat->dmat));
    dense *res = new dense(1, nco);
    
    temp = dtemp->cnrms((int)nrm->num);
    
    int i;
    for (i=0;i<nco;i++)
      { res->set_element(1, i+1, temp[i]); }
    
    delete [] temp;
    dtemp->free();
    
    return new value(res);
  }

}

/*============================================================================
#			    COSCAL/ROSCAL :		
#			    -------------    	
# in entry  :   nrm    -> norm indicator
#		mat    -> the matrix
#		err 
#
# in return :	a value (a sparse matrix)
#
# details   :	It scales the rows (or the columns) of a matrix by their norms
#		- It checks to make sure that mat is a matrix and that nrm
#		  equals to 0, 1, 2 or 3
#		- It calls the function roscal (or coscal) from sparse.C
#
============================================================================*/

value * coscal(value *mat, value *nrm, int &err)

{
if (!(  (mat->type == SPARSE)  || (mat->type == DENSE)  )) {
    	yyerror ("Error, the 1st argument of coscal must be a dense matrix.");
    	err = 1;
    	return mat;
  	} // end if

if (nrm->type != NUM) {
    	yyerror ("Error, the 2nd argument of coscal must be a number.");
    	err = 1;
    	return mat;
  	} // end if

if (!( (nrm->num == 0) || (nrm->num ==1)  || (nrm->num == 2) || (nrm->num ==3)))
{
    	yyerror ("Error, the 2nd argument of coscal must be an integer between 0 and 3.");
    	err = 1;
    	return nrm;
  	} // end if


if (mat->type == SPARSE)
{	sparse *sptemp = new sparse(*(mat->smat));
	sptemp->coscal((int)nrm->num);
	return new value(sptemp);
}
else 
{	dense *dtemp = new dense(*(mat->dmat));
	dtemp->coscal((int)nrm->num);
	return new value(dtemp);
}
}

value * roscal(value *mat, value *nrm, int &err)

{
if (!(  (mat->type == SPARSE)  || (mat->type == DENSE)  )) {
    	yyerror ("Error, the 1st argument of coscal must be a dense matrix.");
    	err = 1;
    	return mat;
  	} // end if


if (nrm->type != NUM) {
    	yyerror ("Error, the 2nd argument of roscal must be a number.");
    	err = 1;
    	return mat;
  	} // end if

if (!( (nrm->num == 0) || (nrm->num ==1)  || (nrm->num == 2) || (nrm->num ==3)))
{
    	yyerror ("Error, the 2nd argument of roscal must be an integer between 0 and 3.");
    	err = 1;
    	return nrm;
  	} // end if

if (mat->type == SPARSE)
{	sparse *sptemp = new sparse(*(mat->smat));
	sptemp->roscal((int)nrm->num);
	return new value(sptemp);
}
else 
{	dense *dtemp = new dense(*(mat->dmat));
	dtemp->roscal((int)nrm->num);
	return new value(dtemp);
}
}

/*==========================  ADDBLK  =================================*/

value * addblk(value *mata, value *matb, value *ipos, value *jpos, int &err)

{
if (mata->type != SPARSE) {
    yyerror ("Error, 1st argument of addblk must be a sparse matrix.");
    err = 1;
    return mata;
  } // end if

if (matb->type != SPARSE) {
    yyerror ("Error, 2nd argument of addblk must be a sparse matrix.");
    err = 1;
    return matb;
  } // end if

if (ipos->type != NUM) {
    yyerror ("Error, 3rd argument of addblk must be an integer.");
    err = 1;
    return ipos;
  } // end if

if (jpos->type != NUM) {
    yyerror ("Error, 4st argument of addblk must be an integer.");
    err = 1;
    return jpos;
  } // end if

if ( (ipos->num<=0) || (ipos->num>(mata->smat)->get_nrow()) || (jpos->num<=0) || (jpos->num>(mata->smat)->get_ncol()) ) {
    yyerror ("Error, at least one of the indecis exceed the dimension od the first matrix. ");
    err = 1;
    return ipos;
  } // end if

if ( (ipos->num + (matb->smat)->get_nrow() > (mata->smat)->get_nrow()) || (jpos->num + (matb->smat)->get_ncol() > (mata->smat)->get_ncol())) {
    yyerror ("Error, these arguments aren't correct!.");
    err = 1;
    return jpos;
  } // end if

	sparse *sptemp = new sparse(*(mata->smat));
	sptemp->addblk(*(matb->smat), (int)ipos->num, (int)jpos->num);
	return (new value (sptemp));

}

/*================== GET1UP =============*/

value * get1up(value *mat, int &err)

{
if (mat->type != SPARSE) {
    yyerror ("Error, 1st argument of get1up must be a sparse matrix.");
    err = 1;
    return mat;
  } // end if

	int nro = (mat->smat)->get_nrow();
	int *temp = new int[nro];
	sparse *sptemp = new sparse(*(mat->smat));
	dense *dtemp = new dense(nro,1);
	temp = sptemp->get1up();
	int i;
	for (i=1;i<=nro;i++) { dtemp->set_element(1,i,temp[i-1]); }
	delete [] temp;
	sptemp->free();
	return new value(dtemp);
	
}



/*============================================================================
#				GEN57PT :		
#				--------    	
# in entry  :   a value    : nx (number of grid points in x direction)
#		a value    : ny (number of grid points in y direction)
#		a value    : nz (number of grid points in z direction)
#		a value    : alpha (coefficient alpha of the boundaries
#			     conditions)
#		an integer : err
#
# in return :	a value (the sparse matrix)
#
# details   :	It generates a 5-point or 7-point matrix (see genmat.f)
#		- checks to make sure that nx,ny and nz are numbers
#		- checks to make sure that al is a vector (a dense matrix 
#		  whose nrow=1 )
#		- checks to make sure that al contains six elements
#		- calls the function gen57pt of sparse.c and return its result
#
============================================================================*/

value * gen57pt (value * nx, value *ny, value *nz, value *al, int &err) {

  

  if (nx->type != NUM) {
    yyerror ("Error, 1st argument of gen57pt must be a number.");
    err = 1;
    return nx;
  } // end if

   if (ny->type != NUM) {
    yyerror ("Error, 2nd argument of gen57pt must be a number.");
    err = 1;
    return ny;
  } // end if

   if (nz->type != NUM) {
    yyerror ("Error, 3rd argument of gen57pt must be a number.");
    err = 1;
    return nz;
  } // end if

   if (al->type != DENSE) {
    yyerror ("Error, 4st argument of gen57pt must be a vector.");
    err = 1;
    return al;
  } // end if

   if ((al->dmat)->get_nrow() != 1) {
    yyerror ("Error, 4st argument of gen57pt must be a vector.");
    err = 1;
    return al;
  } // end if

   if ((al->dmat)->get_ncol() != 6) {
    yyerror ("Error, 4st argument of gen57pt must contain 6 elements");
    err = 1;
    return al;
  } // end if




 

  sparse *sptemp = new sparse();
  sptemp->gen57pt((int)nx->num, (int)ny->num, (int)nz->num, (al->dmat)->get_element(1,1), (al->dmat)->get_element(1,2), (al->dmat)->get_element(1,3), (al->dmat)->get_element(1,4), (al->dmat)->get_element(1,5), (al->dmat)->get_element(1,6));
  return new value (sptemp);
} // end gen57pt



/*============================================================================
#				PRTMT :		
#				------    	
# in entry  :   a value    : mat (sparse matrix to store)
#		a value    : file (the name of the file where to store mat)
#		an integer : err
#
# in return :	a value (0 if there is an error, otherwise 1)
#
# details   :	It stores the matrix mat into the Harwell-Boeing file 
#		- checks to make sure that mat is a SPARSE matrix
#		- checks to make sure that file is a string
#		- checks to make sure that the file doesn't exist
#		- calls the function prtmt of sparse.c
#
============================================================================*/

value * prtmt (value *mat, value *file, int &err) {

  struct stat statbuff;  // buffer to hold the status of the file

  if (mat->type != SPARSE) {
    yyerror ("Error, 1st argument of prtmt must be a sparse matrix.");
    err = 1;
    return file;
  } // end if

  if (file->type != STRING) {
    yyerror ("Error, 2nd argument of prtmt must be a string.");
    err = 1;
    return file;
  } // end if

  if (stat (file->str, &statbuff) == 0) {
    yyerror ("Error, file already exists.");
    err = 1;
    return file;
  } // end if

  //  mat->smat->prtmt(file->str);

  save_matrix_to_file(file->str, *(mat->smat), "RUA\0");

  return new value((1));

} // end prtmt

/*============================================================================
#				VEC :		
#				----    	
# in entry  :   a value    : left (name of the file where is stored the matrix)
#		a value	   : middle 
#		a value	   : right 
#		an integer : err
#
# in return :	a value (the vector)
#
# details   :	It creates a vector
#		Example : vec(a,b,c,0) : a is the first element of the vector
#					 c is the last one
#					 the i'th element = a+(i-1)*b 
#		- checks to make sure that the arguments are numbers
#		- counts the length of the vector ([(c-a)/b])
#		- creates a vector of this length
#		- set all the elements result[i]=left + (i-1)*middle 
#
============================================================================*/

value * vec (value *left, value *middle, value *right, int &err) {
  // Create a vector


  if (left->type == NUM && middle->type == NUM && right->type == NUM) {
    int start = int(left->num);
    int inc = int (middle->num);
    int end = int (right->num);
    int i = start;
    int count = 0;
    while (i <= end) {
      i += inc;
      count += 1;
    } // end while
 
    if (count <= 0)
      return new value();

    dense result (1,count);
    i = start;
    count = 1;
    while (i <= end) {
      result.set_element(1,count,i);
      i += inc;
      count += 1;
    } // end while

    return new value (&result);
  } // end if

  err = 1;
  yyerror ("Error, incompatible types for operator :");
  return left;

} // end vec

/*============================================================================
#				SOLVE_SYSTEM :		
#				-------------    	
# in entry  :   10 values: 
#		mat -> the sparse matrix
#               rhs-> right hand side
#               guess
#		solver_name
#               prec_name -> the name of the preconditionner
#		arg_prec -> arguments of this preconditionner
#		prec -> status of the preconditionning
#               krylov -> size of the krylov subspace
#               maxmult
#               tols -> relative tolerance, absolute tolerance
#
# in return :	a value (the solution of the linear system)
#
# details   :	It solves a linear system Ax = y where A is sparse 
#
============================================================================*/

value * solve_system(value *mat, value *rhs, value *guess, 
		     value *solver_name, value *prec_name, 
		     value *arg_prec, value *prec, value *krylov, 
		     value *maxmult, value *tols, int &err)
{
  if (mat->type != SPARSE) {
    yyerror ("Error, 1st argument of solve_system must be a sparse matrix.");
    err = 1;
    return mat;
  } // end if

  if (rhs->type != DENSE) {
    yyerror ("Error, 2nd argument of solve_system must be a dense matrix.");
    err = 1;
    return mat;
  } // end if

  if (guess->type != DENSE) {
    yyerror ("Error, 3rd argument of solve_system must be a dense matrix.");
    err = 1;
    return mat;
  } // end if

  if (solver_name->type != STRING) {
    yyerror ("Error, 4th argument of solve_system must be a string.");
    err = 1;
    return mat;
  } // end if

  if (prec_name->type != STRING) {
    yyerror ("Error, 5th argument of solve_system must be a string.");
    err = 1;
    return mat;
  } // end if

  if (arg_prec->type != DENSE) {
    yyerror ("Error, 6th argument of solve_system must be a dense array.");
    err = 1;
    return mat;
  } // end if

  string pr(prec->str);
  if ( (pr != *(new string("none"))  ) && 
       (pr != *(new string("left"))  ) && 
       (pr != *(new string("right")) ) &&
       (pr != *(new string("both"))  )) {

    yyerror ("Error, 7st argument of solve_system must be \"none\", "
	     "\"left\", \"right\" or \"both\"");
    err = 1;
    return mat;
  } // end if

  if (krylov->type != NUM) {
    yyerror ("Error, 8th argument of solve_system must be an integer.");
    err = 1;
    return mat;
  } // end if

  if (maxmult->type != NUM) {
    yyerror ("Error, 9th argument of solve_system must be an integer.");
    err = 1;
    return mat;
  } // end if

  if (tols->type != DENSE) {
    yyerror ("Error, 10th argument of solve_system must be a two-element vector.");
    err = 1;
    return mat;
  } else if(tols->dmat->get_ncol() != 2) {
    yyerror ("Error, 10th argument must be a two-element vector");
    err = 1;
    return mat;
  }

  if (  ((rhs->dmat)->get_nrow() != (mat->smat)->get_nrow()) ||  
	((guess->dmat)->get_nrow() != (mat->smat)->get_nrow()) ) {
    yyerror ("Error, either the right hand side or the guess has wrong dimension.");
    err = 1;
    return mat;
  } // end if

  string s(solver_name->str);

  if ( (s !=*(new string("cg")))      && 
       (s !=*(new string("bcg")))     && 
       (s !=*(new string("dbcg")))    && 
       (s !=*(new string("cgnr")))    && 
       (s !=*(new string("bcgstab"))) && 
       (s !=*(new string("fom")))     && 
       (s !=*(new string("tfqmr")))   && 
       (s !=*(new string("gmres")))   && 
       (s !=*(new string("fgmres")))  && 
       (s !=*(new string("dqgmres"))) ) {

    sp_error("Error, solver \"");
    sp_error(s);
    sp_error("\" doesn\'t exist\n");

    err = 1;
    return mat;
  }

  string p(prec_name->str); 
  double *arg_prec2 = NULL;

  // WHEN ADDING A PRECONDITIONNER
  //
  //  Add a new test to the following if statement
  // to test if p is equal to the name of the new
  // preconditionner.

  if ( (p !=*(new string("ilut"))) && 
       (p !=*(new string("iluk"))) &&
       (p !=*(new string("ilud"))) && 
       (p !=*(new string("ilu0"))) &&
       (p !=*(new string("ilutp")))) {

    sp_error("Error, preconditionner \"");
    sp_error(p);
    sp_error("\" doesn\'t exist\n");

    err = 1;
    return mat;
  }

  // WHEN ADDING A PRECONDITIONNER
  //
  //  Add a new else-if clause here to ensure that
  // the right number of arguments have been passed
  // to the preconditioner...

  if (p ==*(new string("ilut"))) {

    if ( (arg_prec->dmat)->get_ncol() != 2  ) {

      yyerror ("ILUT: Wrong number of arguments for the preconditionner");
      err = 1;
      return mat;
    }
    else {

      arg_prec2 = new double[2]; 
      arg_prec2[0]=(arg_prec->dmat)->get_element(1,1);
      arg_prec2[1]=(arg_prec->dmat)->get_element(1,2);
    }
  }
  else if (p ==*(new string("iluk"))) {
    if ( (arg_prec->dmat)->get_ncol() != 1  ) {
      yyerror ("ILUK: Wrong number of arguments for the preconditionner");
      err = 1;
      return mat;
    }
    else {
      arg_prec2 = new double[1]; 
      arg_prec2[0]=(arg_prec->dmat)->get_element(1,1);
    }
  }
  else if (p ==*(new string("ilu0"))) {
    if ( (arg_prec->dmat)->get_ncol() != 1  ) {
      yyerror ("ILU0: wrong number of arguments for the preconditionner");
      err = 1;
      return mat;
    }
  }
  else if (p ==*(new string("ilud"))) {
    if ( (arg_prec->dmat)->get_ncol() != 2  ) {
      yyerror ("ILUD: Wrong number of arguments for the preconditionner");
      err = 1;
      return mat;
    }
    else {
      arg_prec2 = new double[2]; 
      arg_prec2[0]=(arg_prec->dmat)->get_element(1,1);
      arg_prec2[1]=(arg_prec->dmat)->get_element(1,2);
    }
  }
  else if (p == *(new string("ilutp"))) {
    if ((arg_prec->dmat)->get_ncol() != 3) {
      sp_error("Error, ilutp preconditionner takes 3 arguments\n");
      err = 1;
      return mat;
    }
    else {
      arg_prec2 = new double[3]; 
      arg_prec2[0]=(arg_prec->dmat)->get_element(1,1);
      arg_prec2[1]=(arg_prec->dmat)->get_element(1,2);
      arg_prec2[2]=(arg_prec->dmat)->get_element(1,3);
    }
  }

  int i;
  int n = (mat->smat)->get_nrow();

  //  sp_out("nrow: "); sp_out(n); sp_out("\n");

  double *sol = new double[n];
  sparse *sptemp(mat->smat);
  double *rhs2=new double[n];

  for (i=0;i<n;i++) {
    rhs2[i] = (rhs->dmat)->get_element(1,i+1);
  }

  double *guess2=new double[n];

  for (i=0;i<n;i++) {
    guess2[i] = (guess->dmat)->get_element(1,i+1);
  }

  int nprec = 0;

  if      (pr == *(new string("none")) )  { nprec = 0; } 
  else if (pr == *(new string("left")) )  { nprec = 1; }
  else if (pr == *(new string("right")) ) { nprec = 2; }
  else if (pr == *(new string("both")) )  { nprec = 3; }

  double tolrel = (tols->dmat)->get_element(1,1);
  double tolabs = (tols->dmat)->get_element(1,2);

  /*
  cout << "LARGE DEBUG MESSAGE" << endl
       << "Solver name: " << solver_name->str << endl
       << "Prec name: " << prec_name->str << endl
       << "Prec type: " << nprec << endl
       << "Krylob subspace: " << (int)krylov->num << endl
       << "Maxmult: " << (int)maxmult->num << endl
       << "Tolrel: " << tolrel << endl
       << "Tolabs: " << tolabs << endl;
  */

  sol = sptemp->solve_system(rhs2, guess2, solver_name->str, 
			     prec_name->str, arg_prec2, nprec, 
			     (int)krylov->num, (int)maxmult->num, 
			     tolrel, tolabs);

  delete [] rhs2;
  delete [] guess2;

  sptemp->free();

  // If an error occured, output nothing
  if(sol == NULL) {
    return new value();
  }

  dense *solution = new dense(n,1);

  for (i=0;i<n;i++) {
    solution->set_element(i+1,1, sol[i]);
  }

  delete [] sol;

  return new value(solution);
}

/*============================================================================
#				SOLVE_SYSTEM :		
#				-------------    	
# in entry  :   3 values: 
#		mat -> the sparse matrix
#               rhs-> right hand side
#               guess
#
#               All other parameters (solver name, preconditioner, etc)
#               are stored in variables in the symbol table and are
#               retrived in this function.
#
# in return :	a value (the solution of the linear system)
#
# details   :	It solves a linear system Ax = y where A is sparse 
#
============================================================================*/

value * solve_system(value *mat, value *rhs, value *guess, int &err)
{

  value tsolver_name = syms->lookup("slvr_solver_name").get_value();
  value *solver_name = &tsolver_name;

  if(solver_name->type == NO_TYPE) {
    yyerror("Error, variable slvr_solver_name must be set to the solver "
	    "name for the 3-argument version of solve_system.");
    err = 1;
    return mat;
  }

  value tprec_name = syms->lookup("slvr_prec_name").get_value();
  value *prec_name = &tprec_name;

  if(prec_name->type == NO_TYPE) {
    yyerror("Error, variable slvr_prec_name must be set to the preconditioner "
	    "name for the 3-argument version of solve_system.");
    err = 1;
    return mat;
  }

  value targ_prec = syms->lookup("slvr_prec_args").get_value();
  value *arg_prec = &targ_prec;

  if(arg_prec->type == NO_TYPE) {
    yyerror("Error, variable slvr_prec_args must be set to the preconditioner "
	    "arguments for the 3-argument version of solve_system.");
    err = 1;
    return mat;
  }

  value tprec = syms->lookup("slvr_prec_type").get_value();
  value *prec = &tprec;

  if(prec->type == NO_TYPE) {
    yyerror("Error, variable slvr_prec_type must be set to the\n "
	    "preconditioner type (\"none\", \"left\", \"right\", or\n"
	    "\"both\")  for the 3-argument version of solve_system.");
    err = 1;
    return mat;
  }

  value tkrylov = syms->lookup("slvr_krylov").get_value();
  value *krylov = &tkrylov;

  if(krylov->type == NO_TYPE) {
    yyerror("Error, variable slvr_krylov must be set to the krylov subspace "
	    "for the 3-argument version of solve_system.");
    err = 1;
    return mat;
  }

  value tmaxmult = syms->lookup("slvr_maxmult").get_value();
  value *maxmult = &tmaxmult;

  if(maxmult->type == NO_TYPE) {
    yyerror("Error, variable slvr_maxmult must be set to the ... "
	    "for the 3-argument version of solve_system.");
    err = 1;
    return mat;
  }

  value ttols = (syms->lookup("slvr_tols")).get_value();
  value *tols = &ttols;

  if(tols->type == NO_TYPE) {
    yyerror("Error, variable slvr_tols must be set to the relative and "
	    "maximum tolerances for the 3-argument version of solve_system.");
    err = 1;
    return mat;
  }

  return solve_system(mat, rhs, guess, solver_name, prec_name, arg_prec, 
		      prec, krylov, maxmult, tols, err);
}

/*======================= RANDOM ======================================*/

value * rando(value *nr, value *nc, value *nbnnz, value *max, int &err)
{
if (nr->type != NUM) {
    yyerror ("Error, 1st argument of random must be a number.");
    err = 1;
    return nr;
  } // end if

if (nc->type != NUM) {
    yyerror ("Error, 2nd argument of random must be a number.");
    err = 1;
    return nc;
  } // end if

if (nbnnz->type != NUM) {
    yyerror ("Error, 3rd argument of random must be a number.");
    err = 1;
    return nbnnz;
  } // end if

if (max->type != NUM) {
    yyerror ("Error, 4st argument of random must be a number.");
    err = 1;
    return max;
  } // end if

sparse *sptemp = new sparse((int)nr->num, (int)nc->num, (int)nbnnz->num, (int)max->num);
return new value(sptemp);

}

value * rando(value *nr, value *nc, value *max, int &err)
{
if (nr->type != NUM) {
    yyerror ("Error, 1st argument of random must be a number.");
    err = 1;
    return nr;
  } // end if

if (nc->type != NUM) {
    yyerror ("Error, 2nd argument of random must be a number.");
    err = 1;
    return nc;
  } // end if

if (max->type != NUM) {
    yyerror ("Error, 3st argument of random must be a number.");
    err = 1;
    return max;
  } // end if

dense *dtemp = new dense((int)nr->num, (int)nc->num);
dtemp->rando((int)max->num);
return new value(dtemp);

}

/*========================= INIT ===========================*/

value * init(value *nr, value *nc, value *val, int &err)
{
if (nr->type != NUM) {
    yyerror ("Error, 1st argument of init must be a number.");
    err = 1;
    return nr;
  } // end if

if (nc->type != NUM) {
    yyerror ("Error, 2nd argument of init must be a number.");
    err = 1;
    return nc;
  } // end if

if (val->type != NUM) {
    yyerror ("Error, 3rd argument of init must be a number.");
    err = 1;
    return val;
  } // end if

dense *dtemp = new dense((int)nr->num, (int)nc->num, val->num);
return new value(dtemp);

}


/*============================ MULTIC /COLORS ===============================*/

value * multic (value *mat, int &err) {

  if (mat->type != SPARSE) {
    yyerror ("Error, 1st argument of multic must be a sparse matrix.");
    err = 1;
    return mat;
  } // end if
  if (mat->smat->get_nrow() != mat->smat->get_ncol()) {
    yyerror ("Erorr, sparse matrix passed to multic must be square");
    err = 1;
    return mat;
  }
  
  sparse *sptemp =new sparse(*(mat->smat));
  sptemp->multic();
  return new value(sptemp); 
}

value * colors (value *mat, int &err) {

  if (mat->type != SPARSE) {
    yyerror ("Error, 1st argument of colors must be a sparse matrix.");
    err = 1;
    return mat;
  } // end if
  if (mat->smat->get_nrow() != mat->smat->get_ncol()) {
    yyerror ("Erorr, sparse matrix passed to colors must be square");
    err = 1;
    return mat;
  }
  
  sparse *sptemp =new sparse(*(mat->smat));
  int n = (mat->smat)->get_nrow();
  int *res= new int[n+1];
  res = sptemp->colors();
  sptemp->free();
  dense *dtemp = new dense(1,n+1);
  int i;
  for (i = 0; i<n+1; i++) dtemp->set_element(1,i+1,res[i]);
  delete [] res;

  return new value(dtemp); 
}

value * il_multic (value *mat, int &err) {

  if (mat->type != SPARSE) {
    yyerror ("Error, 1st argument of il_multic must be a sparse matrix.");
    err = 1;
    return mat;
  } // end if
  if (mat->smat->get_nrow() != mat->smat->get_ncol()) {
    yyerror ("Error, 1st argument must be a square sparse matrix");
    err = 1;
    return mat;
  }
  
  sparse *sptemp =new sparse(*(mat->smat));
  int n = mat->smat->get_nrow();
  int *res= new int[n+1];
  res = sptemp->il_multic();
  sptemp->free();
  int size = res[0];
  
  dense *dtemp = new dense(1,size+1);

  int i;
  for (i = 1; i<size+2; i++)
    dtemp->set_element(1,i,res[i]);

  delete [] res;
  
  return new value(dtemp); 
}

value * iord_multic (value *mat, int &err) {

  if (mat->type != SPARSE) {
    yyerror ("Error, 1st argument of iord_multic must be a sparse matrix.");
    err = 1;
    return mat;
  } // end if
  if (mat->smat->get_nrow() != mat->smat->get_ncol()) {
    yyerror ("Error, 1st argument must be a square sparse matrix");
    err = 1;
    return mat;
  }
  
  sparse *sptemp =new sparse(*(mat->smat));
  int n = mat->smat->get_nrow();
  int *res= new int[n];
  res = sptemp->iord_multic();
  sptemp->free();
  dense *dtemp = new dense(1,n);
  int i;
  for (i = 0; i<n; i++) dtemp->set_element(1,i+1,res[i]);
  delete [] res;
  
  return new value(dtemp); 
}


value * colors2 (value *mat, int &err, int &nco) {

  if (mat->type != SPARSE) {
    yyerror ("Error, 1st argument of colors2 must be a sparse matrix.");
    err = 1;
    return mat;
  } // end if

  sparse *sptemp =new sparse(*(mat->smat));

  int n = (mat->smat)->get_nrow();
  int *res= new int[n+1];

  res = sptemp->colors();

  sptemp->free();
  dense *dtemp = new dense(1,n);

  int i;
  for (i = 0; i<n; i++)
    dtemp->set_element(1,i+1,res[i]);

  nco = res[n];
  delete [] res;

  return new value(dtemp); 
}

/*=========================== INDSETX ===================================*/

value * indset0 (value *mat, int &err) {

  if (mat->type != SPARSE) {
    yyerror ("Error, 1st argument of indset0 must be a sparse matrix.");
    err = 1;
    return mat;
  } // end if
  
  sparse *sptemp = new sparse(*(mat->smat));
  sptemp->indset0();
  return new value(sptemp);
}

value * indset1 (value *mat, int &err) {

  if (mat->type != SPARSE) {
    yyerror ("Error, 1st argument of indset1 must be a sparse matrix.");
    err = 1;
    return mat;
  } // end if
  
  sparse *sptemp = new sparse(*(mat->smat));
  sptemp->indset1();
  return new value(sptemp);
}

