//////////////////////////////////////////////////////////
//
// io.C
// ~~~~
//  This source file contains the functions which handle
// the reading and writing of sparse matrices on the disk.
// It makes use of the MatrixMarket and Harwell-Boeing io 
// function libraries found online at Matrix Market
// (found in io subdirectory).
//
// Author: Ben Langmead (btl10@columbia.edu)
// Date:   June 29, 2001
//
//////////////////////////////////////////////////////////

#include "io/iohb.h"
#include "io/mmio.h"
#include "spmath.h"
#include <stdio.h>
#include <stdlib.h>

int readMM(FILE *, sparse *&);
int readHB(FILE *, sparse *&);

int save_matrix_to_file(char * filename, sparse &sp, char type[3]);

int load_matrix_from_file(char *filename, sparse * &spm) {
  FILE *f;
  MM_typecode mmtp;
  int err;

  if((f = fopen(filename, "r")) == NULL) {
    char second_try[128];
    sprintf(second_try, "%s/%s/%s", spmathpath, MATRIX_DIR_NAME, filename);

    if((f = fopen(second_try, "r")) == NULL) {
      sp_error("Error, could not open file \"");
      sp_error(filename);
      sp_error("\"\n");

      spm = NULL;
      return 0;
    }
  }

  if((err = mm_read_banner(f, &mmtp))) {
    // fprintf(stderr, "mm_read_banner aborted, error %d, trying HB...\n", err);
  }
  else if(!mm_is_valid(mmtp)) {
    //    fprintf(stderr, "mm_is_valid is false, trying HB...\n");
    return 0;
  }
  else {
    /* It's in matrix market format, read it in */

    if(!readMM(f, spm)) {
      goto bail;
    }

    fclose(f);
    return 1;
  }

  if(!checkHB(f)) {
    //    fprintf(stderr, "checkHB says file is not in HB format, aborting...\n");
  }
  else {
    /* It's in HB format, read it in */

    if(!readHB(f, spm)) {
      goto bail;
    }

    //    printf("readHB done, nz is %d\n", spm->get_nnz());
    //    cout << " READ HB DONE spm = " << spm << endl;

    fclose(f);
    return 1;
  }

 bail:

  fclose(f);
  spm = NULL;
  return (0);
}

/* readMM - reads in from an already opened matrix market type
   matrix file and stores the contents in sparse matrix m */

int readMM(FILE *f, sparse *&spm) {
  int M = 0, N = 0, nz = 0;
  int i[1], j[1];
  double val[1];
  int c;

  spm = new sparse();

  rewind(f);

  //  cout << "ABOUT TO mm_read_mtx_crd_size" << endl;
  
  if (mm_read_mtx_crd_size(f, &M, &N, &nz) != 0) {
    fprintf(stderr, "Error reading matrix file: mm_read_mtx_crd_size failed\n");
    return 0;
  }

  /* NOTE: when reading in doubles, ANSI C requires the use of the "l"  */
  /*   specifier as in "%lg", "%lf", "%le", otherwise errors will occur */
  /*  (ANSI C X3.159-1989, Sec. 4.9.6.2, p. 136 lines 13-15)            */

  //  cout << "ABOUT TO SO SOME fscanf's" << endl;

  for (c=0; c < nz; c++) {
    fscanf(f, "%d %d %lg\n", i, j, val);

    //    cout << "ABOUT TO NUDGE with i=" << i[0]
    //	 << " j=" << j[0] << " and val=" << val[0] << "...";

    i[0]--;  /* adjust from 1-based to 0-based */
    j[0]--;

    //    cout << "FINISHED" << endl;

    //    cout << "BLAH BLAH i=" << 17;

    //    cout << " j=" << 27 << " and val=" << 33.234 << "...";

    spm->set_element(i[0], j[0], val[0]);

    //    cout << "FINISHED" << endl;
  }

  //  cout << "ABOUT TO DECLARE SUCCESS" << endl;
  return 1;
}

/* readHB - reads in from an already opened Harwell-Boeing type
   matrix file and stores the contents in sparse matrix m */

int readHB(FILE *f, sparse *&spm) {
  int M = 0, N = 0, nz = 0, Nrhs = 0;
  int *ia, *ja;
  double *a;
  char *type;

  rewind(f);

  if(!readHB_info(f, &M, &N, &nz, &type, &Nrhs)) {
    fprintf(stderr, "Error reading matrix file: readHB_info failed\n");
    return 0;
  }

  spm = new sparse(M, N, nz);

  ia = spm->get_ia();
  ja = spm->get_ja();
  a = spm->get_a();

  rewind(f);

  if(!readHB_mat_double(f, 
			(char *)spm->get_title(), 
			(char *)spm->get_key(), 
			(char *)spm->get_type(), 
			ia, ja, a)) {

    fprintf(stderr, "Error reading matrix file: readHB_mat_double failed\n");
    delete spm;
    return 0;
  }

  //  cout << "About to transpose matrix" << endl;
  (*spm) = transp(*spm);

  return 1;
}


int save_matrix_to_file(char * filename, sparse &sp, 
			char type[4]) {

  // These are the fortran-like formatting strings that
  // determine the output of the values.

   char Ptrfmt[]="(13I6)";
   char Indfmt[]="(16I5)";
   char Valfmt[]="(3E26.18)";
   char Rhsfmt[]="(3E26.18)";

   // Transpose the matrix so it can be stored in CSR

   sparse sp2 = transp(sp);

   // See io/iohb.C for definition of writeHB_mat_double

   return (writeHB_mat_double(filename, sp2.get_nrow(), sp2.get_ncol(),
			      sp2.get_nnz(), sp2.get_ia(), sp2.get_ja(), 
			      sp2.get_a(), 0, NULL, NULL, NULL,
			      sp2.get_title(), sp2.get_key(), type,
			      Ptrfmt, Indfmt, Valfmt, Rhsfmt,
			      NULL));
}

