//****************************************************************
//
//      Class list
//
//      List is a class for implementing linked lists.  It is
//      written as a template for supporting multiple types of
//      data.  
//
//      If you discover any bugs, it would be very helpful if
//      you sent me e-mail.  Thank you.
//
//      Author:                Peter Braaten
//      e-mail:                braa0007@mermaid.micro.umn.edu
//      Date last modified:    July 12, 1995
//
//****************************************************************

/*============================================================================
			PRIVATE :
			--------
link * current_pointer
link * previous_pointer
link * top_pointer

			FUNCTIONS :
			----------
constructors
destructor

get_ptr : 	returns the pointer to the first element of the list
top : 		returns the first element of the list
pop : 		returns the first element of the list and remove it from the
		list
push(v) : 	puts v onto the front of the list
empty : 	returns 1 if the list is empty
end : 		returns 1 if the current pointer is at the last element of the
		list
init_curr_prev :initializes the current pointer to the top item of the
		list and initialize the previous pointer to NULL
get_current : 	returns the value of the item that is in the link pointed by the
		current pointer
get_next : 	returns the value of the item that is in the link following the
		link pointed by the current pointer
get_previous : 	returns the value of the item that is in the link pointed by the
		previous pointer
remove_top : 	removes the first item of the list
remove_current :removes the item which is in the link pointed by the current
		pointer
remove_all : 	removes all the items of the list

============================================================================*/


#include "list.h"

//#define debugging


//***********************************************************
//   Constructors
//***********************************************************

list::list (link *pt) {
  // Create an empty list
  top_pointer = pt;
  current_pointer = pt;
  previous_pointer = NULL;
} // end list

list::list (const list & x) {
  // Create a list.
  top_pointer = x.get_ptr();
  current_pointer = x.get_ptr();
  previous_pointer = NULL;
} // end list

list::list () {
  // Create an empty list;
  top_pointer = NULL;
  current_pointer = NULL;
  previous_pointer = NULL;
} // end list

//***********************************************************
//  selector
//***********************************************************

/*=========================================================================
#				GET_PTR :
#				---------
# Entries :	nothing 	
#
# Returns : 	the pointer to the first element of the list
#
=========================================================================*/

link* list::get_ptr ()  const {
  // return the pointer to the first element.
  return top_pointer;
} // end get_ptr

//***********************************************************
//  member functions.
//***********************************************************

/*=========================================================================
#				TOP :
#				----
# Entries :	nothing 	
#
# Returns : 	the first element of the list
#
=========================================================================*/

entry list::top () {
  // return the top element of list.
  return  top_pointer->get_item();
} // end top

/*=========================================================================
#				POP :
#				----
# Entries :	nothing 	
#
# Returns : 	the first element of the list
#
# Details : 	- It checks to make sure that the list is not empty
#		- It returns the first element of the list and removes it 
#		  from the list
=========================================================================*/

entry list::pop () {
  // return the top element of list and remove it from list.

  #ifdef debugging
    assert (!empty());
  #endif 

  entry temp = top_pointer->get_item();
  link *p = top_pointer;

  top_pointer = top_pointer->get_ptr();

  delete p;
  return temp;
} // end pop

/*=========================================================================
#				PUSH :
#				-----
# Entries :	an entry : val (the entry to put onto the front of the list) 	
#
# Returns : 	nothing
#
# Details :	It puts the value onto the front of the list
#
=========================================================================*/

void list::push (entry val) {
  // put value onto the front of a list.
  top_pointer = new link (val,top_pointer);
} // end push

/*=========================================================================
#				EMPTY :
#				------
# Entries :	nothing 	
#
# Returns : 	1 if the list is empty (top_pointer=NULL)
#		0 else
#
=========================================================================*/

int list::empty () {
  // return 0 if not empty, 1 if empty.
  if (top_pointer == NULL)
     return 1;
  else
     return 0;
} // end empty

/*=========================================================================
#				END :
#				----
# Entries :	nothing 	
#
# Returns : 	1 if the current pointer is at the end of the list
#		0 otherwise
#
=========================================================================*/

int list::end () {
  // Return true (1) if current_pointer is at the
  // last element of the list,  otherwise return
  // false (0).
  
  if (empty())
    return 1;

  return (current_pointer->get_ptr() == NULL);
} // end end

/*=========================================================================
#			INIT_CURR_PREV :
#			---------------
# Entries :	nothing 	
#
# Returns : 	nothing
#
# Details :	- initializes the current pointer to point to the top item 
#		  in the list
#		- initializes the previous pointer to NULL
#
=========================================================================*/

void list::init_curr_prev () {
  // Initialize current_pointer to point to the top item in
  // the list.  Initialize previous_pointer to NULL.

  current_pointer = top_pointer;
  previous_pointer = NULL;
} // end init_curr_prev

/*=========================================================================
#				GET_CURRENT :
#				------------
# Entries :	nothing 	
#
# Returns : 	an entry : the value of the item that is in the link pointed
#		to by the current pointer. (It first checks to make sure that
#		the current pointer is not NULL)
#
=========================================================================*/

entry list::get_current () {
  // return the value of the item that is in the link
  // pointed to by  current_pointer.

  #ifdef debugging
    assert (current_pointer != NULL);
  #endif

  return current_pointer->get_item();
} // end get_current

/*=========================================================================
#				GET_NEXT :
#				---------
# Entries :	nothing 	
#
# Returns : 	an entry : the value of the item that is in the link following
#		the link that is pointed to by the current pointer. 
#
# Details :	- it checks to make sure that the current pointer is not NULL
#		- it moves to the next_pointer which means that the current
#		  pointer becomes the previous one and that the pointer
#		  following the current pointer begins the current pointer.
#		- it checks to make sure that, NOW, the current pointer is not
#		  NULL and returns it
#
=========================================================================*/

entry list::get_next () {
  // return the value of the item that is in the link
  // following the link that is pointed to by 
  // current_pointer.

  #ifdef debugging
    assert (current_pointer != NULL);
  #endif

  previous_pointer = current_pointer;
  current_pointer = current_pointer->get_ptr();

  #ifdef debugging
    assert (current_pointer != NULL);
  #endif

  return current_pointer->get_item();
} // end get_next

/*=========================================================================
#				GET_PREVIOUS :
#				------------
# Entries :	nothing 	
#
# Returns : 	an entry : the value of the item that is in the link that is pointed to by the previous_pointer 
#
# Details :	- it checks to make sure that the previous pointer is not NULL
#		- it returns the previous pointer
#
=========================================================================*/

entry list::get_previous () {
  // return the value of the item that is  in the link
  // that is pointed to by previous_pointer.

  #ifdef debugging
    assert (previous_pointer != NULL);  
  #endif

  return previous_pointer->get_item();
} // end get_previous

/*=========================================================================
#				REMOVE_TOP :
#				----------
# Entries :	nothing 	
#
# Returns :  	nothing 
#
# Details :	- it checks to make sure that the list isn't empty
#		- it stores the top_pointer into a link p
#		- the pointer of the top_pointer becomes the top_pointer
#		- it deletes p
#
=========================================================================*/

void list::remove_top() {
  #ifdef debugging
    assert (!empty());
  #endif 

  link *p = top_pointer;

  top_pointer = top_pointer->get_ptr();

  p->get_item().free();
  delete p;
} // end remove top

/*=========================================================================
#			REMOVE_CURRENT :
#			--------------
# Entries :	nothing 	
#
# Returns :  	nothing 
#
# Details :	- it checks to make sure that the current pointer isn't NULL
#		- if the current_pointer is the top_pointer, it simply calls
#		  remove_top
#		- the pointer of the previous_pointer becomes the pointer
#		  of the current _pointer
#		- it deletes the current_pointer
#		- it calls init_cuur_prev (current_pointer=top_pointer)
#
=========================================================================*/

void list::remove_current () {
  // Remove the link that is pointed to by 
  // current_pointer.

  #ifdef debugging
    assert (current_pointer != NULL);
  #endif

  // If the current element is the front of the list,
  // simply remove the top of the list.
  if (previous_pointer == NULL) {
    remove_top();
    return;
  } // end if
  
  previous_pointer->set_pointer(current_pointer->get_ptr());

  current_pointer->get_item().free();
  delete current_pointer;
  init_curr_prev ();
} // end remove_current

/*=========================================================================
#				REMOVE_ALL :
#				----------
# Entries :	nothing 	
#
# Returns :  	nothing 
#
# Details :	It removes all the elements of the list: While the list 
#		is empty, it removes the top element of the list.
#
=========================================================================*/

void list::remove_all() {
  // Remove all items from the list.

  while (!empty())
    remove_top();

} // end remove_all
