//////////////////////////////////////////////////////////
//
// output.C
// ~~~~~~~~
//  This file contains all the methods that output 
// messages to the user.  The methods are abstracted so
// that the same methods can be called to output in both
// tk and no_tk modes.
//
// Author: Ben Langmead (btl10@columbia.edu)
// Date:   June 29, 2001
//
//////////////////////////////////////////////////////////

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "spmath.h"

// Preferred output character limit (for tk)
unsigned int char_limit = 10;

//
// sp_out(const char *str)
//
//  Outputs a string.  Used for results, man pages,
// help, etc...
//

void sp_out(const char *str) {
  if(!supress_output) {

#ifndef NO_TK_COMPILE

    if(!no_tk) {
      char * buf = new char[3*strlen(str) + 13];
      int len = strlen(str);
    
      strcpy(buf, "text_out \"");
    
      int ind = 0;
    
      // Escape all special characters
    
      while(1) {
	ind = strcspn(str, "$[]{}\"\'\\\t");
	strncat(buf, str, ind);
	if(ind == len)
	  break;
	else {
	  str = str+ind;
	  strcat(buf, "\\");
	  strncat(buf, str, 1);
	  str++;
	  len -= (ind+1);
	}
      }
    
      strcat(buf, "\"");
    
      // Calls text_out in procs.tcl
      TCL( buf );

      delete buf;
    }
    else {
#endif
      cout << str;
#ifndef NO_TK_COMPILE
    }
#endif
  }
}

//
// sp_out(double d)
//
//  Outputs a double (using char_limit characters in
// tk mode)
//

void sp_out(double d) {
  if(!supress_output) {

#ifndef NO_TK_COMPILE

    if(!no_tk) {
      char * buf = new char[char_limit*2];
      int i = 1;
    
      do {
	gcvt(d, char_limit-i, buf);
	i++;
      } while(strlen(buf) > char_limit);
    
      sp_out(buf);
    
      delete buf;
    }
    else {
#endif
      cout << d;
#ifndef NO_TK_COMPILE
    }
#endif
  }
}

//
// sp_out(float f)
//
//  Outputs a float (using char_limit characters in
// tk mode)
//

void sp_out(float f) {
  if(!supress_output) {

#ifndef NO_TK_COMPILE

    if(!no_tk) {
      char * buf = new char[char_limit*2];
      int i = 1;
    
      do {
	gcvt((double)f, char_limit-i, buf);
	i++;
      } while(strlen(buf) > char_limit);
    
      sp_out(buf);
    
      delete buf;
    }
    else {
#endif
      cout << f;
#ifndef NO_TK_COMPILE
    }
#endif
  }
}

//
// sp_out(int i)
//
//  Outputs an int (using char_limit characters in
// tk mode)
//

void sp_out(int i) {
  if(!supress_output) {

#ifndef NO_TK_COMPILE

    if(!no_tk) {
      char *buf = new char[65];
      int j = 3;
    
      snprintf(buf, 64, "%d", i);
    
      do {
	gcvt((double)i, char_limit-j, buf);
	j++;
      } while(strlen(buf) > char_limit);
    
      sp_out(buf);
    
      delete buf;
    }
    else {
#endif
      cout << i;
#ifndef NO_TK_COMPILE
    }
#endif
  }
}

//
// sp_out(long l)
//
//  Outputs a long (using char_limit characters in
// tk mode)
//

void sp_out(long l) {
  if(!supress_output) {

#ifndef NO_TK_COMPILE

    if(!no_tk) {
      char *buf = new char[65];
      int i = 3;
    
      snprintf(buf, 64, "%ld", l);
    
      do {
	gcvt((double)l, char_limit-i, buf);
	i++;
      } while(strlen(buf) > char_limit);
    
      sp_out(buf);
    
      delete buf;
    }
    else {
#endif
      cout << l;
#ifndef NO_TK_COMPILE
    }
#endif
  }
}

//
// sp_error(const char *err)
//
//  Outputs an error message to the user.
//

void sp_error(const char *err) {
  if(!supress_output) {

#ifndef NO_TK_COMPILE

    if(!no_tk) {
      char * buf = new char[strlen(err) + 14];
    
      // Calls error_out in procs.tcl

      strcpy(buf, "error_out {");
      strcat(buf, err);
      strcat(buf, "}\n");
    
      TCL( buf );
    
      delete buf;
    }
    else {
#endif
      cerr << err;
#ifndef NO_TK_COMPILE
    }
#endif
  }
}

//
// sp_error(double d)
//
//  Outputs a double (using char_limit characters in
// tk mode)
//

void sp_error(double d) {
  if(!supress_output) {

#ifndef NO_TK_COMPILE

    if(!no_tk) {
      char * buf = new char[char_limit*2];
      int i = 1;
    
      do {
	gcvt(d, char_limit-i, buf);
	i++;
      } while(strlen(buf) > char_limit);
    
      sp_error(buf);
    
      delete buf;
    }
    else {
#endif
      cout << d;
#ifndef NO_TK_COMPILE
    }
#endif
  }
}

//
// sp_error(float f)
//
//  Outputs a float (using char_limit characters in
// tk mode)
//

void sp_error(float f) {
  if(!supress_output) {
#ifndef NO_TK_COMPILE
    if(!no_tk) {
      char * buf = new char[char_limit*2];
      int i = 1;
    
      do {
	gcvt((double)f, char_limit-i, buf);
	i++;
      } while(strlen(buf) > char_limit);
    
      sp_error(buf);
    
      delete buf;
    }
    else {
#endif
      cout << f;
#ifndef NO_TK_COMPILE
    }
#endif
  }
}

//
// sp_error(int i)
//
//  Outputs an int (using char_limit characters in
// tk mode)
//

void sp_error(int i) {
  if(!supress_output) {
#ifndef NO_TK_COMPILE
    if(!no_tk) {
      char *buf = new char[65];
      int j = 3;
    
      snprintf(buf, 64, "%d", i);
    
      do {
	gcvt((double)i, char_limit-j, buf);
	j++;
      } while(strlen(buf) > char_limit);
    
      sp_error(buf);
    
      delete buf;
    }
    else {
#endif
      cout << i;
#ifndef NO_TK_COMPILE
    }
#endif
  }
}

//
// sp_error(long l)
//
//  Outputs a long (using char_limit characters in
// tk mode)
//

void sp_error(long l) {
  if(!supress_output) {
#ifndef NO_TK_COMPILE
    if(!no_tk) {
      char *buf = new char[65];
      int i = 3;
    
      snprintf(buf, 64, "%ld", l);
    
      do {
	gcvt((double)l, char_limit-i, buf);
	i++;
      } while(strlen(buf) > char_limit);
    
      sp_error(buf);
    
      delete buf;
    }
    else {
#endif
      cout << l;
#ifndef NO_TK_COMPILE
    }
#endif
  }
}

//
// sp_prompt()
//
//  Outputs a prompt, depending on whether we're in
// nested mode or not.
//
void sp_prompt() {
  if(!supress_output) {

#ifndef NO_TK_COMPILE
    if(no_tk) {
#endif

      if(nest_count > 0) {
	cout << "->        ";
      }
      else {
	cout << "spmath:" << hist_num << "> ";
      }
#ifndef NO_TK_COMPILE
    }
    else {
      if(nest_count > 0) {
	TCL("put_nested_prompt");
      }
      else {
	TCL("put_prompt");
      }
    }
#endif
  }
}

//
// sp_command(const char *com)
//
//  Outputs a command onto the command line, overwriting
// the existing command.  Used to retrive history items
// and display them on the current input line.
//

void sp_command(const char *com) {
  if(!supress_output) {

#ifndef NO_TK_COMPILE

    if(!no_tk) {
      char *buf = new char[strlen(com) + 15];
      
      // Calls command_out in procs.tcl
      
      strcpy(buf, "command_out {");
      strcat(buf, com);
      strcat(buf, "}");
      
      TCL( buf );
      
      delete buf;
    }
    else {

#endif
      cout << com;
      // Cannot be called from command-line version...
      //
      // Also is difficult to know how to replace the
      // command with a different one on the console.
#ifndef NO_TK_COMPILE
    }
#endif
  }
}




