// ****************************************************************
//
//    Class sparse
//    
//    Description:  
//      Sparse is a class for storing, and manipulating sparse
//      matrices.  It is realy just an interface to the 
//      SPARSKIT library written by Professor Youcef Saad.
//
//    Problems:
//      Currently it only stores the values of the matrices
//      as doubles.  Later it should be templated to support
//      integers as well.
//    Remarks:
//	The format of the sparse matrices is CSC.
//
//      If you discover any bugs, it would be very helpful if
//      you sent me e-mail.  Thank you.
//
//      Author:                Peter Braaten
//      e-mail:                braa0007@mermaid.micro.umn.edu
//      Date last modified:    July 28, 1994
//
// ****************************************************************

/* ================================================================

				PRIVATE :
				--------

    int nrow;		// row dimension of the sparse matrix
    int ncol;		// col dimension of the sparse matrix
    int nnz;		// number of non-zero elements
    int *ia;		// array containing the pointers to the beginning of
			// each column
    int *ja;		// array containing the row indices of the elements
			// stored in a
    double *a;		// array containing the non zero elements
    char title[73];
    char key[9];
    char type[4];

    double *rhs;
    double *guess;
    double *exact;

				FUNCTIONS :
				----------

constructors
destructors

amux  :		   multilplies a matrix with a vector
atmux :		   multiplies the transpose of a matrix with a vector
cnrms : 	   computes the norms of the columns
coscal : 	   scales the columns of a matrix by their norms
csrmsr :	   create a matrix in modified row format from a matrix 
		   in compressed row format
csort  : 	   sorts the elements in increasing order of columns
filter : 	   filters elements from a matrix according to their magnitude
getu :	  	   extracts the upper triangular part of a matrix
getl :		   extracts the lower triangular part of a matrix
getbwd :	   gives the bandwith of the upper and the lower part of the mat
getbwdu :	   gives the bandwith of the upper part of the matrix
getbwdl :	   gives the bandwith of the lower part of the matrix
getdia : 	   extracts a specified diagonal from a matrix.
gen57pt :	   generates a 5-point or 7-point matrix (see genmat.f)
get_nrow:	   returns the row dimension of the matrix
get_ncol : 	   returns the col dimension of the matrix
get_nnz : 	   returns the number of non zero elements
get_element(i,j) : returns the value of the element whose coordinates are (i,j)
get_ia : 	   returns ia
get_ja : 	   returns ja
get_a : 	   returns a
get_guess :        returns guess vector
get_rhs :          returns right-hand-side vector
get_exact :        returns exact value vector
ilut, iluk : 	   preconditionner
ilud, ilu0 :       preconditionner
init_zero : 	   initializes all the elements of the matrix to zero
init_identity :    makes the matrix into an identity matrix
indset0 :	   greedy algorithm for independent set ordering
indset1 :	   idem  but using minimal degree traversal
msrcsr : 	   create a matrix in compressed row format from a matrix
		   in modified row format 
multic,colors :    greedy algorithm for multicoloring (the only difference
		   between these functions is what they return)
prtmt(f) : 	   prints the matrix in the Harwell-Boeing file f
readmt(f) : 	   reads the matrix stored in the Harwell-Boeing file f
rmdia : 	   removes a specified diagonal from a matrix.
roscal : 	   scales the rows of a matrix by their norms
set_element(i,j,v) : sets the element whose coordinates are (i,j) to the value v
solve_system :	   solve a linear system Ax = y where A is a sparse matrix.
submat :	   extracts a submatrix from a sparse one.

operators:         =, <<=, plus, minus, *

filter(mat,f) :    applies the function f to all the elements of mat
output(mat) :      outputs the matrix mat
rperm :            calculates the row permutation of a sparse matrix
transp :           calculates the transpose of a sparse matrix
floor, trunc, ceiling, round, abs
irand(n,x) :	   generates a random integer between 0 and n
rnd0nb(n,x) :	   generates a random integer between 0 and n

FUNCTIONS ONLY USED BY SOLVE_SYSTEM (not available in spmath)

lusol, lutsol:     preconditionner

========================================================================*/

#include <time.h>
#include "spmath.h"

//#define debugging

//************************************************************************
//    Member constructors
//************************************************************************


sparse::sparse () {
  // Create an empty sparse matrix.

  int i;             // Simple loop counter.

  ia = new int[1];
  ja = new int[1];
  a = new double[1];

  nrow = 0;
  ncol = 0;
  nnz = 0;

  for (i = 0; i <= 72; i++) 
    title[i] = ' ';
  for (i = 0; i <= 8; i++) 
    key[i] = ' ';
  for (i = 0; i <= 3; i++) 
    type[i] = ' ';
  
} // end constructor sparse

sparse::sparse (const sparse &y) {
  // Copy constructor.

  int i;             // Simple loop counter.

  nrow = y.get_nrow();
  ncol = y.get_ncol();
  nnz = y.get_nnz();

  ia = new int[nrow+1];
  if(!ia) { return; }
  ja = new int[nnz];
  if(!ja) { delete[] ia; return; }
  a  = new double[nnz];
  if(!a) { delete[] ia; delete[] ja; return; }

  for (i = 0; i <= nrow; i += 1) 
    ia[i] = y.get_ia()[i];

  for (i = 0; i <= nnz -1; i += 1) {
    ja[i] = y.get_ja()[i];
    a[i] = y.get_a()[i];
  }

  for (i = 0; i <= 72; i++) 
    title[i] = y.get_title()[i];
  for (i = 0; i <= 8; i++) 
    key[i] = y.get_key()[i];
  for (i = 0; i <= 3; i++) 
    type[i] = y.get_type()[i];
  
} // end constructor sparse

sparse::sparse (int nro, int nco, double *mat) {
  // Create a a sparse matrix from a dense one.

  int i,j;             // Simple loop counter.

  init_zero(nro,nco);

  for (i = 1; i <= nrow; i += 1)
    for (j = 1; j <= ncol; j += 1)
      if (mat[(i-1)*ncol + (j-1)] != 0)
        set_element (i,j,mat[(i-1)*ncol + (j-1)]);

  for (i = 0; i <= 72; i++) 
    title[i] = ' ';
  for (i = 0; i <= 8; i++) 
    key[i] = ' ';
  for (i = 0; i <= 3; i++) 
    type[i] = ' ';
  
}  // end constructor sparse

sparse::sparse(int nr, int nc, int nz) {
  int i;

  nrow = nr;
  ncol = nc;
  nnz = nz;

  a = new double[nnz];
  if(!a) { return; }
  ja = new int[nnz];
  if(!ja) { delete[] a; return; }
  ia = new int[nrow+1];
  if(!ia) { delete[] ja; delete[] a; return; }

  for (i = 0; i <= 72; i++) 
    title[i] = ' ';
  for (i = 0; i <= 8; i++) 
    key[i] = ' ';
  for (i = 0; i <= 3; i++) 
    type[i] = ' ';
}


// Note: This constructor is not currently used
//
// Warning: aa, j and i must not be allocated using malloc.  When
//          a sparse instance is deleted, the memory taken by
//          these arrays is freed using delete [], thus they must
//          be allocated using new [].

sparse::sparse (int nr, int nc, int nz, double *aa, int *j, int *i) {
  // Create an initialized sparse matrix.

  int k;       // Simple loop index.

  nrow = nr;
  ncol = nc;
  nnz = nz;
  a = aa;
  ja = j;
  ia = i;

  for (k = 0; k <= 72; k++) 
    title[k] = ' ';
  for (k = 0; k <= 8; k++) 
    key[k] = ' ';
  for (k = 0; k <= 3; k++) 
    type[k] = ' ';
  
  
} // end constructor sparse

// Note: This constructor is not currently used

sparse::sparse(int nr, int nc, int max_nb_nnz_row, int max) {

  // Generate a sparse matrix whose entries are random 
  // doubles between 0 and max

  nrow = nr;
  ncol = nc;
  nnz = 0;
  a = new double[max_nb_nnz_row + nrow];
  ja = new int[max_nb_nnz_row + nrow];
  ia = new int[nrow+1];
  init_zero(nr,nc); 
  
  int ct_nnz=0;  // count the number of nnz elements per row
  int k;         // simple loop index
  int j;
  
  for (k=0;k<nrow;k++) {
    for (ct_nnz=0;ct_nnz<max_nb_nnz_row;ct_nnz++) { 
      j = irnd(ncol+1,0.0);
      set_element(k+1,j, rnd0nb(max,0.0));
    }
  } 
  
}

//************************************************************************
//    Member destructor
//************************************************************************

void sparse::free () {
  // Free up dynamically allocated memory.

  if(ia) { delete[] ia; ia = NULL; }
  if(ja) { delete[] ja; ja = NULL; }
  if(a)  { delete[] a;  a = NULL; }

  nrow = 0;
  ncol = 0;
  nnz = 0;

}

sparse::~sparse() {
  // Free up dynamically allocated memory.

  //  if(ia) { delete[] ia; }
  //  if(ja) { delete[] ja; }
  //  if(a)  { delete[] a;  }
}

//************************************************************************
//    Member selectors
//************************************************************************

/*=========================================================================
#			GET_NROW :
#			---------
# Entries : none
# Returns : the row dimension of the matrix
#
=========================================================================*/

int sparse::get_nrow () const {
 // Return the row dimension of the matrix.

  return nrow;
} // end get_nrow

/*=========================================================================
#			GET_NCOL :
#			---------
# Entries : none
# Returns : the col dimension the matrix
#
=========================================================================*/

int sparse::get_ncol() const {
  // Return the column dimension of the matrix.

  return ncol;
} // end get_ncol

/*=========================================================================
#			GET_NNZ :
#			---------
# Entries : none
# Returns : the number of non zero elements of the matrix
#
=========================================================================*/

int sparse::get_nnz() const {
  // Return the number of non zero elements.

   return nnz;
} // end get_nnz

/*============================================================================
#				GET_ELEMENT		
				-----------    	
# in entry  : 	i -> row of the element
#		j -> column of the element
#
# in return :   The value of the element whose coordinates are (i,j)
#
# details   :	- It calculates upper : the last element of the matrix whose 
#		  row is i. (ia[i-1] is the first element of row i)
#		- among all the elements whose row is i, it searchs if one has 
#		  j as col coordinate.
#
============================================================================*/

double sparse::get_element (int i, int j) const {
  // Return the value of the element at i,j
  
  int k;
  int upper = ia[(i-1) +1] -1 -1;

  for (k = ia[(i-1)]-1; k <= upper; k++) 
    if (ja[k] == j)
      return a[k];
 
  return 0.0;

} // end get_element

/*============================================================================
#				GET_IA		
#				------    	
# in entry  : 	nothing
#
# in return :   ia
#
============================================================================*/

int* sparse::get_ia () const {
  return ia;
} // end get_ia

/*============================================================================
#				GET_JA		
#				------    	
# in entry  : 	nothing
#
# in return :   ja
#
============================================================================*/

int* sparse::get_ja () const {
  return ja;
} // end get_ja

/*============================================================================
#				GET_A		
#				-----    	
# in entry  : 	nothing
#
# in return :   a
#
============================================================================*/

double* sparse::get_a () const {
  return a;
} // end get_a


double* sparse::get_guess() const {
  return guess;
}

double* sparse::get_rhs() const {
  return rhs;
}

double* sparse::get_exact() const {
  return exact;
}

void sparse::set_guess(double *g) {
  guess = g;
}

void sparse::set_rhs(double *r) {
  rhs = r;
}

void sparse::set_exact(double *e) {
  exact = e;
}


const char *sparse::get_title() const {
  return title;
}

const char *sparse::get_key() const {
  return key;
}

const char *sparse::get_type() const {
  return type;
}

//************************************************************************
//    Member Functions
//************************************************************************

/*============================================================================
#				SET_ELEMENT		
#				-----------    	
# in entry  : 	i     -> row of the element
#		j     -> column of the element
#		value -> value you want to give to the element
#
# in return :	nothing
#
# details   :	- it calculates the index of the last element whose row is i
#		- if NNZ = O (matrix composed by 0)
#		    + it inserts the value in a 
#		    + it increments the number of non-zero elements
#		    + it changes the contents of ia from i to the end (the beginning of each of these rows has been "decale")
#		    + it stores j in ja
#		- if the element EXISTS :
#		    + it searchs it (see get_element) and simply changes it
#		- if it DOESN'T EXIST :
#		    + it stores the old matrix in *_old
#		    + it initializes the matrix with the new dimensions (nnz)
#		    + it copies all the element until the end of the i'th row 
#		    + it adds the element (in a and ja) 
#		    + it copies the rest of the element
#		    + it adds 1 to each row pointers (ia) after the i'th row
#		    + it deletes the old arrays (*_old)
#
============================================================================*/

void sparse::set_element (int i, int j, double value) {
  // Set the value of the element i,j
  // If the element already exists, simply change it.
  // If it does not already exist, add the element.
  int k;

  int upper = ia[(i-1) +1] -1 -1;

  if (nnz == 0) { 
    nnz = 1;
    ja[0] = j;
    a[0] = value;
    for (k = i; k <= nrow; k += 1) 
      ia[k] += 1;
    return; 
  }  // end if

  // If the element already exists, simply change it.
  for (k = ia[(i-1)]-1; k <= upper; k++) 
    if (ja[k] == j) {
      a[k] = value;
      return;
    }

  // If it does not already exist, add the element.
  int *ja_old = ja;
  double *a_old = a;
  
  nnz = nnz + 1;
  ja = new int [nnz];
  a = new double [nnz];

  // Copy all of the elements up until the end of the ith row.
  for (k = 0; k <= upper; k += 1) {
    ja[k] = ja_old[k];  
    a[k] = a_old[k];
  }

  // Add in the new element.
    ja[upper+1] = j;
    a[upper+1] = value;

  // Copy the rest of the elements.
  for (k = upper+2; k <= nnz -1; k += 1) {
    ja[k] = ja_old[k-1];
    a[k] = a_old[k-1];    
  }

  // Add one to each of the row pointers after the ith row.
  for (k = i; k <= nrow; k += 1) 
    ia[k] += 1;

  // Free up the space from the old arrays.
  delete [] ja_old;
  delete [] a_old;

  return;
} // set_element

/*============================================================================
#				GEN57PT		
#				------    	
# in entry  : 	nx -> number of grid points in x directions (integer)
#		ny -> number of grid points in y directions (integer)
#		nz -> number of grid points in z directions (integer)
#		a0, a1, a2, a3, a4, a5 -> the coefficients of the 
#		boundaries condition				
#
# in return :	nothing
#
# details   :	- it calculates nb the maximum of the dimension of the matrix
#		- it frees the memory in the sparse matrix
#		- it allocates enough space for the matrix (see genmat.f)
#		  the max of non zero elements is 7*nb
#		- It allocates space for alpha, n, iau, and rhs
#		- It calls gen57pt_matrix (function of csparse.c)
#		- It gets the dimension of the generated matrix 
#		- It frees all the memory
#
============================================================================*/

void sparse::gen57pt (int nx, int ny, int nz, double a0, double a1, double a2, double a3, double a4, double a5, int *iau, double *rhs)
{
	int nb=nx*ny*nz;

  // Free any memory in the sparse matrix.
 	free();

  // Allocate space for the new matrix.
	nrow=nb;
	ncol=nb;
	nnz=7*nb;
	a = new double[nnz];
	ja = new int [nnz];
	ia = new int [nrow+1];

  // Allocate space for n, alpha
	double al[6];
	int n[1];
	

  // Create al
	al[0]=a0; al[1]=a1; al[2]=a2; al[3]=a3; al[4]=a4; al[5]=a5;

  // Call gen57pt_
	gen57pt_matrix(nx, ny, nz, al, 0, n, a, ja, ia, iau, rhs);

  // Get the dimension of the new matrix
	nrow=*n;
	ncol=*n;
	nnz=ia[nrow]-1;

  // Free the memory
	//delete [] n;
	//delete [] al;

}


void sparse::gen57pt (int nx, int ny, int nz, double a0, double a1, double a2, double a3, double a4, double a5) {
 
	int nb=nx*ny*nz;

  // Free any memory in the sparse matrix.
 	free();

  // Allocate space for the new matrix.
	nrow=nb;
	ncol=nb;
	nnz=7*nb;
	a = new double[nnz];
	ja = new int [nnz];
	ia = new int [nrow+1];

  // Allocate space for n, alpha, iau and rhs
	double al[6];
	double *rhs = new double[nb]; //(double *)malloc(nb*sizeof(double));
	int *iau= new int[nb]; // (int *)malloc(nb*sizeof(int));
	int n[1];
	

  // Create al
	al[0]=a0; al[1]=a1; al[2]=a2; al[3]=a3; al[4]=a4; al[5]=a5;

  // Call gen57pt_
	gen57pt_matrix(nx, ny, nz, al, 0, n, a, ja, ia, iau, rhs);

  // Get the dimension of the new matrix
	nrow=*n;
	ncol=*n;
	nnz=ia[nrow]-1;

  // Free the memory
	// delete [] n;
	delete [] iau;
	delete [] rhs;
	// delete [] al;

} 

/*============================================================================
#				SUBMAT		
#				------    	
# in entry  : 	n -> the row dimension of the submatrix
#		i1, i2, j1, j2 : the indecis of the first and the last 
#		element to extract.
#
# in return :	nothing
#
# details   :	- it stores the values of the matrix in another one (input)
#		- it foresees the new dim of the submatrix
#		- it allocates enough space for this submatrix
#		- it calls submat_matrix of csparse.c
#		- it frees all the memor  
#
============================================================================*/

void sparse::submat(int n, int i1, int i2, int j1, int j2)

{

   // Store the values of the matrix
	sparse input(*this);

  //Foresee the dim of the submatrix	
	int nro = i2 - i1 + 1;
	int nco = j2 - j1 +1;

  //Allocate space for the submatrix
	free();
	a= new double[nro * nco];
	ja = new int[nro*nco];
	ia = new int [nrow+1];

  // Allocate enough space for the results
	int nr[1];
	int nc[1];

	submat_matrix(n,1,i1,i2,j1,j2,input.get_a(), input.get_ja(), input.get_ia(), nr, nc, a, ja, ia);

   // Modify the dim 
	nrow=*nr;
	ncol=*nc;
	nnz=ia[nrow]-1;



   // Free the memory
	input.free();
}


double sparse::region_avg(int i1, int j1, int i2, int j2) {
  double tot = 0.0;
  int i, j, klo, khi, k;

  for(i = i1; i < i2; i++) {

    klo = ia[i]-1;
    khi = ia[i+1]-1;

    for(k = klo; k < khi; k++) {
      j = ja[k]-1;

      if(j < j1) {
	continue;
      }
      else if(j >= j2) {
	break;
      }

      tot += fabs(a[k]);
    }
  }

  return tot/((i2-i1)*(j2-j1));
}


//
//
//
//
//

double sparse::region_avg2(int i1, int j1, int i2, int j2, int use_last_k) {
  double tot = 0.0;
  int i, j, klo, khi, k = 0;

  // This variable holds whatever value k ended up
  // at the last time this method was called

  static int lastk = 0;

  for(i = i1; i < i2; i++) {

    klo = ia[i]-1;
    khi = ia[i+1]-1;

    if(use_last_k) {

      for(k = lastk; k < khi; k++) {
	j = ja[k]-1;

	if(j < j1) {
	  continue;
	}
	else if(j >= j2) {
	  break;
	}

	tot += fabs(a[k]);
      }
    }
    else {

      // Do it the regular way

      for(k = klo; k < khi; k++) {
	j = ja[k]-1;
	
	if(j < j1) {
	  continue;
	}
	else if(j >= j2) {
	  break;
	}
	
	tot += fabs(a[k]);
      }      
    }
  }

  lastk = k;
  return tot/((i2-i1)*(j2-j1));
}

/*============================================================================
#				FILTER		
#				------    	
# in entry  : 	n -> the row dimension of the submatrix
#		job -> job indicator
#		drptol -> drop tolerance
#
# in return :	nothing
#
# details   :	This module removes any elements whose absolute value is small
#		from an input matrix A and puts the resulting matrix in B.  
#		The input parameter job selects a definition of small.
#		- it stores the values of the matrix in another one (input)
#		- it allocates enough space for the resulting matrix
#		- it calls filter of csparse.c
#		- it frees all the memory  
#
============================================================================*/

void sparse::filter(int n, int job, double drptol)

{
  // Store the values of the matrix
	sparse input(*this);

 
  //Allocate space for the submatrix
	free();
	nrow = input.get_nrow();
	ncol= input.get_ncol();
	nnz=input.get_nnz();
	a= new double[nnz];
	ja = new int[nnz];
	ia = new int [nrow+1];

  
	filter_matrix(n, job, drptol, input.get_a(), input.get_ja(), input.get_ia(), a, ja, ia, nnz, 0);

   // Modify the dim 
	
	nnz=ia[nrow]-1;



   // Free the memory
	input.free();

}

/*============================================================================
#				CSORT		
#				------    	
# in entry  : 	values -> job indicator
#
# in return :	nothing
#
# details   :	It sorts the elements in increasing order of columns
#		- It allocates enough space for the work array iwork
#		- If values =0
#			- it calls csort_matrix from csparse.c but without a
#			  as an argument. (so with a dummy "a")
#		- Else it just calls csort_matrix from csparse.c
#		  It frees the memory  
#
# Warning: This allocates an amount of memory proprtional to the
# ~~~~~~~  number of nonzero elements in the matrix TIMES its
#          dimension.  This spends a lot of memory.  Look for the
#          "new" statement to fail for large matrices.
#
============================================================================*/

void sparse::csort(int values)
{
  // Allocate enough space for the work array:
	int *iwork;
	iwork = new int[2*nnz*(nrow+1)];
	
  // Call the function:
	
	if (values==0)
		{
		double *atemp;
		atemp=new double[nnz];
		csort_matrix(nrow, atemp, ja, ia, iwork, values);
		delete [] atemp;
		}

	else {
		csort_matrix(nrow, a, ja, ia, iwork, values);
		}

  // Free the memory	
	delete [] iwork;
}

/*============================================================================
#				GETDIA		
#				------    	
# in entry  : 	ioff -> int (number of the diagonal) 
#
# in return : 	an array of double (elements of the diag)
#
# details   : 	It extracts a diagonal from a dense matrix
#		- it allocates enough memory for the work array and the results
#		- It calls getdia_matrix from csparse.c with job=0
#		- It frees the memory 
#
============================================================================*/

double * sparse::getdia(int ioff)

{
  // Allocate enough space
	double *diag;
	int *idiag;
	int len[1];
	diag = new double[nrow];
	idiag = new int[nrow];

  // Call the function
	getdia_matrix(nrow, ncol, 0, a, ja, ia, len, diag, idiag, ioff);
	
  // Free the memory
	delete [] idiag;

  // Return the result
	return diag;

}

/*============================================================================
#				RMDIA		
#				-----    	
# in entry  : 	ioff -> int (number of the diagonal) 
#
# in return : 	nothing
#
# details   : 	It removes a diagonal from a dense matrix
#		- it allocates enough memory for the work array and the results
#		- It calls getdia_matrix from csparse.c with job=1
#		- It frees the memory 
#
============================================================================*/

void sparse::rmdia(int ioff)

{
  // Allocate enough space
	double *diag;
	int *idiag;
	int len[1];
	diag = new double[nrow];
	idiag = new int[nrow];

  // Call the function
	getdia_matrix(nrow, ncol, 1, a, ja, ia, len, diag, idiag, ioff);
	
  // Free the memory
	delete [] idiag;
	delete [] diag; 

}

/*==========================================================================
				MAX_ELEMENT

Preconditions:
 none

Postconditions:
 the element with the greatest magnitude's absolute value is returned
 as a double-precision real

===========================================================================*/

double sparse::max_element() {
  double max = 0.0;
  int i;

  for(i = 0; i < nnz; i++) {
    if(fabs(a[i]) > max) {
      max = fabs(a[i]);
    }
  }

  return max;
}


/*============================================================================
#				GETL/GETU		
#				---------    	
# in entry  : 	nothing 
#
# in return : 	nothing
#
# details   : 	It extracts either the lower or the upper triangular part 
#		of the dense matrix.
#		- It stores the old matrix in input
#		- It frees it and allocates enough space for the new one
#		- It calls the function getu/getl fron csparse.c
#		- It frees the memory
#
============================================================================*/

void sparse::getu ()

{
  // Store the input matrix
	sparse *input = new sparse (*this);

  // Free the memory of the old matrix
	free();

  // Allocate the space for the new matrix
	nrow=input->get_nrow();
	ncol=input->get_ncol();
	nnz=input->get_nnz();
	a = new double[nnz];
	ja=new int[nnz];
	ia=new int[nrow+1];

  // Call
	getu_matrix(nrow, input->get_a(), input->get_ja(), input->get_ia(), a, ja, ia);	 

	nnz=ia[nrow]-1;
	input->free();
}

void sparse::getl ()

{
  // Store the input matrix
	sparse *input = new sparse (*this);

  // Free the memory of the old matrix
	free();

  // Allocate the space for the new matrix
	nrow=input->get_nrow();
	ncol=input->get_ncol();
	nnz=input->get_nnz();
	a = new double[nnz];
	ja=new int[nnz];
	ia=new int[nrow+1];


  // Call
	getl_matrix(nrow, input->get_a(), input->get_ja(), input->get_ia(), a, ja, ia);	 

	nnz=ia[nrow]-1;
	input->free();
}

/*============================================================================
#				  GETBWD		
#				  ------    	
# in entry  : 	nothing 
#
# in return : 	int * or int
#
# details   : 	It returns the bandwith of the upper and of the lower part 
#		of the matrix
#		- getbwdu returns the bandwith of the upper part
#		- getbwdl returns the bandwith of the lower part
#		- getbwd returns an array containing 2 elements: the first
#		  one is the bandwith of the lower part and the 2nd is the
#		  bandwith of the upper part
#
============================================================================*/


int * sparse::getbwd()
	{
	int ml[1], mu[1], *res = new int[2];
	getbwd_matrix(nrow, a, ja, ia, ml, mu);
	res[0]=*ml; res[1]=*mu;
	return res;
	
	}
int sparse::getbwdu()
	{
	int ml[1], mu[1];
	getbwd_matrix(nrow, a, ja, ia, ml, mu);
	return *mu;
	
	}

int sparse::getbwdl()
	{
	int ml[1], mu[1];
	getbwd_matrix(nrow, a, ja, ia, ml, mu);
	return *ml;
	
	}

int sparse::blkfnd()
	{
	int blk[1];
	blk[0] = 4;
	blkfnd_matrix(nrow, ja, ia, blk);
	return *blk;
	
	}

/*============================================================================
#				CNRMS :		
#				------    	
# in entry  :   nrm -> norm indicator 
#
# in return :	a double array (norms of each columns)
#
# details   :	It computes the norms of the columns of a matrix
#		- It allocates enough space for the result (double array)
#		- It calls the function cnrms_matrix from csparse.c and 
#		  returns the result
#
============================================================================*/

double * sparse::cnrms(int nrm)
{
  // Allocate enough space
	double *diag;
	diag = new double[nrow];

  // Call the function
	cnrms_matrix(nrow, nrm, a, ja, ia, diag);
	
  // Free the memory


  // Return the result
	return diag;

}

/*============================================================================
#			    COSCAL/ROSCAL :		
#			    -------------    	
# in entry  :   nrm    -> norm indicator 
#
# in return :	nothing
#
# details   :	It scales the rows (or the columns) of a matrix by their norms
#		- It allocates enough space for the double array containing the
#		  norms of the columns (or of the rows)
#		- It calls coscal_matrix (or roscal_matrix) from csparse.
#
============================================================================*/

void sparse::coscal(int nrm)

{
  // Allocate enough space
	double *diag;
	diag = new double[nrow];

  // Call the function
	coscal_matrix(nrow, 1, nrm, a, ja, ia, diag, a, ja, ia, 0);

  // Free the memory
	delete [] diag;
}

void sparse::roscal(int nrm)

{
  // Allocate enough space
	double *diag;
	diag = new double[nrow];

  // Call the function
	roscal_matrix(nrow, 1, nrm, a, ja, ia, diag, a, ja, ia, 0);

  // Free the memory
	delete [] diag;
}

/*============================  ADDBLK ===============================*/

void sparse::addblk(sparse mat, int ipos, int jpos)

{
  // Store the old matrix
	sparse *input = new sparse(*this);

  // Allocate enough space for the arguments
	int nrowc[1];
	int ncolc[1];

	

  // Allocate enough space for the result
	free();
	nrow = input->get_nrow();
	ncol = input->get_ncol();
	*nrowc=nrow; *ncolc = ncol;
	nnz = input->get_nnz() + mat.get_nrow() * mat.get_ncol();
	a = new double[nnz];
	ia = new int[nrow+1];
	ja = new int[nnz];
	
  // Call the function
	addblk_matrices(mat.get_nrow(), mat.get_ncol(), mat.get_a(), mat.get_ja(), mat.get_ia(), ipos, jpos, input->get_nrow(), input->get_ncol() , input->get_a(), input->get_ja(), input->get_ia(), nrowc, ncolc, a, ja, ia, nnz, 0);

  // Corrects the dimension of the result
	nrow = *nrowc;
	ncol = *ncolc;
	nnz = ia[nrow] - 1;

  // Free the memory
	input->free();

}

/*=================== GET1UP ======================*/

int * sparse::get1up()
{
 	int * ju = new int[nrow];

	get1up_matrix(nrow, ja, ia, ju);

	return ju;
}

/*=================== INFDIA ======================*/

int * sparse::infdia()
{ //returns a segmentation fault!!
 	int * idiag = new int[4*nrow];
 	int * ind = new int[4*nrow];
	
	infdia_matrix(nrow, ja, ia, ind, idiag);

	delete [] idiag;
	return ind;
}
/*============================================================================
#				READMT		
#				------    	
# in entry  : 	filename : a char array
#
# in return :	nothing
#
# details   :	- it checks to make sure the file exists
#		- it free the memory of the matrix
#		- it gets the dimension of the matrix stored in the file:get_dim
#		- it allocates space for the new matrix
#		- it reads the matrix from the file by calling read_matrix, 
#		  a function of the class csparse.c
#
============================================================================*/

void sparse::readmt (char filename[]) {
  // Read the matrix from the file filename.
  #ifdef debugging
  struct stat statbuff;   // Buffer to hold the status of the file.
  assert (stat (filename, &statbuff) == 0);  //  Make sure the file exists.
  #endif

  // Free any memory in the sparse matrix.
  free();

  // Get the dimensions of the matrix to determine how much
  // space to allocate.
  get_dim (filename, &nrow, &ncol, &nnz);

  // Allocate space for the new matrix.
  a = new double[nnz];
  ja = new int [nnz];
  ia = new int [nrow+1];

  read_matrix (filename, nrow, ncol, nnz, a, ja, ia,
	       title, key, type);

} // end readmt

/*============================================================================
#				PRTMT :		
#				------    	
# in entry  :   a string: filename 
#
# in return :	nothing
#
# details   :	It stores the matrix mat into the Harwell-Boeing file 
#		- checks to make sure that the file doesn't exist
#		- calls the function write_matrix of csparse.c
#
============================================================================*/

void sparse::prtmt (const char *filename) {
  // Write the matrix to the  file filename.
  #ifdef debugging

  struct stat statbuff;   // Buffer to hold the status of the file.

    assert (stat (filename, &statbuff) != 0);  //  Make sure the file does not
                                               //  already exist.
  #endif                                          

  write_matrix (strdup(filename), nrow, ncol, nnz, a, ja, ia, title, key,type, 4);
  
} // end prtmt

/*============================================================================
#				INIT_ZERO :		
#				----------   	
# in entry  :   2 integers :i, j (the dimensions of the matrix)
#
# in return :	nothing
#
# details   :	It initializes a matrix to all zeroes
#		- it frees the old matrix
#		- it creates a new matrix with i rows and j columns
#
# Additions by Ben:
#
# Changes   :   No longer frees the old matrix!  Clearly this is not
#               necessary since this method is only ever called from
#               constructors, and is dangerous becuase it will often
#               delete [] arrays that have not yet been initialized.
#
============================================================================*/

void sparse::init_zero (int i, int j) {
  // Initialize a sparse matrice to all zeroes.

  // free();

  nrow = i;
  ncol = j;
  nnz = 0;
  ia = new int[nrow+1];
  ja = new int[1];
  a = new double[1];

  for (i = 0; i <= nrow; i += 1) 
    ia[i] = 1;

  return;

} // end init_zero

/*============================================================================
#			INIT_IDENTITY :		
#			--------------   	
# in entry  :   integer :i (the dimension of the matrix)
#
# in return :	nothing
#
# details   :	It initializes a matrix to an identity matrix
#		- it frees the old matrix
#		- it creates a new matrix with i rows and i columns
#		- it stores i 1 into the array ia
#		- ja[i] = i+1 and ia[i] = i+1
#
============================================================================*/

void sparse::init_identity (int i) {
  // Initialize a sparse matrix to an identity matrix.

  free();
  nrow = i; 
  ncol = i;
  nnz = i;

  ia = new int[nrow+1];
  ja = new int[nnz];
  a = new double[nnz];
  
  for (i = 0; i <= nnz -1; i += 1) {
    a[i] = 1.0;
    ja[i] = i+1;
    ia[i] = i+1;
  }
  ia [nrow] = ia[0] + nnz;

} // end init_identity

//************************************************************************
//   Member Operators
//************************************************************************

/*============================================================================
#				OPERATOR =		
#				----------    	
# in entry  : x -> a sparse matrix 
#
# in return : nothing
#
# details   : copies the value of x into the sparse matrix.
#		- free the memory from the old matrix 
#		- create a new matrix with x's dimensions
#		- copy all the elements of ia, a and ja
#
============================================================================*/

void sparse::operator = (const sparse & x) {
  // Simply  copy the value of x.

  int *xia;         // Pointer to the array ia in x 
  int *xja;         // Pointer to the array ja in x 
  double *xa;       // Pointer to array a in x.
  int i;            // Simple loop counter

  // Free memory from left matrix (if any).
  free();

  // Allocate new memory for the matrix.
  a = new double [x.get_nnz()];
  ja = new int [x.get_nnz()];
  ia = new int [x.get_nrow() +1];

  xia = x.get_ia();
  xja = x.get_ja();
  xa = x.get_a();

  nrow = x.get_nrow();
  ncol = x.get_ncol();
  nnz = x.get_nnz();

  for (i = 0; i <= nrow; i++) 
    ia[i] = xia[i];

  for (i = 0; i <= nnz-1; i++) {
    ja[i] = xja[i];
    a[i] = xa[i];
  } // end for


} // end operator =
  
/*============================================================================
#				OPERATOR <<=		
#				------------    	
# in entry  : x -> a sparse matrix 
#
# in return : nothing
#
# details   : copies the value of the pointers to a, ia and ja of x into the
# 	      pointers a, ia and ja of the sparse matrix
#
============================================================================*/

void sparse::operator <<= (sparse & x) {
  // Simply  copy the value of x.  Does not actually
  // copy the contents of the arrays just the pointers.
  // Be carefull about aliasing.

  // Free memory from left matrix (if any).
  free();

  ia = x.get_ia();
  ja = x.get_ja();
  a = x.get_a();

  nrow = x.get_nrow();
  ncol = x.get_ncol();
  nnz = x.get_nnz();

} // end operator <<=
  
/*============================================================================
#				PRECONDITIONNERS		
#				----------------    	
# functions: ilut, ilud, ilu0, iluk, lusol, lutsol	
#
# details   : It takes a sparse matrix in entry (and other arguments, see
#	      /project/darpa/SPARSKIT2/itsol/ilut.f) and returns another 
#	      sparse matrix in modified row format
#
============================================================================*/

int sparse::ilut(int lfil, double drptol)  {

  // Allocate enough memory
	
	double *a0 = new double[nrow+1+nrow*lfil];
	int *ja0 = new int[nrow + 1 + nrow*lfil];
	int *ia0 = new int[nrow+1];	

	int iwk = nrow+1+nrow*lfil;
	int *jw = new int[2*nrow];
	int ierr[1]; ierr[0]=0;
	double *w = new double[nrow];

  // Call the function
   ilut_prec(nrow, a, ja, ia, lfil, drptol, a0, ja0, ia0, iwk, w, jw, ierr);

 // Free the memory of the work arrays 
	delete [] jw;
	delete [] w;

  // Convert the result in CSR
	this->msrcsr(a0, ja0, nrow);
	nnz = ia[nrow] - 1;
	
  // Free the memory of the msr
	delete [] a0;
	delete [] ia0;
	delete [] ja0;

  // Result:
	return *ierr;
}

int sparse::ilut_msr(int lfil, double drptol)  {

  // Allocate enough memory
	sparse *input = new sparse(*this);
	free();

	nrow=input->get_nrow();	
	ncol=input->get_ncol();	
	nnz=input->get_nnz();

	a = new double[nrow+1+2*nrow*lfil];
	ja = new int[nrow + 1 + 2*nrow*lfil];
	ia = new int[nrow+1];	

	int iwk = nrow+1+2*nrow*lfil;
	int *jw = new int[2*nrow];
	int ierr[1]; ierr[0]=0;
	double *w = new double[nrow+1];

	// Call the function
	// a -> alu
	// ja -> jlu
	// ia -> ju

	//	cout << "CALLING ILUT_PREC: " << endl
	//	     << "lfil is " << lfil << endl
	//	     << "drptol " << drptol << endl;

	ilut_prec(nrow, input->get_a(), input->get_ja(), input->get_ia(), lfil, drptol, a, ja, ia, iwk, w, jw, ierr);

 // Free the memory of the work arrays 
	delete [] jw;
	delete [] w;
	input->free();

  // Result:
	return *ierr;
}

int sparse::ilutp_msr(int lfil, double drptol, double permtol)  {

  // Allocate enough memory
	sparse *input = new sparse(*this);
	free();

	nrow=input->get_nrow();	
	ncol=input->get_ncol();	
	nnz=input->get_nnz();

	a = new double[nrow+1+2*nrow*lfil];
	ja = new int[nrow + 1 + 2*nrow*lfil];
	ia = new int[nrow+1];	

	int iwk = nrow+1+2*nrow*lfil;
	int *jw = new int[2*nrow];
	int ierr[1]; ierr[0]=0;
	double *w = new double[nrow+1];
	int *iperm = new int[2*nrow];

	// Call the function
	// a -> alu
	// ja -> jlu
	// ia -> ju

	ilutp_prec(nrow, input->get_a(), input->get_ja(), input->get_ia(), lfil, drptol, permtol, nrow, a, ja, ia, iwk, w, jw, iperm, ierr);

 // Free the memory of the work arrays 
	delete [] jw;
	delete [] w;
	delete [] iperm;
	input->free();

  // Result:
	return *ierr;
}

int sparse::iluk_msr(int lfil)
{
// Allocate enough memory
	sparse *input = new sparse(*this);
	free();
	int iwk = lfil*lfil*input->get_nnz();
	nrow=input->get_nrow();	
	ncol=input->get_ncol();	
	nnz=nrow*nrow +nrow +1;
	a = new double[nrow*nrow +nrow +1];
	ja = new int[nrow*nrow +nrow +1];
	ia = new int[nrow+1];	

	int ierr[1]; ierr[0]=0;
	int *levs = new int[iwk];
	double *w = new double[nrow];
	int *jw = new int[3*nrow];

  // Call the function
	iluk_prec(nrow, input->get_a(), input->get_ja(), input->get_ia(), lfil, a, ja, ia, levs, iwk, w, jw, ierr);

 // Free the memory of the work arrays 
	delete [] jw;
	delete [] w;
	delete [] levs;
	input->free();

  // Result:
	return *ierr;
}

int sparse::ilu0_msr()
{
// Allocate enough memory
	sparse *input = new sparse(*this);
	free();
	nrow=input->get_nrow();	
	ncol=input->get_ncol();	
	nnz=input->get_nnz();
	a = new double[nrow+1+nnz];
	ja = new int[nrow + 1+nnz];
	ia = new int[nrow + 1];	

	int ierr[1]; ierr[0]=0;
	int *iw = new int[nrow];

  // Call the function
	ilu0_prec(nrow, input->get_a(), input->get_ja(), input->get_ia(), a, ja, ia, iw, ierr);

 // Free the memory of the work arrays 
	delete [] iw;
	input->free();

  // Result:
	return *ierr;
}

int sparse::ilud(double alph, double tol)  {

  // Allocate enough memory
	
	double *a0 = new double[nrow+1+nrow*nrow];
	int *ja0 = new int[nrow + 1 + nrow*nrow];
	int *ia0 = new int[nrow+1];	

	int iwk = nrow+1+nrow*nrow;
	int *jw = new int[2*nrow];
	int ierr[1]; ierr[0]=0;
	double *w = new double[nrow];

  // Call the function
  
  ilut_prec(nrow, a, ja, ia, (int)alph, tol, a0, ja0, ia0, iwk, w, jw, ierr);

 // Free the memory of the work arrays 
	delete [] jw;
	delete [] w;

// Free the memory of the msr
	delete [] a0;
	delete [] ia0;
	delete [] ja0;

  // Convert the result in CSR
	this->msrcsr(a0, ja0, nrow);
	nnz = ia[nrow] - 1;
	
  // Result:
	return *ierr;

}

int sparse::ilud_msr(double alph, double tol)  {

  // Allocate enough memory
	sparse *input = new sparse(*this);
	free();
	nrow=input->get_nrow();	
	ncol=input->get_ncol();	
	nnz=input->get_nnz();
	a = new double[nrow+1+nrow*nrow];
	ja = new int[nrow + 1 + nrow*nrow];
	ia = new int[nrow+1];	

	int iwk = nrow+1+nrow*nrow;
	int *jw = new int[2*nrow];
	int ierr[1]; ierr[0]=0;
	double *w = new double[nrow];

  // Call the function
     ilud_prec(nrow, input->get_a(), input->get_ja(), input->get_ia(), 
	       alph, tol, a, ja, ia, iwk, w, jw, ierr);

 // Free the memory of the work arrays 
	delete [] jw;
	delete [] w;
	input->free();

  // Result:
	return *ierr;

}

void sparse::lusol(int n, double *y, double *x)
{
	lusol_prec(n, y, x, a, ja, ia);
}
void sparse::lutsol(int n, double *y, double *x)
{
	lutsol_prec(n, y, x, a, ja, ia);
}

/*============================================================================
#				CHANGE FORMAT		
#				----------------    	
# functions: msrcsr, csrmsr	
#
# details   : It takes a sparse matrix stored in compressed row format and
#	      returns it in modified row format (or the contrary)
#
============================================================================*/

void sparse::msrcsr(double *a0, int *ja0, int n)

{
  // Allocate the memory of the matrix
	free();
	nrow = n;
	ncol = n;
	nnz = n*n;
	a = new double[nnz];
	ja = new int[nnz];
	ia = new int[nrow+1];

  // Allocate the memory for the work arrays
	double *wk = new double[n];
	int *iwk = new int[n+1];

  // Call the fortran function
	msrcsr_format(n, a0, ja0, a, ja, ia, wk, iwk);

	nnz=ia[n]-1;
  // Free the memory of the work arrays
	delete [] wk;
	delete [] iwk;
}

void sparse::csrmsr( )

{	sparse *input = new sparse(*this);

  // Allocate the memory of the matrix
	free();
	nrow = input->get_nrow();
	ncol = input->get_ncol();
	nnz = nrow*ncol;
	a = new double [nrow + 1 + nnz];
	ja = new int[nrow+1+nnz];
	ia = new int[nrow+1];

  // Allocate the memory for the work arrays
	double *wk = new double[nrow];
	int *iwk = new int[nrow+1];

  // Call the fortran function
	csrmsr_format(nrow, input->get_a(), input->get_ja(), input->get_ia(), a, ja, wk, iwk);

  // Free the memory of the work arrays
	delete [] wk;
	delete [] iwk;
	input->free();
	
	int i;
	for (i=0;i<=nrow;i++)
		{ ia[i] = ja[i];  }
}

/*============================================================================
#				MULT MAT/VECT		
#				-------------    	
# functions:  amux, atmux	
#
# details   : It multiplies a sparse matrix with a vector (amux) A*x = y
#	      or the transpose of a matrix with a vector (atmux) tA*x =y
#	      
============================================================================*/

void sparse::amux(double *x, double *y)
{
	amux_matrix(nrow, x,y,a,ja,ia);
	
}

void sparse::atmux(double *x, double *y)
{
	atmux_matrix(nrow, x,y,a,ja,ia);
	
}

/*============================================================================
#				SOLVERS		
#				-------------    	
# functions:   	runriters, solve_system	
#
# details   :  	runriters is an old function (shouldn't be used) which
#		corresponds to the fortran one riters (in itsol/riters.f)
#		The most important one is solve_system.
#	      
============================================================================*/

double * sparse::runriters(double *rhs, double *guessvec, char *solver_name) {

  // Allocate all the memory ->
  int *ipar = new int[16];
  double *fpar = new double[16];
  double *sol = new double[nrow];
  string s(solver_name);
  double *wk = NULL;  // Given NULL value to ensure that compilers
  // do not give warnings about wk not being used
	
	
  // Initialize
	// int nx=6, ny=6, nz=1;
	// double alpha = 0.0; double gammax = 0.0; double gammay = 0.0;
	
  ipar[1] = 2;
  ipar[2] = 1;
  ipar[3] = 40*5000;
  ipar[4] = 10;
  ipar[5] = 60;
  fpar[0] = 0.00001;
  fpar[1] = 0.0000000001;

  // int lfil = 3;
  // double tol = 0.0001;
  sparse *b = new sparse(*this);
  int ierr;
  ierr = b->ilut_msr(3, 0.0001);
  if (ierr!=0) cout<<ierr<<"erreur dans ilut \n";
	
  if (s == *(new string("bcg"))) {	
    wk = new double[5000];
    rundbcg_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, a, ja, ia, 
		   b->get_a(), b->get_ja(), b->get_ia() );
  }

  if (s == *(new string("cg"))) {	
    wk = new double[5000];
    runcg_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, a, ja, ia, 
		 b->get_a(), b->get_ja(), b->get_ia() );
  }

  if (s == *(new string("dbcg"))) {	
    wk = new double[5000];
    rundbcg_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, a, ja, ia, 
		   b->get_a(), b->get_ja(), b->get_ia() );
  }

  if (s == *(new string("bcgstab"))) {	
    wk = new double[5000];
    runbcgstab_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, a, ja, ia, 
		      b->get_a(), b->get_ja(), b->get_ia() );
  }

  if (s == *(new string("cgnr"))) {	
    wk = new double[5000];
    runcgnr_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, a, ja, ia, 
		   b->get_a(), b->get_ja(), b->get_ia() );
  }

  if (s == *(new string("fom"))) {	
    wk = new double[5000];
    runfom_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, a, ja, ia, 
		  b->get_a(), b->get_ja(), b->get_ia() );
  }

  if (s == *(new string("tfqmr"))) {	
    wk = new double[5000];
    runtfqmr_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, a, ja, ia, 
		    b->get_a(), b->get_ja(), b->get_ia() );
  }

  if (s == *(new string("gmres"))) {	
    wk = new double[5000];
    rungmres_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, a, ja, ia, 
		    b->get_a(), b->get_ja(), b->get_ia() );
  }

  if (s == *(new string("fgmres"))) {	
    wk = new double[5000];
    runfgmres_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, a, ja, ia, 
		     b->get_a(), b->get_ja(), b->get_ia() );
  }

  if (s == *(new string("dqgmres"))) {	
    wk = new double[5000];
    rundqgmres_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, a, ja, ia, 
		      b->get_a(), b->get_ja(), b->get_ia() );
  }


  // Free the memory
  delete [] ipar;
  delete [] fpar;
  delete [] wk;
  delete [] b->get_a();
  delete [] b->get_ja();
  delete [] b->get_ia();

  return sol;
	
}

double * sparse::solve_system(double *rhs, double *guessvec, char *solver_name,
			      char *prec_name, double *arg_prec, int prec, 
			      int krylov, int maxmult, double tolrel,
			      double tolabs) {

  // Allocate all the memory ->
	int *ipar = new int[16];
	double *fpar = new double[16];
	double *sol = new double[nrow];
	string s(solver_name);
	string p(prec_name);
	double *wk = NULL;
	
	
  // Initialize
	// int nx=6, ny=6, nz=1;
	// double alpha = 0.0; double gammax = 0.0; double gammay = 0.0;
	
	ipar[1] = prec;
	ipar[2] = 0;
	ipar[3] = 40*5000;
	ipar[4] = krylov;
	ipar[5] = maxmult;
	fpar[0] = tolrel;
	fpar[1] = tolabs;

	/*
	cout << "ipar[1]: " << ipar[1] << endl;
	cout << "ipar[2]: " << ipar[2] << endl;
	cout << "ipar[3]: " << ipar[3] << endl;
	cout << "ipar[4]: " << ipar[4] << endl;
	cout << "ipar[5]: " << ipar[5] << endl;
	cout << "fpar[0]: " << fpar[0] << endl;
	cout << "fpar[1]: " << fpar[1] << endl;
	*/

	sparse *b = new sparse(*this);
	int ierr = 0;

/*------------------------  preconditionners  ----------------------------*/

	// WHEN ADDING A PRECONDITIONER:
	//
	// Add a new "else if" to this part that checks for
	// the name of the new preconditioner.  Then, store
	// each argument in arg_prec[] in a variable of the
	// appropriate type.  Finally, call the method
	// <prec>_msr where <prec> is the name of the
	// preconditionner.  The <prec>_msr method must be
	// written to set up the scratch space and call the
	// preconditionner.  See ilut_msr for example

	if (p == *(new string("ilut"))) {

	  //	  cout << "Calling ilut" << endl;	
	  int lfil = (int)arg_prec[0];
	  double drptol = arg_prec[1];

	  //  cout << "lfil is " << lfil << " drptol is " << drptol << endl;

	  ierr = b->ilut_msr(lfil, drptol);
	}
	else if (p == *(new string("ilutp"))) {

	  int lfil = (int)arg_prec[0];
	  double drptol = arg_prec[1];
	  double permtol = arg_prec[2];

	  ierr = b->ilutp_msr(lfil, drptol, permtol);
	}
	else if (p == *(new string("iluk"))) {

	  int lfil = (int)arg_prec[0];
	  ierr = (int)b->iluk_msr(lfil);
	}
	else if (p == *(new string("ilu0"))) {

	  ierr = b->ilu0_msr();
	}
	else if (p == *(new string("ilud"))) {

	  double alph = arg_prec[0];
	  double tol  = arg_prec[1];
	  ierr = b->ilud_msr(alph, tol);
	}

	if (ierr != 0) {

	  sp_error("Error in preconditioner: ");
	  sp_error(ierr);
	  sp_error("\n");

	  delete[] ipar;
	  delete[] fpar;
	  delete[] sol;

	  delete b;

	  return NULL;
	}

/*------------------------  solvers  ------------------------------------*/	

	if (s == *(new string("bcg"))) {	
	  wk = new double[7*nrow];
	  rundbcg_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, a, ja, ia, 
			 b->get_a(), b->get_ja(), b->get_ia());
	}

	if (s == *(new string("cg"))) {	
	  wk = new double[5*nrow];
	  runcg_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, a, ja, ia, 
		       b->get_a(), b->get_ja(), b->get_ia() );
	}

	if (s == *(new string("dbcg"))) {	
	  wk = new double[11*nrow];
	  rundbcg_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, a, ja, ia, 
			 b->get_a(), b->get_ja(), b->get_ia() );
	}

	if (s == *(new string("bcgstab"))) {	
	  wk = new double[8*nrow];
	  runbcgstab_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, a, ja, ia,
			    b->get_a(), b->get_ja(), b->get_ia() );
	}

	if (s == *(new string("cgnr"))) {	
	  wk = new double[5*nrow];
	  runcgnr_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, a, ja, ia,
			 b->get_a(), b->get_ja(), b->get_ia() );
	}

	if (s == *(new string("fom"))) {
	  int m = ipar[4];
	  wk = new double[(nrow+3)*(m+2)+(m+1)*m/2];
	  runfom_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, a, ja, ia, 
			b->get_a(), b->get_ja(), b->get_ia() );
	}

	if (s == *(new string("tfqmr"))) {
	  wk = new double[11*nrow];
	  runtfqmr_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, a, ja, ia, 
			  b->get_a(), b->get_ja(), b->get_ia() );
	}

	if (s == *(new string("gmres"))) {
	  //	  cout << "About to call gmres with nrow " << nrow << endl;
	  //	  cout << "rhs[0] " << rhs[0] << endl;
	  //	  cout << "guessvec[0] " << guessvec[0] << endl;
	  //	  cout << "a[0] " << a[0] << endl;
	  //	  cout << "ja[0] " << ja[0] << endl;
	  //	  cout << "ia[0] " << ia[0] << endl;
	  int m = ipar[4];
	  //	  cout << "m " << m << endl;
	  wk = new double[(nrow+3)*(m+2)+(m+1)*m/2];
	  rungmres_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, a, ja, ia, 
			  b->get_a(), b->get_ja(), b->get_ia() );
	}

	if (s == *(new string("fgmres"))) {	
	  int m = ipar[4];
	  wk = new double[2*nrow*(m+1)+(m+1)*m/2+3*m+2];
	  runfgmres_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, a, ja, ia, 
			   b->get_a(), b->get_ja(), b->get_ia() );
	}

	if (s == *(new string("dqgmres"))) {	
	  /*int lb = ipar[4] + 1;
	    wk = new double[nrow+lb*(2*nrow + 4)];*/ 
	  wk = new double[5000];
	  rundqgmres_solver(nrow, rhs, sol, ipar, fpar, wk, guessvec, 
			    a, ja, ia, b->get_a(), b->get_ja(), 
			    b->get_ia() );
	}


  // Free the memory
	delete [] ipar;
	delete [] fpar;
	delete [] wk;
	delete [] b->get_a();
	delete [] b->get_ja();
	delete [] b->get_ia();

	return sol;
	
}

/* ======================== MULTIC / COLORS============================== */

void sparse::multic ()
{
  // Allocate memory
	int ncolor[1];
	int maxcol=0;
	int *kolrs = new int[nrow];
	int *iord = new int[nrow];
	int ierr[1];
	ierr[0]=0;
	int i;
	for (i=0;i<nrow;i++) iord[i]=i+1;

  // Calculates maxcol (max of non zero elements in each row)
	for (i = 1; i<= nrow; i++) 		
	{ if (ia[i]-ia[i-1]>maxcol) maxcol = ia[i]-ia[i-1];		}
	int *il = new int[maxcol+1];

  // Call the fortran function
	multic_matrix (nrow, ja, ia, ncolor, kolrs, il, iord, maxcol, ierr);

  // Prints the other results
	cout<<"\n"<<*ncolor<<" colors found - size of each color set : \n";
	int j;
	for (j=1;j<*ncolor+1;j++) cout<<"color "<<j<<" -> "<<il[j]-il[0]<<"\n";
	cout<<"\n";

  // Free the memory
	delete [] kolrs;
	delete [] il;

 // Call rperm with iord
	sparse *temp = new sparse(*this);
	free();

	*temp = cperm(*temp, iord);
	*this = rperm(*temp, iord);

	delete [] iord;
	temp->free();
}

int *  sparse::colors ()
{
  // Allocate memory
	int ncolor[1];
	int maxcol=0;
	int *kolrs = new int[nrow+1];
	int *iord = new int[nrow];
	int ierr[1];
	ierr[0]=0;
	
	int i;
	for (i=0;i<nrow;i++) iord[i]=i+1;

	for (i = 1; i<= nrow; i++) 		
	{ if (ia[i]-ia[i-1]>maxcol) maxcol = ia[i]-ia[i-1];		}

	int *il = new int[maxcol+1];

  // Call the fortran function
	multic_matrix (nrow, ja, ia, ncolor, kolrs, il, iord, maxcol, ierr);

  // add ncolor at the end of kolrs
	kolrs[nrow] = *ncolor;

  // Free the memory
	delete [] il;
	delete [] iord;

	return kolrs;
}

int *  sparse::il_multic ()
{
  // Allocate memory
	int ncolor[1];
	int maxcol=0;
	int *kolrs = new int[nrow];
	int *iord = new int[nrow];
	int ierr[1];
	ierr[0]=0;
	
	int i;
	for (i=0;i<nrow;i++) iord[i]=i+1;

	for (i = 1; i<= nrow; i++) 		
	{ if (ia[i]-ia[i-1]>maxcol) maxcol = ia[i]-ia[i-1];		}

	int *il = new int[maxcol+2];

  // Call the fortran function
	multic_matrix (nrow, ja, ia, ncolor, kolrs, il+1, iord, maxcol, ierr);

  // Add the number of colors at the beginning of il
	il[0] = *ncolor;

  // Free the memory
	delete [] iord;
	delete [] kolrs;

	return il;
}

int *  sparse::iord_multic ()
{
  // Allocate memory
	int ncolor[1];
	int maxcol=0;
	int *kolrs = new int[nrow];
	int *iord = new int[nrow];
	int ierr[1];
	ierr[0]=0;
	
	int i;
	for (i=0;i<nrow;i++) iord[i]=i+1;

	for (i = 1; i<= nrow; i++) 		
	{ if (ia[i]-ia[i-1]>maxcol) maxcol = ia[i]-ia[i-1];		}

	int *il = new int[maxcol+1];

  // Call the fortran function
	multic_matrix (nrow, ja, ia, ncolor, kolrs, il+1, iord, maxcol, ierr);



  // Free the memory
	delete [] il;
	delete [] kolrs;

	return iord;
}
/* ============================= INDSET ============================== */

void sparse::indset0()
{
int *nset = new int[1]; nset[0]=3;
int *iord = new int[nrow];
int *riord = new int[nrow];
int sym=1;
int iptr = 1;

indset0_matrix(nrow, ja, ia, nset, iord, riord, sym, iptr);
int j;
cout<<"IORD: ";
for (j=0;j<nrow;j++)
cout<<iord[j]<<"  ";
cout<<"\n";
sparse *temp = new sparse(*this);
free();
*this = rperm(*temp, iord);
temp->free();
delete [] nset;
delete [] iord;
delete [] riord;


}

void sparse::indset1()
{
int *nset = new int[1]; nset[0]=3;
int *iord = new int[nrow];
int *riord = new int[nrow];
int *iw = new int[nrow];
int sym=1;
int iptr = 1;

indset1_matrix(nrow, ja, ia, nset, iord, riord, iw, sym, iptr);
int j;
cout<<"IORD: ";
for (j=0;j<nrow;j++)
cout<<iord[j]<<"  ";
cout<<"\n";
sparse *temp = new sparse(*this);
free();
*this = rperm(*temp, iord);
temp->free();

delete [] nset;
delete [] iord;
delete [] riord;
delete [] iw;


}

int sparse::size() {
  return (sizeof(int)*nnz +        // ja
	  sizeof(double)*nnz +     // a
	  sizeof(int *)*(nrow+1) + // ia
	  sizeof(*this));
}

//
// sparse::get_stats(sparse_stats &)
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
//   Calls various fortran functions in order to fill the 
// sparse_stats structure with all the appropriate stats for
// this matrix.  Obtains all the same information as info1.ex
// in SPARSKIT.  See sparse.h for definition of sparse_stats
// structure.  See infofun.f in SPARSKIT's INFO subdirectory
// for the definitions of the subroutines called here.
//
//   This method is meant to be completely analogus to the
// FORTRAN subroutine dinfo1 written by Y. Saad.  Some 
// comments are taken directly from that code.
//
// Written by Ben Langmead, July 5 2001
//

void sparse::get_stats(sparse_stats &stats) {

  int i, j, k, klo, khi;

  int ndiag = nrow+nrow-1;
  int jdim = (nnz > (ndiag+2)) ? nnz : (ndiag+2);

  int *iao = new int[nrow + 1];
  int *jao = new int[jdim];
  double *ao = new double[nnz];

  stats.Dimension = ncol;
  stats.NNz = nnz;

  //  cout << "About to do max_element" << endl;
  stats.Maximum_Element = max_element();

  //  cout << "About to do csrcsc" << endl;
  csrcsc(ncol, 1, 1, a, ja, ia, ao, jao, iao);
  csrcsc(ncol, 1, 1, ao, jao, iao, a, ja, ia);

  // computing max bandwith, max number of nonzero elements per column
  // min nonzero elements per column/row, row/column diagonal dominance 
  // occurences, average distance of an element from diagonal, number of 
  // elemnts in lower and upper parts, ...

  //  cout << "About to do skyline" << endl;
  skyline(ncol, 0, ja, ia, jao, iao,
	  &stats.NNZ_In_Skyline_Storage);

  //  cout << "About to do nonz_lud" << endl;
  nonz_lud(&ncol, ja, ia,
	      &stats.NNZ_In_Strict_Lower_Part,
	      &stats.NNZ_In_Strict_Upper_Part,
	      &stats.NNZ_In_Main_Diagonal);

  //  cout << "About to do avnz_col" << endl;
  avnz_col(&ncol, ja, ia, iao, &stats.NNZ_In_Main_Diagonal, 
	      &stats.Average_NNZ_Per_Column, 
	      &stats.Averate_NNZ_Std_Deviation);

  //  cout << "About to do nonz" << endl;
  nonz(ncol, 0, ja, ia, iao,
       &stats.Weight_Of_Longest_Column,
       &stats.Weight_Of_Shortest_Column,
       &stats.Weight_Of_Longest_Row,
       &stats.Weight_Of_Shortest_Row,
       &stats.N_Zero_Columns,
       &stats.N_Zero_Rows);

  //  cout << "About to do diag_domi" << endl;
  int sym = 0;
  int val = 1;
  double da, db;
  diag_domi(&ncol, &sym, &val, a, ja, ia, ao, jao, iao,
	      &da, &db);

  //  cout << "da is " << da << " and db is " << db << endl;

  stats.Percentage_Diag_Dominant_Columns = da;
  stats.Percentage_Diag_Dominant_Rows = db;

  //
  // symmetry and near symmetry - Frobenius  norms
  //

  //  cout << "About to do frobnorm" << endl;
  frobnorm(ncol, 0, a, ja, ia,
	   &stats.Frobenius_Norm);

  //  cout << "About to do ansym" << endl;
  ansym(ncol, 0, a, ja, ia, ao, jao, iao, 
	&stats.Matching_Elements_In_Symmetry,
	&stats.Relative_Symmetry_Match,
	&stats.Frobenius_Norm_Of_Symmetric,
	&stats.Frobenius_Norm_Of_Nonsymmetric);

  //  cout << "About to do distaij" << endl;
  distaij(ncol, nnz, 0, ja, ia, 
	  &stats.Average_Distance_From_Diag,
	  &stats.Average_Distance_Std_Deviation);

  //
  //  bandedness - main diagonals
  //

  for(i = 0; i < ndiag; i++) {
    jao[i] = 0;
  }
  for(i = 0; i < ncol; i++) {
    klo = ia[i]-1;
    khi = ia[i+1]-1;

    for(k = klo; k < khi; k++) {
      j = ja[k]-1;
      jao[ncol+i-j]++;
    }
  }

  //  cout << "About to do bandwidth" << endl;
  bandwidth(ncol, ja, ia, 
	    &stats.Lower_Bandwidth,
	    &stats.Upper_Bandwidth,
	    &stats.Maximum_Bandwidth,
	    &stats.Average_Bandwidth);

  //  cout << "About to do distdiag" << endl;
  distdiag(nrow, ncol, ja, ia, jao);

  //  cout << "About to calc bandpart 80" << endl;
  bandpart(ncol, ja, ia, jao, 80,
	   &stats.Eighty_Percent_Bandwidth);

  stats.Eighty_Percent_Bandwidth = stats.Eighty_Percent_Bandwidth*2 + 1;

  //  cout << "About to calc bandpart 90" << endl;
  bandpart(ncol, ja, ia, jao, 90,
	   &stats.Ninety_Percent_Bandwidth);

  stats.Ninety_Percent_Bandwidth = stats.Ninety_Percent_Bandwidth*2 + 1;

  //  cout << "About to calc nonvoid diagonals" << endl;
  stats.N_Nonvoid_Diagonals = 0;

  for(i = 0; i < ndiag; i++) {
    if(jao[i] != 0) {
      stats.N_Nonvoid_Diagonals++;
    }
  }

  //  cout << "About to do n_imp_diag" << endl;
  n_imp_diag(&ncol, &nnz, jao, &val, 
		   &stats.N_Important_Diagonals,
		   stats.Most_Important_Diagonals,
		   stats.Important_Diag_Accumulated_Percentages);

  //  cout << "About to do blkfnd_matrix" << endl;
  blkfnd_matrix(ncol, ja, ia, 
		&stats.Block_Size);

  if(stats.Block_Size > 1) {
    stats.Has_Block_Structure = 1;
  } else {
    stats.Has_Block_Structure = 0;
  }
}


// ************************************************************************
//   Operators
// ************************************************************************

/*=========================================================================
#				OPERATOR +		
#				----------    	
# in entry  : left -> a sparse matrix
#	      right -> a sparse matrix 
#
# in return : a sparse matrix : result
#
# details   : calculates result = left + right
#	      	- it checks if the two entries have the same dimension.
#		- it calculates the maximum dimension possibly needed for 
#		  the result (number max of non zero elements)
#		- it allocates the space for a work array
#		- it initializes ja to zero
#		- it adds the 2 matrices by calling the function add_matrices
#		  from the class csparse.c
#		- it determines how many non zero elements there are (by
#		  counting the zeroes of ja)
#		- it returns the matrix
#
=========================================================================*/

sparse operator + (sparse & left, sparse & right) {
  // Add two matrices and return the result.

  #ifdef debugging
    assert (left.get_nrow() == right.get_nrow());
    assert (left.get_ncol() == right.get_ncol());
  #endif

  int i;    // Simple loop array.

  // Calculate the maximum dimension possibly needed for c.
  int nzmax = left.get_nnz() + right.get_nnz();

  int nrow = left.get_nrow();
  int ncol = left.get_ncol();
  int nnz;

  // allocate space for a work array.
  int *iw = new int [left.get_ncol()];

  int *ia = new int[nrow+1];
  int *ja = new int[nzmax];
  double *a = new double[nzmax];

  // Initialize ja to zeros.  This is done, so as to determine the
  // number of nnz elements later.
  for (i = 0; i <= nzmax - 1; i++)
    ja[i] = 0; 

  // Perform the addition.
  add_matrices (left.get_nrow(), left.get_ncol(), left.get_a(),
	        left.get_ja(), left.get_ia(),
                right.get_a(), right.get_ja(), right.get_ia(), 
		a, ja, ia, nzmax, iw);

  // Determine how many nonzero elements there are.
  nnz = 0;
  for (i = 0; ja[i] != 0; i++) 
    nnz += 1;   

  // Free the memory from the work array
  delete [] iw;
		
  return sparse (nrow, ncol, nnz, a, ja, ia);

} // end operator +
  
/*============================================================================
#				OPERATOR -		
#				----------    	
# in entry  : left -> a sparse matrix
#	      right -> a sparse matrix 
#
# in return : a sparse matrix : result
#
# details   : calculates result = left - right
#	      	- it checks if the two entries have the same dimension.
#		- it calculates the maximum dimension possibly needed for 
#		  the result (number max of non zero elements)
#		- it allocates the space for a work array
#		- it initializes ja to zero
#		- it substracts the 2 matrices by calling the function
#		  substract_matrices
#		  from the class csparse.c
#		- it determines how many non zero elements there are (by
#		  counting the zeroes of ja)
#		- it returns the matrix
#
============================================================================*/

sparse operator - (sparse & left, sparse & right) {
  // Subtract two matrices and return the result.

  #ifdef debugging
    assert (left.get_nrow() == right.get_nrow());
    assert (left.get_ncol() == right.get_ncol());
  #endif

  int i;    // Simple loop array.

  // Calculate the maximum dimension possibly needed for c.
  int nzmax = left.get_nnz() + right.get_nnz();

  int nrow = left.get_nrow();
  int ncol = left.get_ncol();
  int nnz;

  int *ia = new int[nrow+1];
  int *ja = new int[nzmax];
  double *a = new double[nzmax];

  // Initialize ja to zeros.  This is done, so as to determine the
  // number of nnz elements later.
  for (i = 0; i <= nzmax - 1; i++)
    ja[i] = 0; 

  // Perform the substraction
  subtract_matrices (left.get_nrow(), left.get_ncol(), left.get_a(),
	             left.get_ja(), left.get_ia(),
                     right.get_a(), right.get_ja(), right.get_ia(), 
		     a, ja, ia, nzmax);

  // Determine how many nonzero elements there are.
  nnz = 0;
  for (i = 0; ja[i] != 0; i++) 
    nnz += 1;   
		
  return sparse (nrow, ncol, nnz, a, ja, ia);

} // end operator -

/*============================================================================
#				OPERATOR -		
#				----------    	
# in entry  : x -> a sparse matrix
#
# in return : a sparse matrix : result
#
# details   : calculates result = -left 
#		- it allocates space for the nex matrix
#		- it copies ia and ja from x
#		- a[i] = -xa[i]
#
============================================================================*/



sparse operator - (sparse &x) {
  // Calculate the negative of a matrix, and return the
  // result.

  int i;                 // Simple loop index

  int nnz = x.get_nnz();
  int nrow = x.get_nrow();
  int ncol = x.get_ncol();

  // allocate space for the new matrix.
  int *ja = new int[nnz];
  int *ia = new int[nrow+1];
  double *a = new double[nnz];

  int *xja = x.get_ja();
  int *xia = x.get_ia();
  double *xa = x.get_a();

  for (i = 0; i <= nrow; i++)
    ia[i] = xia[i];

  for (i = 0; i <= nnz -1; i++) {
    ja[i] = xja[i];
    a[i] = -xa[i];
  }

  return sparse (nrow, ncol, nnz, a, ja, ia);

} // end unary operator -  

/*============================================================================
#				OPERATOR *		
#				----------    	
# in entry  :	s -> a double 
#		x -> a sparse matrix
#
# in return : 	a sparse matrix : result = s*x
#
# details   : 	- it gets the dimensions of the matrix x and allocates enough
#		  space for the new matrix
#		- it stores the contents of the matrix x in xia, xja, xa
#		- it copies xia and xja into the ia and ja of the new matrix
#		- a[i] = s*xa[i] 
#
============================================================================*/

sparse operator * (double s, sparse & x) {
  // Calculate s * x and return the result

  int i;                 // Simple loop index

  int nnz = x.get_nnz();
  int nrow = x.get_nrow();
  int ncol = x.get_ncol();

  // allocate space for the new matrix.
  int *ja = new int[nnz];
  int *ia = new int[nrow+1];
  double *a = new double[nnz];

  int *xja = x.get_ja();
  int *xia = x.get_ia();
  double *xa = x.get_a();

  for (i = 0; i <= nrow; i++)
    ia[i] = xia[i];

  for (i = 0; i <= nnz -1; i++) {
    ja[i] = xja[i];
    a[i] = s*xa[i];
  }

  return sparse (nrow, ncol, nnz, a, ja, ia);

} // end operator *

/*============================================================================
#				OPERATOR *		
#				----------    	
# in entry  :	s -> a double 
#		x -> a sparse matrix
#
# in return : 	a sparse matrix : result = s*x
#
# details   : 	it calls s*x (previous function)
#
============================================================================*/

sparse operator * (sparse & x, double s) {
  // Calculate s * x and return the result

  return s * x;

} // end operator *


sparse operator / (sparse & x, double s) {
  // Calculate x / s and return the result.

  return (1/s) * x;

} // end operator /

/*============================================================================
#				OPERATOR *		
#				----------    	
# in entry  :	left -> a sparse matrix 
#		right -> a sparse matrix
#
# in return : 	a sparse matrix : result = left * right
#
# details   : 	- it checks if the dimensions of the 2 entries are OK
#		- it allocates enough space for the new matrix (as if dense)
#		- it initializes ja to zeros
#		- it calls multiply_matrices from csparse.c
#		- it determines how many non zero element there are (by 
#		  counting the number of non zero elements of ja)
#		- it creates a new matrix with this new number and copies 
#		  the result of multiply_matrices
#
============================================================================*/

sparse operator * (sparse &left, sparse &right) {
  // Calculate left * right and return the result.

  #ifdef debugging
    assert (left.get_ncol() == right.get_nrow());
  #endif
  
  int i;                             // Simple loop index.
  int nnz;
  int nrow = left.get_nrow();
  int ncol = right.get_ncol();

  // Allocate space for the new matrix.  Currently, this routine
  // allocates enough space for a dense matrice.   Hopefully this
  // will be changed later.
  int *ja = new int[nrow * ncol];
  int *ia = new int[nrow+1];
  double *a = new double[nrow * ncol];

  // Initialize ja to zeros.  This is done, so as to determine the
  // number of nnz elements later.
  for (i = 0; i <= ncol*nrow -1; i++)
    ja[i] = 0; 

  // Allocate space for a work array.
  int *iw = new int[left.get_ncol()];

  // Perform the multiplication
  multiply_matrices (nrow, ncol, left.get_a(), left.get_ja(), left.get_ia(),
	             right.get_a(), right.get_ja(), right.get_ia(),
		     a, ja, ia, nrow*ncol, iw);

  // Determine how many nonzero elements there are.
  nnz = 0;
  for (i = 0; ja[i] != 0; i++) 
    nnz += 1;   

  int *jja = new int[nnz];
  double *aa = new double [nnz];

  for (i = 0; i <= nnz -1; i++) {
    jja[i] = ja[i];
    aa[i] = a[i];
  }

  delete [] ja;
  delete [] a;

  return sparse (nrow, ncol, nnz, aa, jja, ia);

} // end operator *


//*********************************************************************
//     Functions
//*********************************************************************

/*============================================================================
#				FILTER		
#				------    	
# in entry  :	mat    -> a sparse matrix
#		funct  -> a function whose argument is a double and which
#		          returns also a double
#
# in return : 	a sparse matrix : result
#
# details   : 	It returns a matrix where each element (i,j) is calculated 
#		by applying the function funct to tje (i,j) element of mat
#		- It checks to make sure that funct(0)=0 (that the result
#	 	  will be a sparse matrix)	
#		- result[i] = funct (mat[i])	
#
============================================================================*/

sparse filter (const sparse &mat, ffunc funct) {
  // Return a new matrix where each element (i,j) is  calculated
  // by applying the function funct to the i,j element of mat.
  // This functions should only be called with functions that return
  // zero on the input zero.  This is since the function is only applied 
  // to the elements of the array ao.

  assert (funct(0) == 0);

  int i;          // Simple loop index.
  int nnz = mat.get_nnz();
  sparse result (mat);

  for (i = 0; i <= nnz-1; i++)
    result.get_a() [i] = funct (mat.get_a()[i]);

  return result;

} // end filter

/*============================================================================
#				OUTPUT :		
#				--------    	
# in entry  : 	mat -> a sparse matrix
#
# in return : 	nothing
#
# details   : 	It outputs the sparse matrix by calling mat.get_element(i,j) 
#
============================================================================*/

void output (sparse & mat) {
  int i, j, klo, khi, k;      // Simple loop counters
  int nrow = mat.get_nrow();  // The row dimension of the matrix.

  int *ia = mat.get_ia();
  int *ja = mat.get_ja();
  double *a = mat.get_a();

  char buf[128];

  for(i = 0; i < nrow; i++) {
    klo = ia[i]-1;
    khi = ia[i+1]-1;

    for(k = klo; k < khi; k++) {
      j = ja[k]-1; 

      if(a[k] != 0.0) {
	sprintf(buf, "(%d,%d)\t%g\n", i+1,j+1,a[k]);
	sp_out(buf);
      }
    }
  }
  
} // end output

/*============================================================================
#				RPERM :		
#				-------    	
# in entry  :	perm -> a pointer to an integer 
#		x -> a sparse matrix
#
# in return : 	a sparse matrix : result
#
# details   : 	It calculates the row permutation of a sparse matrix and 
#		returns the result : It simply calls rperm_matrix from csparse.c
#
============================================================================*/

sparse rperm (sparse &x, int *perm) {
  // Calculate the row permutation of a sparse matrix
  // and return the result.

  sparse y;
  y = x;

  rperm_matrix (x.get_nrow(), x.get_a(), x.get_ja(), x.get_ia(),
		y.get_a(), y.get_ja(), y.get_ia(), perm);

  return y;

} // end rperm

sparse cperm (sparse &x, int *perm) {
  // Calculate the column permutation of a sparse matrix
  // and return the result.

  sparse y;
  y = x;

  cperm_matrix (x.get_nrow(), x.get_a(), x.get_ja(), x.get_ia(),
		y.get_a(), y.get_ja(), y.get_ia(), perm);

  return y;

} // end cperm

sparse dperm (sparse &x, int *perm, int *qperm, int job) {
  // Calculate the row and col permutation of a sparse matrix
  // and return the result.

  sparse y;
  y = x;

  dperm_matrix (x.get_nrow(), x.get_a(), x.get_ja(), x.get_ia(),
		y.get_a(), y.get_ja(), y.get_ia(), perm, qperm, job);

  return y;

} // end dperm


/*============================================================================
#				TRANSP :		
#				-------    	
# in entry  : 	x -> a sparse matrix
#
# in return : 	a sparse matrix : result
#
# details   : 	It calculates the transpose of x and returns the result
#		It simply calls transpose_matrix from csparse.c
#
============================================================================*/

sparse transp (sparse & x) {
  // Calculate the transpose of x and return the result.

  sparse y;
  y = x;
  int *iwk = new int [x.get_nnz()];

  transpose_matrix (y.get_nrow(), y.get_ncol(), y.get_a(), y.get_ja(),
		    y.get_ia(), iwk);

  delete [] iwk;
  return y;
} // end transp

/*============================================================================
#				FLOOR :		
#				-------    	
# in entry  :	x -> a sparse matrix
#
# in return : 	a sparse matrix : result
#
# details   : 	It rounds all the elements of x down and returns the result
#		It just casts each element of x into an integer
#
============================================================================*/

sparse floor (sparse & x) {
  // Round all of the elements of x down and return the result

  sparse y;
  int k;
  y = x;
  int upper = x.get_nnz() -1;

  for (k = 0; k <= upper; k += 1) 
    y.get_a()[k] = int (x.get_a()[k]);

  return y;
} // end floor    

/*============================================================================
#				TRUNC :		
#				-------    	
# in entry  :	x -> a sparse matrix
#
# in return : 	a sparse matrix : result
#
# details   : 	It rounds all the elements of x down and returns the result
#		It just casts each element of x into an integer
#
============================================================================*/

sparse trunc (sparse & x) {
  // Round all of the elements of x down and return the result

  sparse y;
  int k;
  y = x;
  int upper = x.get_nnz() -1;

  for (k = 0; k <= upper; k += 1) 
    y.get_a()[k] = int (x.get_a()[k]);

  return y;
} // end trunc

/*============================================================================
#				CEIL :		
#				------    	
# in entry  :	x -> a sparse matrix
#
# in return : 	a sparse matrix : result
#
# details   : 	It rounds all the elements of x up and returns the result
#		It just calls ceil for each element of x
#
============================================================================*/

sparse ceiling (sparse & x) {
  // Round all of the elements of x up and return the result

  sparse y;
  int k;
  y = x;
  int upper = x.get_nnz() -1;

  for (k = 0; k <= upper; k += 1) 
    y.get_a()[k] = ceil (x.get_a()[k]);

  return y;
} // end ceiling

/*============================================================================
#				ROUND :		
#				-------    	
# in entry  :	x -> a sparse matrix
#
# in return : 	a sparse matrix : result
#
# details   : 	It rounds all the elements of x and returns the result
#		It just casts (each element+0.5) into an integer
#
============================================================================*/

sparse round (sparse & x) {
  // Round all of the elements of x and return the result

  sparse y;
  int k;
  y = x;
  int upper = x.get_nnz() -1;

  for (k = 0; k <= upper; k += 1) 
    y.get_a()[k] = int (x.get_a()[k] + .5);

  return y;
} // end round

/*============================================================================
#				ABS :		
#				-----    	
# in entry  :	x -> a sparse matrix
#
# in return : 	a sparse matrix : result
#
# details   : 	It returns the matrix where each element (i,j) is the abs of 
#		x[i,j]
#
============================================================================*/

sparse abs (sparse & x) {
  // Round all of the elements of x down and return the result

  sparse y;
  int k;
  double value;
  y = x;
  int upper = x.get_nnz() -1;

  for (k = 0; k <= upper; k += 1) {
    value = x.get_a()[k];
    if (value >= 0)
      y.get_a()[k] = value;
    else 
      y.get_a()[k] = -value;
  }

  return y;
} // end floor    

/*==========================================================================
				RANDOM
===========================================================================*/

int irnd (int nb, double x)
  {// Return a random number (integer) between 0 and nb
   // x is a dummy variable
  static int rnd_flag = 0;
  time_t *tloc = NULL;

  // The first time this routine is executed, set the seed
  // the the current time.  This prevents the same sequence from
  // being used twice.
  if (rnd_flag == 0) {
    srand(int (time(tloc)));
    rnd_flag = 1;
  }

  return ((int)((double (rand())/32767.0)*nb));
}

double rnd0nb (int nb, double x)
{ // Returns a random number (double) between 0 and nb
  // x is a dummy variable
static int rnd_flag = 0;
  time_t *tloc = NULL;

  // The first time this routine is executed, set the seed
  // the the current time.  This prevents the same sequence from
  // being used twice.
  if (rnd_flag == 0) {
    srand(int (time(tloc)));
    rnd_flag = 1;
  }
  return ((double (rand())/32767.0)*nb);


}


