/* ***************************************************************
//
//    Class symtab
//
//    Description:
//      Symtab is a class for implementing a symbol table.
//      The symbol table is implemented as an array of linked
//      lists of entries. (see class entry.)  Each linked list
//      corresponds to a unique hash number.
//
//    Remark: spmath uses this class (symtab) to store the variables and the
//	      functions
//
//      If you discover any bugs, it would be very helpful if
//      you sent me e-mail.  Thank you.
//
//      Author:                Peter Braaten
//      e-mail:                braa0007@mermaid.micro.umn.edu
//      Date last modified:    July 18, 1995
//
***************************************************************** */

/*===========================================================================

			PRIVATE :
			---------
    list syms[SIZE];	// array of entry lists 
    symtab *parent;	// each symtab has a parent which is also a symtab

			FUNCTIONS :
			----------

constructors
destructor

is_root : 	returns 1 if the parent is NULL
get_parent :	returns the pointer to the parent symbol table
contains(n) : 	returns 1 if the symbol table contains an entry whose name is n
lookup(n) : 	verifies if the symbol table contains an entry whose name is n
		and returns this entry
out() : 	outputs the contents of the symtab
insert(e) : 	inserts the entry e into the symbol table
remove(n) : 	removes the entry with the name n
remove_all : 	removes all the entries of the symbol table
hash(n) : 	assigns a number to each name (this will be the number of the
		list in which an entry with this name will be stored)

===========================================================================*/


#include "spmath.h"


// ******************************************************************
//       Member Constructors
// ******************************************************************
 
symtab::symtab () {
  // Create an empty symbol table.

  parent = NULL;

} // end constructor symtab

symtab::symtab (symtab *p) {
  // Create an empty symbol table.

  parent = p;

} // end constructor symtab


// ******************************************************************
//       Member Selectors
// ******************************************************************

/*=========================================================================
#				IS_ROOT :
#				---------
# Entries :	nothing 	
#
# Returns : 	an integer -> 1 if parent = NULL
#
=========================================================================*/

int symtab::is_root() {
  // Boolean function.  Return true if parent == NULL
  // otherwise false

  if (parent == NULL)
    return 1;
  else
    return 0;

} // end is_root

/*=========================================================================
#				GET_PARENT :
#				-----------
# Entries :	nothing 	
#
# Returns : 	the pointer to the parent symbol table
#
=========================================================================*/

symtab * symtab::get_parent() {
  // Return the pointer to the parent symbol table

  assert (!is_root());
  
  return parent;
} 

/*=========================================================================
#				LOOKUP :
#				-------
# Entries :	name -> the string to search 	
#
# Returns : 	an entry with the same name
#
# Details :	- It checks to make sure that the symtab contains the name
#		- It calculates the hash number of this name (this is the
#		  number of the list where is stored the name)
#		- it goes to the beginning of this list
#		- it searchs (in this list) an entry with the same name and
#	 	  returns it
#		- If it doesn't find any entry with the same name, it returns
#		  an error
#
=========================================================================*/

entry symtab::lookup (const char *name) {
  // Search for the entry with the same name,
  // and return it's entry.

  #ifdef DEBUGGING
    assert (contains (name));
  #endif

  unsigned int num = hash (name);

  syms[num].init_curr_prev();
  
  if (!syms[num].empty()) {
    if (strcmp (syms[num].get_current().get_name(), name) == 0)
      return syms[num].get_current();
  } // end if

  while (!syms[num].end()) {
    if (strcmp (syms[num].get_next().get_name(), name) == 0)
      return syms[num].get_current();
  } // end while


  // Function should never get this far.
  #ifdef DEBUGGING
    assert (0);
  #endif

    return *(new entry);
} // end selector lookup

/*=========================================================================
#				CONTAINS :
#				---------
# Entries :	name -> the string to search 	
#
# Returns : 	an integer (1 if symtab contains an entry with the same name)
#
# Details :	- It calculates the hash number of this name (this is the
#		  number of the only list where can be stored the name)
#		- it goes to the beginning of this list
#		- it searchs (in this list) an entry with the same name 
#		- If it has found something, it returns 1, else 0
#
=========================================================================*/

int symtab::contains (const char *name) {
  // Search for the entry with the name name,
  // if it is present, return 1, otherwise 0.

  int num = hash (name);

  syms[num].init_curr_prev();

  if (!syms[num].empty()) {
    if (strcmp (syms[num].get_current().get_name(), name) == 0)
      return 1;
  } // end if

  while (!syms[num].end()) {
    if (strcmp (syms[num].get_next().get_name(), name) == 0)
      return 1;
  } // end while

  // If the function got this far, then the name is not present.
  return 0;

} // end selector lookup

/*=========================================================================
#				OUTPUT :
#				-------
# Entries :	nothing 	
#
# Returns : 	nothing
#
# Details :	It outputs the contents of the symbol table
#
=========================================================================*/

void symtab::out () {

  // Output the contents of the symbol table.

  int i;

  sp_out("Current Variables:\n\n");

  for (i = 0; i < SIZE; i += 1) {

    entry e;
    syms[i].init_curr_prev();

    if(!syms[i].empty()) {
      e = syms[i].get_current();
      sp_out(e.get_name()); sp_out("\n");
    }
    while(!syms[i].end()) {
      e = syms[i].get_next();
      sp_out(e.get_name()); sp_out("\n");
    }
  }

  sp_out("\n");
}

void symtab::out_fancy() {

  int i;

  sp_out("  Name                 Size        Bytes  Class\n\n");

  for (i = 0; i <= SIZE-1; i += 1) {

    char buf[128];

    entry e;
    syms[i].init_curr_prev();

    if(!syms[i].empty()) {
      e = syms[i].get_current();

      int spaces;
      value v = e.get_value();

      strcpy(buf, "  ");
      strcat(buf, e.get_name());
      buf[22]='\0';

      spaces = 23-strlen(buf);
      for(int s = 0; s < spaces; s++)
	strcat(buf, " ");

      switch(v.type) {

      case NUM:	

	strcat(buf, "1x1       ");
	sprintf(buf, "%s%7d  number", buf, sizeof(double)); 
	sp_out(buf);
	break;

      case DENSE:

	sprintf(buf, "%s%dx%d", buf, 
		v.dmat->get_nrow(), v.dmat->get_ncol());

	spaces = 33-strlen(buf);
	for(int s = 0; s < spaces; s++)
	  strcat(buf, " ");

	sprintf(buf, "%s%7d  dense matrix", buf, v.dmat->size()); 

	sp_out(buf);	
	break;

      case SPARSE:

	sprintf(buf, "%s%dx%d", buf, 
		v.smat->get_nrow(), v.smat->get_ncol());

	spaces = 33-strlen(buf);
	for(int s = 0; s < spaces; s++)
	  strcat(buf, " ");

	sprintf(buf, "%s%7d  sparse matrix", buf, v.smat->size()); 

	sp_out(buf);	
	break;

      case STRING:

	sprintf(buf, "%s%d", buf, strlen(v.str));

	spaces = 33-strlen(buf);
	for(int s = 0; s < spaces; s++)
	  strcat(buf, " ");

	sprintf(buf, "%s%7d  string", buf, strlen(v.get_str()));

	sp_out(buf);	
	break;

      case FUNCT:

      	strcat(buf, "          ");
	sprintf(buf, "%s%7d  function", buf, v.funct->size()); 
	sp_out(buf);
	break;

      }
      sp_out("\n");
    }

    while(!syms[i].end()) {

      e = syms[i].get_next();

      int spaces;
      value v = e.get_value();

      strcpy(buf, "  ");
      strcat(buf, e.get_name());
      buf[22]='\0';

      spaces = 23-strlen(buf);
      for(int s = 0; s < spaces; s++)
	strcat(buf, " ");

      switch(v.type) {

      case NUM:	

	strcat(buf, "1x1       ");
	sprintf(buf, "%s%7d  number", buf, sizeof(double)); 
	sp_out(buf);
	break;

      case DENSE:

	sprintf(buf, "%s%dx%d", buf, 
		v.dmat->get_nrow(), v.dmat->get_ncol());

	spaces = 33-strlen(buf);
	for(int s = 0; s < spaces; s++)
	  strcat(buf, " ");

	sprintf(buf, "%s%7d  dense matrix", buf, v.dmat->size()); 

	sp_out(buf);	
	break;

      case SPARSE:

	sprintf(buf, "%s%dx%d", buf, 
		v.smat->get_nrow(), v.smat->get_ncol());

	spaces = 33-strlen(buf);
	for(int s = 0; s < spaces; s++)
	  strcat(buf, " ");

	sprintf(buf, "%s%7d  sparse matrix", buf, v.smat->size()); 

	sp_out(buf);	
	break;

      case STRING:

	sprintf(buf, "%s%d", buf, strlen(v.str));

	spaces = 33-strlen(buf);
	for(int s = 0; s < spaces; s++)
	  strcat(buf, " ");

	sprintf(buf, "%s%7d  string", buf, strlen(v.get_str()));

	sp_out(buf);	
	break;

      case FUNCT:

      	strcat(buf, "          ");
	sprintf(buf, "%s%7d  function", buf, v.funct->size()); 
	sp_out(buf);
	break;

      }
      sp_out("\n");
    }
  }
}

void symtab::out_alias () {
  // Output the contents of the symbol table. (alias)

  int i;

  for (i = 0; i <= SIZE-1; i += 1) {
    syms[i].init_curr_prev();
    if (!syms[i].empty()) {
      sp_out(syms[i].get_current().get_name());
      sp_out("\t\t");
      sp_out(syms[i].get_current().get_value().str);
      sp_out("\n");
    } // endif
    while (!syms[i].end()) {
      sp_out(syms[i].get_next().get_name());
      sp_out("\t\t");
      sp_out(syms[i].get_current().get_value().str);
      sp_out("\n");
    } // end while
  } // end for

} // end output

//******************************************************************
//       Member Mutators
//******************************************************************

/*=========================================================================
#				INSERT :
#				-------
# Entries :	entry -> the entry to insert 	
#
# Returns : 	nothing
#
# Details :	It inserts the entry into the symbol table
#		- If symtab CONTAINS the name of this entry
#		    + it removes this entry from the symbol table
#		    + it calculates h : the hash number of the name of the entry
#		    + it inserts the netry at the beginning of the list of
#		      symtab (whose number is h )
#		- else
#		    + it does the same thing without removing anything
#
=========================================================================*/

void symtab::insert(entry ent) {
  // Insert the entry into the symbol table

  if (contains(ent.get_name())) {
    remove (ent.get_name());
    syms[hash(ent.get_name())].push(ent);
  } // end if
  else
    syms[hash(ent.get_name())].push(ent);

} // end insert

/*=========================================================================
#				REMOVE :
#				-------
# Entries :	name -> the string to remove 	
#
# Returns : 	nothing
#
# Details :	- It checks to make sure that the symtab contains the name
#		- It calculates h : the hash number of this name (this is the
#		  number of the list where is stored the name)
#		- It looks if the current item of the list h has the same name
#		  If it's the case, it removes it
#		  Otherwise, it goes to the next item of the list and restart 
#
=========================================================================*/

void symtab::remove (const char *name) {
  // Remove the entry with the name name.

  #ifdef DEBUGGING
    assert (contains (name));
  #endif

  unsigned int num = hash (name);
  
  syms[num].init_curr_prev();
  
  if (!syms[num].empty()) {
    if (strcmp (syms[num].get_current().get_name(), name) == 0) {
      syms[num].remove_current();
      return;
    } // end if
  } // end if

  while (!syms[num].end()) {
    if (strcmp (syms[num].get_next().get_name(), name) == 0) {
      syms[num].remove_current();
      return;
    } // end if
  } // end while


  // Function should never get this far.
  #ifdef DEBUGGING
    assert (0);
  #endif

} // end remove

/*=========================================================================
#				REMOVE_ALL :
#				-----------
# Entries :	nothing 	
#
# Returns : 	nothing
#
# Details :	It removes all the elements of the symbol table by calling
#		the function remove_all (of list.C) for each list  
#
=========================================================================*/

void symtab::remove_all() {
  // Empty out the symbol table.

  int i;

  for (i = 0; i <= SIZE - 1; i += 1) 
    syms[i].remove_all();

} // end remove_all


//******************************************************************
//       Functions
//******************************************************************

/*==================================================================
#				HASH :
#				-----
# Entries :	name -> a string 	
#
# Returns : 	an integer
#
# Details :	It assigns a number to each name (the entry whith 
#               this name will be stored in the list whose number is 
#               this hash code)
#
===================================================================*/

unsigned int hash (const char *name) {
  // Hash function.  Assign a number to each name.

  unsigned int i;
  unsigned int h = 0;
  int alpha = 3;

  for (i = 0; i <= strlen(name) - 1; i += 1) {
    h = h*alpha + name[i];
  } // end for

  return h % (SIZE - 1);

} // end hash        
 

